=== modified file '.bzrignore'
--- .bzrignore	2010-09-30 06:24:20 +0000
+++ .bzrignore	2012-05-17 01:55:58 +0000
@@ -3,6 +3,7 @@
 *.8mandos
 confdir
 keydir
+statedir
 man
 plugin-runner
 plugins.d/askpass-fifo

=== modified file 'INSTALL'
--- INSTALL	2011-03-08 19:09:03 +0000
+++ INSTALL	2014-06-22 02:19:30 +0000
@@ -12,7 +12,7 @@
    server and client programs themselves *could* be run in other
    distributions, but they *are* specific to GNU/Linux systems, and
    are not written with portabillity to other Unixes in mind.
-  
+   
 ** Libraries
    
    The following libraries and packages are needed.  (It is possible
@@ -35,7 +35,7 @@
     To build just the documentation, run the command "make doc".  Then
     the manual page "mandos.8", for example, can be read by running
     "man -l mandos.8".
-   
+    
 *** Mandos Server
     + GnuTLS 2.4          http://www.gnu.org/software/gnutls/
     + Avahi 0.6.16        http://www.avahi.org/
@@ -47,11 +47,12 @@
     
     Strongly recommended:
     + fping 2.4b2-to-ipv6 http://www.fping.com/
+    + ssh-keyscan from OpenSSH http://www.openssh.com/
     
     Package names:
     python-gnutls avahi-daemon python python-avahi python-dbus
-    python-gobject python-urwid
-   
+    python-gobject python-urwid ssh-client
+    
 *** Mandos Client
     + initramfs-tools 0.85i
                   http://packages.qa.debian.org/i/initramfs-tools.html
@@ -60,9 +61,12 @@
     + GnuPG 1.4.9         http://www.gnupg.org/
     + GPGME 1.1.6         http://www.gnupg.org/related_software/gpgme/
     
+    Strongly recommended:
+    + OpenSSH             http://www.openssh.com/
+    
     Package names:
     initramfs-tools libgnutls-dev libavahi-core-dev gnupg
-    libgpgme11-dev
+    libgpgme11-dev ssh
 
 * Installing the Mandos server
   
@@ -132,6 +136,6 @@
   
   You may want to tighten or loosen the timeouts in the server
   configuration files; see mandos.conf(5) and mandos-clients.conf(5).
-  If IPsec is not used, it is suggested that a more cryptographically
-  secure checker program is used and configured, since without IPsec
-  ping packets can be faked.
+  If IPsec is not used and SSH is not installed, it is suggested that
+  a more cryptographically secure checker program is used and
+  configured, since, without IPsec, ping packets can be faked.

=== modified file 'Makefile'
--- Makefile	2012-01-15 21:07:44 +0000
+++ Makefile	2014-06-22 02:19:30 +0000
@@ -1,14 +1,17 @@
-WARN=-O -Wall -Wformat=2 -Winit-self -Wmissing-include-dirs \
-	-Wswitch-default -Wswitch-enum -Wunused-parameter \
-	-Wstrict-aliasing=1 -Wextra -Wfloat-equal -Wundef -Wshadow \
+WARN=-O -Wall -Wextra -Wdouble-promotion -Wformat=2 -Winit-self \
+	-Wmissing-include-dirs -Wswitch-default -Wswitch-enum \
+	-Wunused -Wuninitialized -Wstrict-overflow=5 \
+	-Wsuggest-attribute=pure -Wsuggest-attribute=const \
+	-Wsuggest-attribute=noreturn -Wfloat-equal -Wundef -Wshadow \
 	-Wunsafe-loop-optimizations -Wpointer-arith \
 	-Wbad-function-cast -Wcast-qual -Wcast-align -Wwrite-strings \
-	-Wconversion -Wstrict-prototypes -Wold-style-definition \
-	-Wpacked -Wnested-externs -Winline -Wvolatile-register-var
-#	-Wunreachable-code
+	-Wconversion -Wlogical-op -Waggregate-return \
+	-Wstrict-prototypes -Wold-style-definition \
+	-Wmissing-format-attribute -Wnormalized=nfc -Wpacked \
+	-Wredundant-decls -Wnested-externs -Winline -Wvla \
+	-Wvolatile-register-var -Woverlength-strings
 #DEBUG=-ggdb3
-# For info about _FORTIFY_SOURCE, see
-# <http://www.kernel.org/doc/man-pages/online/pages/man7/feature_test_macros.7.html>
+# For info about _FORTIFY_SOURCE, see feature_test_macros(7)
 # and <http://gcc.gnu.org/ml/gcc-patches/2004-09/msg02055.html>.
 FORTIFY=-D_FORTIFY_SOURCE=2 -fstack-protector-all -fPIC
 LINK_FORTIFY_LD=-z relro -z now
@@ -20,10 +23,10 @@
 LINK_FORTIFY += -pie
 endif
 #COVERAGE=--coverage
-OPTIMIZE=-Os
+OPTIMIZE=-Os -fno-strict-aliasing
 LANGUAGE=-std=gnu99
 htmldir=man
-version=1.5.3
+version=1.6.5
 SED=sed
 
 USER=$(firstword $(subst :, ,$(shell getent passwd _mandos || getent passwd nobody || echo 65534)))
@@ -36,6 +39,7 @@
 # MANDIR=$(PREFIX)/man
 # INITRAMFSTOOLS=$(DESTDIR)/etc/initramfs-tools
 # STATEDIR=$(DESTDIR)/var/lib/mandos
+# LIBDIR=$(PREFIX)/lib
 ##
 
 ## These settings are for a package-type install
@@ -45,8 +49,19 @@
 MANDIR=$(PREFIX)/share/man
 INITRAMFSTOOLS=$(DESTDIR)/usr/share/initramfs-tools
 STATEDIR=$(DESTDIR)/var/lib/mandos
+LIBDIR=$(shell \
+	for d in \
+	"/usr/lib/`dpkg-architecture -qDEB_HOST_MULTIARCH 2>/dev/null`" \
+	"`rpm --eval='%{_libdir}' 2>/dev/null`" /usr/lib; do \
+		if [ -d "$$d" -a "$$d" = "$${d%/}" ]; then \
+			echo "$(DESTDIR)$$d"; \
+			break; \
+		fi; \
+	done)
 ##
 
+SYSTEMD=$(DESTDIR)$(shell pkg-config systemd --variable=systemdsystemunitdir)
+
 GNUTLS_CFLAGS=$(shell pkg-config --cflags-only-I gnutls)
 GNUTLS_LIBS=$(shell pkg-config --libs gnutls)
 AVAHI_CFLAGS=$(shell pkg-config --cflags-only-I avahi-core)
@@ -56,10 +71,10 @@
 	getconf LFS_LDFLAGS)
 
 # Do not change these two
-CFLAGS=$(WARN) $(DEBUG) $(FORTIFY) $(COVERAGE) $(OPTIMIZE) \
+CFLAGS+=$(WARN) $(DEBUG) $(FORTIFY) $(COVERAGE) $(OPTIMIZE) \
 	$(LANGUAGE) $(GNUTLS_CFLAGS) $(AVAHI_CFLAGS) $(GPGME_CFLAGS) \
 	-DVERSION='"$(version)"'
-LDFLAGS=-Xlinker --as-needed $(COVERAGE) $(LINK_FORTIFY) $(foreach flag,$(LINK_FORTIFY_LD),-Xlinker $(flag))
+LDFLAGS+=-Xlinker --as-needed $(COVERAGE) $(LINK_FORTIFY) $(foreach flag,$(LINK_FORTIFY_LD),-Xlinker $(flag))
 
 # Commands to format a DocBook <refentry> document into a manual page
 DOCBOOKTOMAN=$(strip cd $(dir $<); xsltproc --nonet --xinclude \
@@ -68,13 +83,12 @@
 	--param make.single.year.ranges		1 \
 	--param man.output.quietly		1 \
 	--param man.authors.section.enabled	0 \
-	 /usr/share/xml/docbook/stylesheet/nwalsh/manpages/docbook.xsl \
+	/usr/share/xml/docbook/stylesheet/nwalsh/manpages/docbook.xsl \
 	$(notdir $<); \
-	$(MANPOST) $(notdir $@);\
-	LANG=en_US.UTF-8 MANWIDTH=80 man --warnings --encoding=UTF-8 \
-	--local-file $(notdir $@) >/dev/null)
-# DocBook-to-man post-processing to fix a '\n' escape bug
-MANPOST=$(SED) --in-place --expression='s,\\\\en,\\en,g;s,\\n,\\en,g'
+	if locale --all 2>/dev/null | grep --regexp='^en_US\.utf8$$' \
+	&& type man 2>/dev/null; then LANG=en_US.UTF-8 MANWIDTH=80 \
+	man --warnings --encoding=UTF-8 --local-file $(notdir $@); \
+	fi >/dev/null)
 
 DOCBOOKTOHTML=$(strip xsltproc --nonet --xinclude \
 	--param make.year.ranges		1 \
@@ -225,9 +239,11 @@
 	$(LINK.c) $^ -lrt $(GNUTLS_LIBS) $(AVAHI_LIBS) $(strip\
 		) $(GPGME_LIBS) $(LOADLIBES) $(LDLIBS) -o $@
 
-.PHONY : all doc html clean distclean run-client run-server install \
-	install-server install-client uninstall uninstall-server \
-	uninstall-client purge purge-server purge-client
+.PHONY : all doc html clean distclean mostlyclean maintainer-clean \
+	check run-client run-server install install-html \
+	install-server install-client-nokey install-client uninstall \
+	uninstall-server uninstall-client purge purge-server \
+	purge-client
 
 clean:
 	-rm --force $(CPROGS) $(objects) $(htmldocs) $(DOCS) core
@@ -239,6 +255,7 @@
 
 check:	all
 	./mandos --check
+	./mandos-ctl --check
 
 # Run the client with a local config and key
 run-client: all keydir/seckey.txt keydir/pubkey.txt
@@ -247,15 +264,18 @@
 	@echo "# ignored.  The messages are caused by not running as root, but   #"
 	@echo "# you should NOT run \"make run-client\" as root unless you also    #"
 	@echo "# unpacked and compiled Mandos as root, which is NOT recommended. #"
-	@echo "# From plugin-runner: setuid: Operation not permitted             #"
+	@echo "# From plugin-runner: setgid: Operation not permitted             #"
+	@echo "#                     setuid: Operation not permitted             #"
 	@echo "# From askpass-fifo:  mkfifo: Permission denied                   #"
-	@echo "# From mandos-client: setuid: Operation not permitted             #"
-	@echo "#                     seteuid: Operation not permitted            #"
-	@echo "#                     klogctl: Operation not permitted            #"
+	@echo "# From mandos-client:                                             #"
+	@echo "#             Failed to raise privileges: Operation not permitted #"
+	@echo "#             Warning: network hook \"*\" exited with status *      #"
 	@echo "###################################################################"
+# We set GNOME_KEYRING_CONTROL to block pam_gnome_keyring
 	./plugin-runner --plugin-dir=plugins.d \
 		--config-file=plugin-runner.conf \
 		--options-for=mandos-client:--seckey=keydir/seckey.txt,--pubkey=keydir/pubkey.txt,--network-hook-dir=network-hooks.d \
+		--env-for=mandos-client:GNOME_KEYRING_CONTROL= \
 		$(CLIENTARGS)
 
 # Used by run-client
@@ -264,7 +284,7 @@
 	./mandos-keygen --dir keydir --force
 
 # Run the server with a local config
-run-server: confdir/mandos.conf confdir/clients.conf
+run-server: confdir/mandos.conf confdir/clients.conf statedir
 	./mandos --debug --no-dbus --configdir=confdir \
 		--statedir=statedir $(SERVERARGS)
 
@@ -276,7 +296,7 @@
 	install --directory confdir
 	install --mode=u=rw $< $@
 # Add a client password
-	./mandos-keygen --dir keydir --password >> $@
+	./mandos-keygen --dir keydir --password --no-ssh >> $@
 statedir:
 	install --directory statedir
 
@@ -289,8 +309,12 @@
 
 install-server: doc
 	install --directory $(CONFDIR)
-	install --directory --mode=u=rwx --owner=$(USER) \
-		--group=$(GROUP) $(STATEDIR)
+	if install --directory --mode=u=rwx --owner=$(USER) \
+		--group=$(GROUP) $(STATEDIR); then \
+		:; \
+	elif install --directory --mode=u=rwx $(STATEDIR); then \
+		chown -- $(USER):$(GROUP) $(STATEDIR) || :; \
+	fi
 	install --mode=u=rwx,go=rx mandos $(PREFIX)/sbin/mandos
 	install --mode=u=rwx,go=rx --target-directory=$(PREFIX)/sbin \
 		mandos-ctl
@@ -304,6 +328,9 @@
 		$(DESTDIR)/etc/dbus-1/system.d/mandos.conf
 	install --mode=u=rwx,go=rx init.d-mandos \
 		$(DESTDIR)/etc/init.d/mandos
+	if [ "$(SYSTEMD)" != "$(DESTDIR)" -a -d "$(SYSTEMD)" ]; then \
+		install --mode=u=rw,go=r mandos.service $(SYSTEMD); \
+	fi
 	install --mode=u=rw,go=r default-mandos \
 		$(DESTDIR)/etc/default/mandos
 	if [ -z $(DESTDIR) ]; then \
@@ -323,36 +350,36 @@
 		> $(MANDIR)/man8/intro.8mandos.gz
 
 install-client-nokey: all doc
-	install --directory $(PREFIX)/lib/mandos $(CONFDIR)
+	install --directory $(LIBDIR)/mandos $(CONFDIR)
 	install --directory --mode=u=rwx $(KEYDIR) \
-		$(PREFIX)/lib/mandos/plugins.d
-	if [ "$(CONFDIR)" != "$(PREFIX)/lib/mandos" ]; then \
+		$(LIBDIR)/mandos/plugins.d
+	if [ "$(CONFDIR)" != "$(LIBDIR)/mandos" ]; then \
 		install --mode=u=rwx \
 			--directory "$(CONFDIR)/plugins.d"; \
 	fi
 	install --mode=u=rwx,go=rx --directory \
 		"$(CONFDIR)/network-hooks.d"
 	install --mode=u=rwx,go=rx \
-		--target-directory=$(PREFIX)/lib/mandos plugin-runner
+		--target-directory=$(LIBDIR)/mandos plugin-runner
 	install --mode=u=rwx,go=rx --target-directory=$(PREFIX)/sbin \
 		mandos-keygen
 	install --mode=u=rwx,go=rx \
-		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
+		--target-directory=$(LIBDIR)/mandos/plugins.d \
 		plugins.d/password-prompt
 	install --mode=u=rwxs,go=rx \
-		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
+		--target-directory=$(LIBDIR)/mandos/plugins.d \
 		plugins.d/mandos-client
 	install --mode=u=rwxs,go=rx \
-		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
+		--target-directory=$(LIBDIR)/mandos/plugins.d \
 		plugins.d/usplash
 	install --mode=u=rwxs,go=rx \
-		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
+		--target-directory=$(LIBDIR)/mandos/plugins.d \
 		plugins.d/splashy
 	install --mode=u=rwxs,go=rx \
-		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
+		--target-directory=$(LIBDIR)/mandos/plugins.d \
 		plugins.d/askpass-fifo
 	install --mode=u=rwxs,go=rx \
-		--target-directory=$(PREFIX)/lib/mandos/plugins.d \
+		--target-directory=$(LIBDIR)/mandos/plugins.d \
 		plugins.d/plymouth
 	install initramfs-tools-hook \
 		$(INITRAMFSTOOLS)/hooks/mandos
@@ -404,13 +431,13 @@
 	! grep --regexp='^ *[^ #].*keyscript=[^,=]*/mandos/' \
 		$(DESTDIR)/etc/crypttab
 	-rm --force $(PREFIX)/sbin/mandos-keygen \
-		$(PREFIX)/lib/mandos/plugin-runner \
-		$(PREFIX)/lib/mandos/plugins.d/password-prompt \
-		$(PREFIX)/lib/mandos/plugins.d/mandos-client \
-		$(PREFIX)/lib/mandos/plugins.d/usplash \
-		$(PREFIX)/lib/mandos/plugins.d/splashy \
-		$(PREFIX)/lib/mandos/plugins.d/askpass-fifo \
-		$(PREFIX)/lib/mandos/plugins.d/plymouth \
+		$(LIBDIR)/mandos/plugin-runner \
+		$(LIBDIR)/mandos/plugins.d/password-prompt \
+		$(LIBDIR)/mandos/plugins.d/mandos-client \
+		$(LIBDIR)/mandos/plugins.d/usplash \
+		$(LIBDIR)/mandos/plugins.d/splashy \
+		$(LIBDIR)/mandos/plugins.d/askpass-fifo \
+		$(LIBDIR)/mandos/plugins.d/plymouth \
 		$(INITRAMFSTOOLS)/hooks/mandos \
 		$(INITRAMFSTOOLS)/conf-hooks.d/mandos \
 		$(INITRAMFSTOOLS)/scripts/init-premount/mandos \
@@ -422,8 +449,8 @@
 		$(MANDIR)/man8/splashy.8mandos.gz \
 		$(MANDIR)/man8/askpass-fifo.8mandos.gz \
 		$(MANDIR)/man8/plymouth.8mandos.gz \
-	-rmdir $(PREFIX)/lib/mandos/plugins.d $(CONFDIR)/plugins.d \
-		 $(PREFIX)/lib/mandos $(CONFDIR) $(KEYDIR)
+	-rmdir $(LIBDIR)/mandos/plugins.d $(CONFDIR)/plugins.d \
+		 $(LIBDIR)/mandos $(CONFDIR) $(KEYDIR)
 	update-initramfs -k all -u
 
 purge: purge-server purge-client
@@ -433,6 +460,8 @@
 		$(DESTDIR)/etc/dbus-1/system.d/mandos.conf
 		$(DESTDIR)/etc/default/mandos \
 		$(DESTDIR)/etc/init.d/mandos \
+		$(SYSTEMD)/mandos.service \
+		$(DESTDIR)/run/mandos.pid \
 		$(DESTDIR)/var/run/mandos.pid
 	-rmdir $(CONFDIR)
 

=== modified file 'NEWS'
--- NEWS	2012-01-15 21:07:44 +0000
+++ NEWS	2014-05-11 20:18:48 +0000
@@ -1,6 +1,83 @@
 This NEWS file records noteworthy changes, very tersely.
 See the manual for detailed information.
 
+Version 1.6.5 (2014-05-11)
+* Client
+** Work around bug in GnuPG <http://bugs.g10code.com/gnupg/issue1622>
+** Give better error messages when run without sufficient privileges
+** Only warn if workaround for Debian bug #633582 was necessary and
+   failed, not if it failed and was unnecessary.
+
+Version 1.6.4 (2014-02-16)
+* Server
+** Very minor fix to self-test code.
+
+Version 1.6.3 (2014-01-21)
+* Server
+** Add systemd support.
+** For PID file, fall back to /var/run if /run does not exist.
+* Client
+** Moved files from /usr/lib/mandos to whatever the architecture
+   specifies, like /usr/lib/x86_64-linux-gnu/mandos or
+   /usr/lib64/mandos.
+
+Version 1.6.2 (2013-10-24)
+* Server
+** PID file moved from /var/run to /run.
+** Bug fix: Handle long secrets when saving client state.
+** Bug fix: Use more magic in the GnuTLS priority string to handle
+   both old DSA/ELG 2048-bit keys and new RSA/RSA 4096-bit keys.
+* Client
+** mandos-keygen: Bug fix: now generate RSA keys which GnuTLS can use.
+   		  Bug fix: Output passphrase prompts even when
+   		  redirecting standard output.
+
+Version 1.6.1 (2013-10-13)
+* Server
+** All client options for time intervals now also take an RFC 3339
+   duration.  The same for all options to mandos-ctl.
+** Bug fix: Handle fast checkers (like ":") correctly.
+** Bug fix: Don't print output from checkers when running in
+   foreground.
+** Bug fix: Do not fail when client is removed from clients.conf but
+   saved settings remain.
+** Bug fix: mandos-monitor now displays standout (reverse video) again
+   using new version of Urwid.
+** Bug fix: Make boolean options work from the config file again.
+** Bug fix: Make --no-ipv6 work again.
+** New default priority string to be slightly more compatible with
+   older versions of GnuTLS.
+* Client
+** Bug fix: Fix bashism in mandos-keygen.
+** Default key and subkey types are now RSA and RSA, respectively.
+   Also, new default key size is 4096 bits.
+
+Version 1.6.0 (2012-06-18)
+* Server
+** Takes new --foreground option
+** Init script supports new "status" action.
+* Client
+** Now uses all interfaces by default; the --interface option can
+   still be used to restrict it, and the argument to --interface (as
+   well as the $DEVICE environment variable for the network hooks) is
+   now a comma-separated list of interfaces to use.
+
+Version 1.5.5 (2012-06-01)
+* Server
+** Server takes new --socket option
+
+Version 1.5.4 (2012-05-20)
+* Server
+** Bug fix: Regression fix: Make non-zero approval timeout values work.
+** Bug fix: Regression fix: Allow changing the Timeout D-Bus property.
+** Fall back to not bind to an interface if an invalid interface name
+   is given.
+** Removed support for undocumented feature of using plain "%%s" in
+   "checker" client option.
+** Old D-Bus interface are now marked as deprecated.
+** mandos-monitor: Bug fix: show approval timers correctly.
+** mandos-ctl: Show "Extended Timeout" correctly, not as milliseconds.
+
 Version 1.5.3 (2012-01-15)
 * Server
 ** Add D-Bus property se.recompile.Client.LastCheckerStatus and use it

=== modified file 'TODO'
--- TODO	2012-01-15 20:27:28 +0000
+++ TODO	2014-06-22 02:19:30 +0000
@@ -1,16 +1,28 @@
 -*- org -*-
 
+* GIT
+** General: [[https://www.atlassian.com/git/workflows][Git Workflows]], [[http://gitimmersion.com/][Git Immersion]], [[https://news.ycombinator.com/item?id=7036628][Simple git workflow is simple]]
+** Intro: [[http://www.eyrie.org/~eagle/notes/debian/git.html#combine][Using Git for Debian Packaging]]
+** Use: [[https://honk.sigxcpu.org/piki/projects/git-buildpackage/][git-buildpackage]]
+** Migration
+   tailor?
+   Using bzr-fastimport: [[http://www.fusonic.net/en/blog/2013/03/26/migrating-from-bazaar-to-git/][Migrating from Bazaar to Git]]
+** Unresolved: [[http://jameswestby.net/bzr/builddeb/user_manual/split.html][bzr builddeb split mode]]
+   Maybe: [[http://honk.sigxcpu.org/projects/git-buildpackage/manual-html/gbp.import.html#GBP.IMPORT.UPSTREAM.GIT.NOTARBALL][git-buildpackage - No upstream tarballs]]
+   [[http://www.python.org/dev/peps/pep-0374/][PEP 374 - Choosing a distributed VCS for the Python project]]
+   [[http://www.emacswiki.org/emacs/GitForEmacsDevs][Git For Emacs Devs]]
+
 * [[http://www.undeadly.org/cgi?action=article&sid=20110530221728][OpenBSD]]
 
+* Testing
+** python-nemu
+
 * mandos-applet
 
 * mandos-client
-** TODO [#A] Wireless network hook
 ** TODO [#B] Use capabilities instead of seteuid().
-** TODO [#B] Use struct sockaddr_storage instead of a union
 ** TODO [#B] Use getaddrinfo(hints=AI_NUMERICHOST) instead of inet_pton()
-** TODO [#B] Use getnameinfo(serv=NULL, NI_NUMERICHOST) instead of inet_ntop()
-** TODO [#B] Prefer /run/tmp over /tmp, if it exists
+** TODO [#C] Make start_mandos_communication() take "struct server".
 
 * splashy
 ** TODO [#B] use scandir(3) instead of readdir(3)
@@ -18,7 +30,6 @@
 * usplash (Deprecated)
 ** TODO [#A] Make it work again
 ** TODO [#B] use scandir(3) instead of readdir(3)
-** TODO Use [[info:libc:Argz%20Functions][argz_extract]]
 
 * askpass-fifo
 ** TODO [#B] Drop privileges after opening FIFO.
@@ -32,36 +43,24 @@
 
 * plugin-runner
 ** TODO handle printing for errors for plugins
-*** Hook up stderr of plugins, buffer them, and prepend mandos pluig [plugin name]
-** TODO [#B] use scandir(3) instead of readdir(3)
+*** Hook up stderr of plugins, buffer them, and prepend "Mandos Plugin [plugin name]"
 ** TODO [#C] use same file name rules as run-parts(8)
 ** kernel command line option for debug info
-** TODO [#B] Use openat()
 
 * mandos (server)
-** TODO Document why we ignore sigint
 ** TODO [#B] Log level							  :BUGS:
 *** TODO /etc/mandos/clients.d/*.conf
     Watch this directory and add/remove/update clients?
 ** TODO [#C] config for TXT record
 ** TODO Log level dbus option
    SetLogLevel D-Bus call
-** TODO Implement --foreground						  :BUGS:
-   [[info:standards:Option%20Table][Table of Long Options]]
-** TODO Implement --socket
-   [[info:standards:Option%20Table][Table of Long Options]]
 ** TODO [#C] DBusServiceObjectUsingSuper
 ** TODO [#B] Global enable/disable flag
 ** TODO [#B] By-client countdown on number of secrets given
-** TODO [#B] Support RFC 3339 time duration syntax
-** More D-Bus methods
-*** NeedsPassword(50) - Timeout, default disapprove
+** D-Bus Client method NeedsPassword(50) - Timeout, default disapprove
     + SetPass(u"gazonk", True) -> Approval, persistent
     + Approve(False) -> Close client connection immediately
 ** TODO [#C] python-parsedatetime
-** TODO [#C] systemd/launchd
-   http://0pointer.de/blog/projects/systemd.html
-   http://wiki.debian.org/systemd
 ** TODO Separate logging logic to own object
 ** TODO [#A] Limit approval_delay to max gnutls/tls timeout value
 ** TODO [#B] break the wait on approval_delay if connection dies
@@ -69,15 +68,24 @@
 ** TODO Allow %%(checker)s as a runtime expansion
 ** TODO Use python-tlslite?
 ** TODO D-Bus AddClient() method on server object
-** TODO Use org.freedesktop.DBus.Method.NoReply annotation on async methods.
-** TODO Emit [[http://dbus.freedesktop.org/doc/dbus-specification.html#standard-interfaces-properties][org.freedesktop.DBus.Properties.PropertiesChanged]] signal
+** TODO Use org.freedesktop.DBus.Method.NoReply annotation on async methods. :2:
+** TODO Emit [[http://dbus.freedesktop.org/doc/dbus-specification.html#standard-interfaces-properties][org.freedesktop.DBus.Properties.PropertiesChanged]] signal	  :2:
    TODO Deprecate se.recompile.Mandos.Client.PropertyChanged - annotate!
    TODO Can use "invalidates" annotation to also emit on changed secret.
-** TODO Support [[http://dbus.freedesktop.org/doc/dbus-specification.html#standard-interfaces-objectmanager][org.freedesktop.DBus.ObjectManager]] interface on server object
+** TODO Support [[http://dbus.freedesktop.org/doc/dbus-specification.html#standard-interfaces-objectmanager][org.freedesktop.DBus.ObjectManager]] interface on server object :2:
    Deprecate methods GetAllClients(), GetAllClientsWithProperties()
    and signals ClientAdded and ClientRemoved.
 ** TODO Save state periodically to recover better from hard shutdowns
 ** TODO CheckerCompleted method, deprecate CheckedOK
+** TODO Secret Service API?
+   http://standards.freedesktop.org/secret-service/
+** TODO Remove D-Bus interfaces with old domain name			  :2:
+** TODO Remove old string_to_delta format				  :2:
+** TODO http://0pointer.de/blog/projects/stateless.html
+*** tmpfiles snippet to create /var/lib/mandos with right user+perms
+*** File in /usr/lib/sysusers.d to create user+group "_mandos"
+** TODO Error handling on error parsing config files
+** TODO init.d script error handling
 
 * mandos.xml
 ** Add mandos contact info in manual pages
@@ -85,7 +93,7 @@
 * mandos-ctl
 *** Handle "no D-Bus server" and/or "no Mandos server found" better
 *** [#B] --dump option
-** TODO Support RFC 3339 time duration syntax
+** TODO Remove old string_to_delta format				  :2:
 
 * TODO mandos-dispatch
   Listens for specified D-Bus signals and spawns shell commands with
@@ -97,7 +105,8 @@
    Better view of client data in the listing
 *** Properties popup
 ** Print a nice "We are sorry" message, save stack trace to log.
-** Show timeout countdown for approval
+** Rename module "gobject" to "GObject".
+** TODO Optional verbose mode to see checkers starting and succeeding
 
 * mandos-keygen
 ** TODO "--secfile" option
@@ -105,10 +114,6 @@
 ** TODO [#B] "--test" option
    For testing decryption before rebooting.
 
-* Makefile
-** TODO [#C] Implement DEB_BUILD_OPTIONS
-   http://www.debian.org/doc/debian-policy/ch-source.html#s-debianrules-options
-
 * Package
 ** /usr/share/initramfs-tools/hooks/mandos
 *** TODO [#C] use same file name rules as run-parts(8)

=== modified file 'clients.conf'
--- clients.conf	2011-11-26 23:08:17 +0000
+++ clients.conf	2012-06-23 00:58:49 +0000
@@ -4,19 +4,19 @@
 
 # How long until a client is disabled and not be allowed to get the
 # data this server holds.
-;timeout = 5m
+;timeout = PT5M
 
 # How often to run the checker to confirm that a client is still up.
 # Note: a new checker will not be started if an old one is still
 # running.  The server will wait for a checker to complete until the
 # above "timeout" occurs, at which time the client will be disabled,
 # and any running checker killed.
-;interval = 2m
+;interval = PT2M
 
 # Extended timeout is an added timeout that is given once after a
 # password has been sent sucessfully to a client.  This allows for
 # additional delays caused by file system checks and quota checks.
-;extended_timeout = 15m
+;extended_timeout = PT15M
 
 # What command to run as "the checker".
 ;checker = fping -q -- %%(host)s
@@ -25,10 +25,10 @@
 ;approved_by_default = True
 
 # How long to wait for approval.
-;approval_delay = 0s
+;approval_delay = PT0S
 
 # How long one approval will last.
-;approval_duration = 1s
+;approval_duration = PT1S
 
 # Whether this client is enabled by default
 ;enabled = True
@@ -78,10 +78,10 @@
 ;host = 192.0.2.3
 ;
 ;# Parameters from the [DEFAULT] section can be overridden per client.
-;interval = 1m
+;interval = PT1M
 ;
 ;# This client requires manual approval before it receives its secret.
 ;approved_by_default = False
 ;# Require approval within 30 seconds.
-;approval_delay = 30s
+;approval_delay = PT30S
 ;####

=== modified file 'common.ent'
--- common.ent	2012-01-15 21:07:44 +0000
+++ common.ent	2014-05-11 20:18:48 +0000
@@ -1,3 +1,3 @@
 <?xml version="1.0" encoding="UTF-8"?>
 <!-- Edited by Makefile -->
-<!ENTITY version "1.5.3">
+<!ENTITY version "1.6.5">

=== modified file 'debian/changelog'
--- debian/changelog	2012-02-21 21:20:12 +0000
+++ debian/changelog	2014-06-06 02:29:14 +0000
@@ -1,3 +1,131 @@
+mandos (1.6.5-3) unstable; urgency=medium
+
+  * debian/control (mandos-client/Depends): Add "dpkg-dev (>=1.16.0)";
+    initramfs-tools-hook runs "dpkg-architecture -qDEB_HOST_MULTIARCH".
+    (Closes: #750221)
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Fri, 06 Jun 2014 04:27:15 +0200
+
+mandos (1.6.5-2) unstable; urgency=medium
+
+  * debian/rules (override_dh_auto_test-arch): New; does nothing.  Fixes
+    FTBFS for build-indep.
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Tue, 13 May 2014 08:08:31 +0200
+
+mandos (1.6.5-1) unstable; urgency=medium
+
+  * New upstream release.
+  * debian/copyright: Change year to "2014".
+  * debian/control (Build-Depends, Build-Depends-Indep): Moved build
+    dependencies of "mandos" package to "Build-Depends-Indep".
+  * debian/upstream/signing-key.asc: New; upstream source public key.
+  * debian/control (Standards-Version): Updated to "3.9.5".
+  * debian/control (mandos/Depends): Remove the dependency on
+    "avahi-daemon (>= 0.6.31-3) | systemd-sysv".  It is unnecessary
+    since we have a workaround in debian/mandos.postinst anyway.
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Sun, 11 May 2014 22:16:33 +0200
+
+mandos (1.6.4-1) unstable; urgency=medium
+
+  * New upstream release.
+  * debian/control (Build-Depends): Add Python dependencies to
+                                    successfully run self-tests.
+  * debian/copyright: GPLv3 now has its own license file - use it.
+  * debian/watch: Set PGP signature URL.
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Sun, 16 Feb 2014 14:09:25 +0100
+
+mandos (1.6.3-1) unstable; urgency=low
+
+  * New upstream release.
+  * debian/control (Build-Depends): Added "systemd".
+  * debian/mandos.dirs (lib/systemd/system): New.
+  * debian/mandos-client.README.Debian: Refer to architecture libdir.
+  * debian/control (mandos/Depends): Add "avahi-daemon (>= 0.6.31-3) |
+                                     systemd-sysv".
+  * debian/mandos.postinst: If avahi-daemon is version 0.6.31-2 or older,
+                            edit /etc/init.d script headers Required-Start
+                            and Required-Stop to have "avahi" instead of
+                            "avahi-daemon", before insserv(8) sees it.
+  * debian/mandos-client.lintian-overrides: Libdir changes.
+  * debian/rules (override_dh_fixperms): - '' -
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Tue, 21 Jan 2014 22:01:30 +0100
+
+mandos (1.6.2-1) unstable; urgency=low
+
+  * New upstream release.
+  * debian/compat: Changed to "9".
+  * debian/control (Build-Depends): Changed debhelper version to (>= 9).
+    (Standards-Version): Updated to "3.9.4".
+    (DM-Upload-Allowed): Removed.
+    (mandos/Depends): Add "initscripts (>= 2.88dsf-13.3)" to be able to
+                      use the "/run" directory (for mandos.pid).
+  * debian/copyright (Copyright): Update year.
+  * Fix "Mandos/gnutls fails to establish connection, "an algorithm that
+    is not enabled was negotiated"" fixed by upstream. (Closes: #702120)
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Thu, 24 Oct 2013 22:33:40 +0200
+
+mandos (1.6.1-1) unstable; urgency=low
+
+  * New upstream release.
+  * debian/control (mandos/Depends): No longer depends on
+                                     python-gnupginterface, but does
+                                     depend on gnupg (<< 2).
+    (Build-Depends): Depend on debhelper 8.9.7 for using "override-*-arch"
+                     and "override-*-indep" targets in debian/rules.
+  * debian/mandos-client.README: Update Linux documentation link.
+  * debian/rules: Completely rewritten to use debhelper v7.
+  * initramfs-tools-hook: Bug fix: Make sure the right version of GnuPG is
+    copied into the initramfs image.  Always assume that GPGME is used to
+    avoid searching for it since the path might not be /usr/lib.  Thanks
+    to Félix Sipma <felix+debian@gueux.org> for the initial bug report,
+    and also thanks to Dick Middleton <dick@lingbrae.com> for some more
+    debugging. (Closes:  #721903)
+  * Fix "bashism in /bin/sh script" fixed by upstream. (Closes: #690639)
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Sun, 13 Oct 2013 19:03:23 +0200
+
+mandos (1.6.0-1) unstable; urgency=low
+
+  * New upstream release.
+  * debian/copyright (Copyright): Join the two lines to a single line.
+  * debian/mandos-client.README.Debian: Update to refer to the new
+    location of the example network hooks, and the new feature of using
+    all network interfaces.
+  * debian/mandos-client.docs (network-hooks.d): Removed.
+  * debian/mandos-client.examples (network-hooks.d): New.
+  * debian/rules (binary-common): Added "dh_installexamples".
+    (binary-common/dh_fixperms): Exclude new location of
+    "network-hooks.d".
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Mon, 18 Jun 2012 00:15:23 +0200
+
+mandos (1.5.5-1) unstable; urgency=low
+
+  * New upstream release.
+  * debian/copyright (Format): Updated to
+    "http://www.debian.org/doc/packaging-manuals/copyright-format/1.0/".
+  * debian/control (Build-Depends): Removed "man, locales-all".
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Fri, 01 Jun 2012 20:30:41 +0200
+
+mandos (1.5.4-1) unstable; urgency=low
+
+  * New upstream release.
+
+ -- Teddy Hogeborn <teddy@recompile.se>  Sun, 20 May 2012 15:38:34 +0200
+
+mandos (1.5.3-1.2) unstable; urgency=low
+
+  * Non-maintainer upload.
+  * Set Architecture to linux-any.  (Closes: #647670)
+
+ -- Robert Millan <rmh@debian.org>  Sun, 22 Apr 2012 16:22:01 +0200
+
 mandos (1.5.3-1.1) unstable; urgency=low
 
   * Non-maintainer upload.

=== modified file 'debian/compat'
--- debian/compat	2008-09-17 00:34:09 +0000
+++ debian/compat	2013-10-20 15:25:09 +0000
@@ -1,1 +1,1 @@
-7
+9

=== modified file 'debian/control'
--- debian/control	2012-02-21 21:20:12 +0000
+++ debian/control	2014-07-12 12:54:13 +0000
@@ -4,22 +4,24 @@
 Maintainer: Mandos Maintainers <mandos@recompile.se>
 Uploaders: Teddy Hogeborn <teddy@recompile.se>,
 	   Björn Påhlsson <belorn@recompile.se>
-Build-Depends: debhelper (>= 7), docbook-xml, docbook-xsl,
+Build-Depends: debhelper (>= 9), docbook-xml, docbook-xsl,
 	libavahi-core-dev, libgpgme11-dev, libgnutls-dev, xsltproc,
-	pkg-config, man, locales-all
-Standards-Version: 3.9.2
+	pkg-config
+Build-Depends-Indep: systemd, python (>=2.6), python-gnutls,
+	python-dbus, python-avahi, python-gobject,
+	python (>=2.7) | python-argparse
+Standards-Version: 3.9.5
 Vcs-Bzr: http://ftp.recompile.se/pub/mandos/trunk
 Vcs-Browser: http://bzr.recompile.se/loggerhead/mandos/trunk/files
 Homepage: http://www.recompile.se/mandos
-DM-Upload-Allowed: yes
 
 Package: mandos
 Architecture: all
 Depends: ${misc:Depends}, python (>=2.6), python-gnutls, python-dbus,
 	python-avahi, python-gobject, avahi-daemon, adduser,
-	python-urwid, python (>=2.7) | python-argparse,
-	python-gnupginterface
-Recommends: fping
+	python-urwid, python (>=2.7) | python-argparse, gnupg (<< 2),
+	initscripts (>= 2.88dsf-13.3)
+Recommends: ssh-client | fping
 Description: server giving encrypted passwords to Mandos clients
  This is the server part of the Mandos system, which allows
  computers to have encrypted root file systems and at the
@@ -36,9 +38,10 @@
  whereupon the computers can continue booting normally.
 
 Package: mandos-client
-Architecture: any
+Architecture: linux-any
 Depends: ${shlibs:Depends}, ${misc:Depends}, adduser, cryptsetup,
-	gnupg (<< 2), initramfs-tools
+	gnupg (<< 2), initramfs-tools, dpkg-dev (>=1.16.0)
+Recommends: ssh
 Breaks: dropbear (<= 0.53.1-1)
 Enhances: cryptsetup
 Description: do unattended reboots with an encrypted root file system

=== modified file 'debian/copyright'
--- debian/copyright	2011-12-31 23:05:34 +0000
+++ debian/copyright	2014-03-01 09:39:25 +0000
@@ -1,11 +1,11 @@
-Format: http://anonscm.debian.org/viewvc/dep/web/deps/dep5.mdwn?revision=202
+Format: http://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
 Upstream-Name: Mandos
 Upstream-Contact: Mandos <mandos@recompile.se>
 Source: <http://www.recompile.se/mandos>
 
 Files: *
-Copyright: Copyright © 2008-2012 Teddy Hogeborn
-Copyright: Copyright © 2008-2012 Björn Påhlsson
+Copyright: Copyright © 2008-2014 Teddy Hogeborn
+	   Copyright © 2008-2014 Björn Påhlsson
 License: GPL-3+
     This program is free software: you can redistribute it and/or
     modify it under the terms of the GNU General Public License as
@@ -22,4 +22,4 @@
     <http://www.gnu.org/licenses/>.
     .
     On Debian systems, the complete text of the GNU General Public
-    License can be found in "/usr/share/common-licenses/GPL".
+    License can be found in "/usr/share/common-licenses/GPL-3".

=== modified file 'debian/mandos-client.README.Debian'
--- debian/mandos-client.README.Debian	2012-01-01 04:52:07 +0000
+++ debian/mandos-client.README.Debian	2013-10-28 10:04:05 +0000
@@ -16,7 +16,8 @@
   is possible to verify that the correct password will be received by
   this client by running the command, on the client:
   
-	/usr/lib/mandos/plugins.d/mandos-client \
+	/usr/lib/$(dpkg-architecture -qDEB_HOST_MULTIARCH \
+	)/mandos/plugins.d/mandos-client \
 		--pubkey=/etc/keys/mandos/pubkey.txt \
 		--seckey=/etc/keys/mandos/seckey.txt; echo
   
@@ -32,9 +33,9 @@
 
 * Specifying a Client Network Interface
   
-  At boot time the network interface to use will by default be
-  automatically detected.  If this should result in an incorrect
-  interface, edit the DEVICE setting in the
+  At boot time the network interfaces to use will by default be
+  automatically detected.  If this should result in incorrect
+  interfaces, edit the DEVICE setting in the
   "/etc/initramfs-tools/initramfs.conf" file.  (The default setting is
   empty, meaning it will autodetect the interface.)  *If* the DEVICE
   setting is changed, it will be necessary to update the initrd image
@@ -42,21 +43,21 @@
   
 	update-initramfs -k all -u
   
-  The device can be overridden at boot time on the Linux kernel
+  The device can also be overridden at boot time on the Linux kernel
   command line using the sixth colon-separated field of the "ip="
   option; for exact syntax, read the documentation in the file
-  "/usr/share/doc/linux-doc-*/Documentation/filesystems/nfsroot.txt",
+  "/usr/share/doc/linux-doc-*/Documentation/filesystems/nfs/nfsroot.txt",
   available in the "linux-doc-*" package.
   
-  Note that since this network interface is used in the initial RAM
-  disk environment, the network interface *must* exist at that stage.
-  Thus, the interface can *not* be a pseudo-interface such as "br0" or
-  "tun0"; instead, only real interface (such as "eth0") can be used.
+  Note that since the network interfaces are used in the initial RAM
+  disk environment, the network interfaces *must* exist at that stage.
+  Thus, an interface can *not* be a pseudo-interface such as "br0" or
+  "tun0"; instead, only real interfaces (such as "eth0") can be used.
   This can be overcome by writing a "network hook" program to create
-  the interface (see mandos-client(8mandos)) and placing it in
+  an interface (see mandos-client(8mandos)) and placing it in
   "/etc/mandos/network-hooks.d", from where it will be copied into the
   initial RAM disk.  Example network hook scripts can be found in
-  "/usr/share/doc/mandos-client/network-hooks.d".
+  "/usr/share/doc/mandos-client/examples/network-hooks.d".
 
 * User-Supplied Plugins
   
@@ -90,4 +91,4 @@
   work, "--options-for=mandos-client:--connect=<ADDRESS>:<PORT>" needs
   to be manually added to the file "/etc/mandos/plugin-runner.conf".
 
- -- Teddy Hogeborn <teddy@recompile.se>, Mon, 28 Nov 2011 23:07:22 +0100
+ -- Teddy Hogeborn <teddy@recompile.se>, Mon, 28 Oct 2013 11:02:26 +0100

=== modified file 'debian/mandos-client.docs'
--- debian/mandos-client.docs	2011-11-27 02:32:20 +0000
+++ debian/mandos-client.docs	2012-06-01 21:48:12 +0000
@@ -1,4 +1,3 @@
 NEWS
 README
 TODO
-network-hooks.d

=== added file 'debian/mandos-client.examples'
--- debian/mandos-client.examples	1970-01-01 00:00:00 +0000
+++ debian/mandos-client.examples	2012-06-01 21:48:12 +0000
@@ -0,0 +1,1 @@
+network-hooks.d

=== modified file 'debian/mandos-client.lintian-overrides'
--- debian/mandos-client.lintian-overrides	2010-10-13 06:12:52 +0000
+++ debian/mandos-client.lintian-overrides	2014-01-20 21:50:11 +0000
@@ -2,12 +2,12 @@
 #
 mandos-client binary: non-standard-dir-perm etc/keys/mandos/ 0700 != 0755
 
-# The directory /usr/lib/mandos/plugins.d contains setuid binaries
-# which are not meant to be run outside an initial RAM disk
+# The directory /usr/lib/<arch>/mandos/plugins.d contains setuid
+# binaries which are not meant to be run outside an initial RAM disk
 # environment (except for test purposes).  It would be insecure to
 # allow anyone to run them.
 #
-mandos-client binary: non-standard-dir-perm usr/lib/mandos/plugins.d/ 0700 != 0755
+mandos-client binary: non-standard-dir-perm usr/lib/*/mandos/plugins.d/ 0700 != 0755
 
 # These binaries must be setuid root, since they need root powers, but
 # are started by plugin-runner(8mandos), which runs all plugins as
@@ -15,14 +15,14 @@
 # system, but in an initial RAM disk environment.  Here they are
 # protected from non-root access by the directory permissions, above.
 #
-mandos-client binary: setuid-binary usr/lib/mandos/plugins.d/mandos-client 4755 root/root
-mandos-client binary: setuid-binary usr/lib/mandos/plugins.d/askpass-fifo 4755 root/root
-mandos-client binary: setuid-binary usr/lib/mandos/plugins.d/splashy 4755 root/root
-mandos-client binary: setuid-binary usr/lib/mandos/plugins.d/usplash 4755 root/root
-mandos-client binary: setuid-binary usr/lib/mandos/plugins.d/plymouth 4755 root/root
+mandos-client binary: setuid-binary usr/lib/*/mandos/plugins.d/mandos-client 4755 root/root
+mandos-client binary: setuid-binary usr/lib/*/mandos/plugins.d/askpass-fifo 4755 root/root
+mandos-client binary: setuid-binary usr/lib/*/mandos/plugins.d/splashy 4755 root/root
+mandos-client binary: setuid-binary usr/lib/*/mandos/plugins.d/usplash 4755 root/root
+mandos-client binary: setuid-binary usr/lib/*/mandos/plugins.d/plymouth 4755 root/root
 
 # The directory /etc/mandos/plugins.d can be used by local system
 # administrators to place plugins in, overriding and complementing
-# /usr/lib/mandos/plugins.d, and must be likewise protected.
+# /usr/lib/<arch>/mandos/plugins.d, and must be likewise protected.
 #
 mandos-client binary: non-standard-dir-perm etc/mandos/plugins.d/ 0700 != 0755

=== modified file 'debian/mandos.dirs'
--- debian/mandos.dirs	2011-11-26 22:22:20 +0000
+++ debian/mandos.dirs	2013-10-27 17:42:23 +0000
@@ -5,3 +5,4 @@
 etc/dbus-1/system.d
 usr/sbin
 var/lib/mandos
+lib/systemd/system

=== modified file 'debian/mandos.postinst'
--- debian/mandos.postinst	2011-11-26 22:22:20 +0000
+++ debian/mandos.postinst	2014-06-07 20:29:36 +0000
@@ -47,6 +47,16 @@
 	;;
 esac
 
+# Avahi version 0.6.31-2 and older provides "avahi" (instead of
+# "avahi-daemon") in its /etc/init.d script header.  To make
+# insserv(8) happy, we edit our /etc/init.d script header to contain
+# the correct string before the code added by dh_installinit calls
+# update.rc-d, which calls insserv.
+avahi_version="`dpkg-query --showformat='${Version}' --show avahi-daemon`"
+if dpkg --compare-versions "$avahi_version" le 0.6.31-2; then
+    sed --in-place --expression='/^### BEGIN INIT INFO$/,/^### END INIT INFO$/s/^\(# Required-\(Stop\|Start\):.*avahi\)-daemon\>/\1/g' /etc/init.d/mandos
+fi
+
 #DEBHELPER#
 
 exit 0

=== modified file 'debian/rules'
--- debian/rules	2012-02-21 21:20:12 +0000
+++ debian/rules	2014-05-12 21:02:38 +0000
@@ -1,107 +1,31 @@
 #!/usr/bin/make -f
-# Sample debian/rules that uses debhelper.
-#
-# This file was originally written by Joey Hess and Craig Small.
-# As a special exception, when this file is copied by dh-make into a
-# dh-make output file, you may use that output file without restriction.
-# This special exception was added by Craig Small in version 0.37 of dh-make.
-#
-# Modified to make a template file for a multi-binary package with separated
-# build-arch and build-indep targets  by Bill Allombert 2001
-
-# Uncomment this to turn on verbose mode.
-#export DH_VERBOSE=1
-
-# This has to be exported to make some magic below work.
-export DH_OPTIONS
-
-# -pie was broken briefly on the mips and mipsel architectures, see
-# <http://bugs.debian.org/526961>
-BINUTILS_V := $(shell dpkg-query --showformat='$${Version}' \
-			--show binutils)
-ifeq (yes,$(shell dpkg --compare-versions $(BINUTILS_V) lt 2.20-3 \
-		&& dpkg --compare-versions $(BINUTILS_V) ge 2.19.1-1 \
-		&& echo yes))
-  ifneq (,$(strip $(findstring :$(DEB_HOST_ARCH):,:mips:mipsel:) \
-		$(findstring :$(DEB_BUILD_ARCH):,:mips:mipsel:)))
-    BROKEN_PIE := yes
-    export BROKEN_PIE
-  endif
-endif
-
-configure: configure-stamp
-configure-stamp:
-	dh_testdir
-	touch configure-stamp
-
-build: build-arch build-indep
-
-build-arch: build-arch-stamp
-build-arch-stamp: configure-stamp
+%:
+	dh $@
+
+override_dh_auto_build-arch:
 	LC_ALL=en_US.utf8 dh_auto_build -- all doc
-	touch $@
-
-build-indep: build-indep-stamp
-build-indep-stamp: configure-stamp
-	LC_ALL=en_US.UTF-8 dh_auto_build -- doc
-	touch $@
-
-clean:
-	dh_testdir
-	dh_testroot
-	rm -f build-arch-stamp build-indep-stamp configure-stamp
-	dh_auto_clean
-	dh_clean
-
-install: install-indep install-arch
-install-indep:
-	dh_testdir
-	dh_testroot
-	dh_prep
-	dh_installdirs --indep
-	$(MAKE) DESTDIR=$(CURDIR)/debian/mandos install-server
-	dh_lintian
+
+override_dh_auto_build-indep:
+	LC_ALL=en_US.utf8 dh_auto_build -- doc
+
+override_dh_installinit-indep:
 	dh_installinit --onlyscripts \
 		--update-rcd-params="defaults 25 15"
-	dh_install --indep
-
-install-arch:
-	dh_testdir
-	dh_testroot
-	dh_prep
-	dh_installdirs --same-arch
-	$(MAKE) DESTDIR=$(CURDIR)/debian/mandos-client install-client-nokey
-	dh_lintian
-	dh_install --same-arch
-
-binary-common:
-	dh_testdir
-	dh_testroot
-	dh_installchangelogs
-	dh_installdocs
-	dh_link
-	dh_strip
-	dh_compress
+
+override_dh_auto_install-indep:
+	$(MAKE) DESTDIR=$(CURDIR)/debian/mandos install-server
+
+override_dh_auto_install-arch:
+	$(MAKE) DESTDIR=$(CURDIR)/debian/mandos-client \
+		install-client-nokey
+
+override_dh_fixperms:
 	dh_fixperms --exclude etc/keys/mandos \
 		--exclude etc/mandos/clients.conf \
 		--exclude etc/mandos/plugins.d \
-		--exclude usr/lib/mandos/plugins.d \
-		--exclude usr/share/doc/mandos-client/network-hooks.d
-	dh_installdeb
-	dh_shlibdeps
-	dh_gencontrol
-	dh_md5sums
-	dh_builddeb
-
-# Build architecture independant packages using the common target.
-binary-indep: build-indep install-indep
-	$(MAKE) -f debian/rules DH_OPTIONS=--indep binary-common
-
-# Build architecture dependant packages using the common target.
-binary-arch: build-arch install-arch
-	$(MAKE) -f debian/rules DH_OPTIONS=--same-arch binary-common
-
-binary: binary-arch binary-indep
-
-.PHONY: build clean binary-indep binary-arch binary install \
-		install-indep install-arch configure
+		--exclude usr/lib/$(shell dpkg-architecture -qDEB_HOST_MULTIARCH 2>/dev/null)/mandos/plugins.d \
+		--exclude usr/share/doc/mandos-client/examples/network-hooks.d
+	chmod --recursive g-w -- \
+	"$(CURDIR)/debian/mandos-client/usr/share/doc/mandos-client/examples/network-hooks.d"
+
+override_dh_auto_test-arch: ;

=== added directory 'debian/upstream'
=== added file 'debian/upstream/signing-key.asc'
--- debian/upstream/signing-key.asc	1970-01-01 00:00:00 +0000
+++ debian/upstream/signing-key.asc	2014-03-28 22:32:21 +0000
@@ -0,0 +1,52 @@
+-----BEGIN PGP PUBLIC KEY BLOCK-----
+Version: GnuPG v1.4.12 (GNU/Linux)
+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+=9Noe
+-----END PGP PUBLIC KEY BLOCK-----

=== modified file 'debian/watch'
--- debian/watch	2011-10-05 16:00:56 +0000
+++ debian/watch	2014-02-16 02:42:42 +0000
@@ -1,2 +1,3 @@
 version=3
-ftp://ftp.recompile.se/pub/mandos/mandos[-_]([^\s]+?)(?:\.orig)?\.tar\.(?:gz|bz2|7z|xz)
+opts=pgpsigurlmangle=s/$/.asc/ \
+ ftp://ftp.recompile.se/pub/mandos/mandos[-_]([^\s]+?)(?:\.orig)?\.tar\.(?:gz|bz2|7z|xz)

=== modified file 'init.d-mandos'
--- init.d-mandos	2011-10-05 16:00:56 +0000
+++ init.d-mandos	2014-01-06 15:56:54 +0000
@@ -1,12 +1,12 @@
 #! /bin/sh
 ### BEGIN INIT INFO
 # Provides:          mandos
-# Required-Start:    $remote_fs $syslog avahi
-# Required-Stop:     $remote_fs $syslog avahi
+# Required-Start:    $remote_fs $syslog avahi-daemon
+# Required-Stop:     $remote_fs $syslog avahi-daemon
 # Default-Start:     2 3 4 5
 # Default-Stop:      0 1 6
 # Short-Description: Mandos server
-# Description:       Gives encrypted passwords to Mandos clients
+# Description:       Server of encrypted passwords to Mandos clients
 ### END INIT INFO
 
 # Author: Teddy Hogeborn <teddy@recompile.se>
@@ -23,7 +23,11 @@
 NAME=mandos
 DAEMON=/usr/sbin/$NAME
 DAEMON_ARGS=""
-PIDFILE=/var/run/$NAME.pid
+if [ -d /run/. ]; then
+    PIDFILE=/run/$NAME.pid
+else
+    PIDFILE=/var/run/$NAME.pid
+fi
 SCRIPTNAME=/etc/init.d/$NAME
 
 # Exit if the package is not installed
@@ -40,7 +44,8 @@
 . /lib/init/vars.sh
 
 # Define LSB log_* functions.
-# Depend on lsb-base (>= 3.0-6) to ensure that this file is present.
+# Depend on lsb-base (>= 3.2-14) to ensure that this file is present
+# and status_of_proc is working.
 . /lib/lsb/init-functions
 
 #
@@ -118,6 +123,9 @@
 		2) [ "$VERBOSE" != no ] && log_end_msg 1 ;;
 	esac
 	;;
+  status)
+	status_of_proc "$DAEMON" "$NAME" -p "$PIDFILE" && exit 0 || exit $?
+	;;
   #reload|force-reload)
 	#
 	# If do_reload() is not implemented then leave this commented out
@@ -144,14 +152,14 @@
 		esac
 		;;
 	  *)
-	  	# Failed to stop
+		# Failed to stop
 		log_end_msg 1
 		;;
 	esac
 	;;
   *)
 	#echo "Usage: $SCRIPTNAME {start|stop|restart|reload|force-reload}" >&2
-	echo "Usage: $SCRIPTNAME {start|stop|restart|force-reload}" >&2
+	echo "Usage: $SCRIPTNAME {start|stop|status|restart|force-reload}" >&2
 	exit 3
 	;;
 esac

=== modified file 'initramfs-tools-hook'
--- initramfs-tools-hook	2011-11-29 18:19:31 +0000
+++ initramfs-tools-hook	2013-10-28 08:38:47 +0000
@@ -3,7 +3,7 @@
 # This script will be run by 'mkinitramfs' when it creates the image.
 # Its job is to decide which files to install, then install them into
 # the staging area, where the initramfs is being created.  This
-# happens when a new 'linux-image' package is installed, or when the
+# happens when a new 'linux-image' package is installed, or when an
 # administrator runs 'update-initramfs' by hand to update an initramfs
 # image.
 
@@ -29,13 +29,15 @@
 
 . /usr/share/initramfs-tools/hook-functions
 
-for d in /usr /usr/local; do
-    if [ -d "$d"/lib/mandos ]; then
-	prefix="$d"
+for d in /usr/lib \
+    "/usr/lib/`dpkg-architecture -qDEB_HOST_MULTIARCH 2>/dev/null`" \
+    "`rpm --eval='%{_libdir}' 2>/dev/null`" /usr/local/lib; do
+    if [ -d "$d"/mandos ]; then
+	libdir="$d"
 	break
     fi
 done
-if [ -z "$prefix" ]; then
+if [ -z "$libdir" ]; then
     # Mandos not found
     exit 1
 fi
@@ -77,12 +79,12 @@
     --mode=u=rwx "${DESTDIR}${PLUGINDIR}"
 
 # Copy the Mandos plugin runner
-copy_exec "$prefix"/lib/mandos/plugin-runner "${MANDOSDIR}"
+copy_exec "$libdir"/mandos/plugin-runner "${MANDOSDIR}"
 
 # Copy the plugins
 
 # Copy the packaged plugins
-for file in "$prefix"/lib/mandos/plugins.d/*; do
+for file in "$libdir"/mandos/plugins.d/*; do
     base="`basename \"$file\"`"
     # Is this plugin overridden?
     if [ -e "/etc/mandos/plugins.d/$base" ]; then
@@ -147,12 +149,18 @@
     fi
 done
 
-# GPGME needs /usr/bin/gpg
-if [ ! -e "${DESTDIR}/usr/bin/gpg" \
-    -a -n "`ls \"${DESTDIR}\"/usr/lib/libgpgme.so* \
-		2>/dev/null`" ]; then
-    copy_exec /usr/bin/gpg
-fi
+# GPGME needs GnuPG
+libgpgme11_version="`dpkg-query --showformat='${Version}' --show libgpgme11`"
+if dpkg --compare-versions "$libgpgme11_version" ge 1.4.1-0.1; then
+    gpg=/usr/bin/gpg2
+else
+    gpg=/usr/bin/gpg
+fi
+if [ ! -e "${DESTDIR}$gpg" ]; then
+    copy_exec "$gpg"
+fi
+unset gpg
+unset libgpgme11_version
 
 # Config files
 for file in /etc/mandos/plugin-runner.conf; do

=== added file 'initramfs-unpack'
--- initramfs-unpack	1970-01-01 00:00:00 +0000
+++ initramfs-unpack	2013-10-13 15:43:42 +0000
@@ -0,0 +1,67 @@
+#!/bin/bash
+# 
+# Initramfs unpacker - unpacks initramfs images into /tmp
+# 
+# Copyright © 2013 Teddy Hogeborn
+# Copyright © 2013 Björn Påhlsson
+# 
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+#     This program is distributed in the hope that it will be useful,
+#     but WITHOUT ANY WARRANTY; without even the implied warranty of
+#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#     GNU General Public License for more details.
+# 
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see
+# <http://www.gnu.org/licenses/>.
+# 
+# Contact the authors at <mandos@recompile.se>.
+
+cpio="cpio --extract --make-directories --unconditional --preserve-modification-time"
+
+if [ -z "$*" ]; then
+    set -- /boot/initrd.img-*
+fi
+
+for imgfile in "$@"; do
+    if ! [ -f "$imgfile" ]; then
+	echo "Error: Not an existing file: $imgfile" >&2
+	continue
+    fi
+    imgdir="${TMPDIR:-/tmp}/${imgfile##*/}"
+    if [ -d "$imgdir" ]; then
+	rm --recursive -- "$imgdir"
+    fi
+    mkdir --parents "$imgdir"
+    # Does this image contain microcode?
+    if $cpio --quiet --list --file="$imgfile" >/dev/null 2>&1; then
+	# Number of bytes to skip to get to the compressed archive
+	skip=$(($(LANG=C $cpio --io-size=1 --list --file="$imgfile" 2>&1 \
+	    | sed --quiet --expression='s/^\([0-9]\+\) blocks$/\1/p')+8))
+	catimg="dd if=$imgfile bs=$skip skip=1 status=noxfer"
+    else
+	catimg="cat -- $imgfile"
+    fi
+    # Determine the compression method
+    if { $catimg 2>/dev/null | zcat --test >/dev/null 2>&1;
+	    [ ${PIPESTATUS[-1]} -eq 0 ]; }; then
+	decomp="zcat"
+    elif { $catimg 2>/dev/null | bzip2 --test >/dev/null 2>&1;
+	    [ ${PIPESTATUS[-1]} -eq 0 ]; }; then
+	decomp="bzip2 --stdout --decompress"
+    elif { $catimg 2>/dev/null | lzop --test >/dev/null 2>&1;
+	    [ ${PIPESTATUS[-1]} -eq 0 ]; }; then
+	decomp="lzop --stdout --decompress"
+    else
+	echo "Error: Could not determine type of $imgfile" >&2
+	continue
+    fi
+    $catimg 2>/dev/null | $decomp | ( cd -- "$imgdir" && $cpio --quiet )
+    if [ ${PIPESTATUS[-1]} -eq 0 ]; then
+	echo "$imgfile unpacked into $imgdir"
+    fi
+done

=== modified file 'intro.xml'
--- intro.xml	2011-12-31 23:05:34 +0000
+++ intro.xml	2014-06-22 02:19:30 +0000
@@ -1,7 +1,7 @@
 <?xml version="1.0" encoding="UTF-8"?>
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 "http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
-<!ENTITY TIMESTAMP "2012-01-01">
+<!ENTITY TIMESTAMP "2014-06-22">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -215,17 +215,22 @@
       </para>
     </refsect2>
     
-    <refsect2 id="fakeping">
-      <title>Faking ping replies?</title>
+    <refsect2 id="fakecheck">
+      <title>Faking checker results?</title>
       <para>
-	The default for the server is to use
+	If the Mandos client does not have an SSH server, the default
+	is for the Mandos server to use
 	<quote><literal>fping</literal></quote>, the replies to which
 	could be faked to eliminate the timeout.  But this could
 	easily be changed to any shell command, with any security
-	measures you like.  It could, for instance, be changed to an
-	SSH command with strict keychecking, which could not be faked.
-	Or IPsec could be used for the ping packets, making them
-	secure.
+	measures you like.  If the Mandos client
+	<emphasis>has</emphasis> an SSH server, the default
+	configuration (as generated by
+	<command>mandos-keygen</command> with the
+	<option>--password</option> option) is for the Mandos server
+	to use an <command>ssh-keyscan</command> command with strict
+	keychecking, which can not be faked.  Alternatively, IPsec
+	could be used for the ping packets, making them secure.
       </para>
     </refsect2>
   </refsect1>

=== modified file 'mandos'
--- mandos	2012-02-26 11:47:19 +0000
+++ mandos	2014-06-15 02:48:49 +0000
@@ -11,8 +11,8 @@
 # "AvahiService" class, and some lines in "main".
 # 
 # Everything else is
-# Copyright © 2008-2012 Teddy Hogeborn
-# Copyright © 2008-2012 Björn Påhlsson
+# Copyright © 2008-2014 Teddy Hogeborn
+# Copyright © 2008-2014 Björn Påhlsson
 # 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -34,6 +34,8 @@
 from __future__ import (division, absolute_import, print_function,
                         unicode_literals)
 
+from future_builtins import *
+
 import SocketServer as socketserver
 import socket
 import argparse
@@ -66,6 +68,7 @@
 import binascii
 import tempfile
 import itertools
+import collections
 
 import dbus
 import dbus.service
@@ -76,7 +79,6 @@
 import ctypes.util
 import xml.dom.minidom
 import inspect
-import GnuPGInterface
 
 try:
     SO_BINDTODEVICE = socket.SO_BINDTODEVICE
@@ -86,13 +88,11 @@
     except ImportError:
         SO_BINDTODEVICE = None
 
-version = "1.5.3"
+version = "1.6.5"
 stored_state_file = "clients.pickle"
 
 logger = logging.getLogger()
-syslogger = (logging.handlers.SysLogHandler
-             (facility = logging.handlers.SysLogHandler.LOG_DAEMON,
-              address = str("/dev/log")))
+syslogger = None
 
 try:
     if_nametoindex = (ctypes.cdll.LoadLibrary
@@ -114,6 +114,11 @@
 def initlogger(debug, level=logging.WARNING):
     """init logger and add loglevel"""
     
+    global syslogger
+    syslogger = (logging.handlers.SysLogHandler
+                 (facility =
+                  logging.handlers.SysLogHandler.LOG_DAEMON,
+                  address = str("/dev/log")))
     syslogger.setFormatter(logging.Formatter
                            ('Mandos [%(process)d]: %(levelname)s:'
                             ' %(message)s'))
@@ -137,19 +142,17 @@
 class PGPEngine(object):
     """A simple class for OpenPGP symmetric encryption & decryption"""
     def __init__(self):
-        self.gnupg = GnuPGInterface.GnuPG()
         self.tempdir = tempfile.mkdtemp(prefix="mandos-")
-        self.gnupg = GnuPGInterface.GnuPG()
-        self.gnupg.options.meta_interactive = False
-        self.gnupg.options.homedir = self.tempdir
-        self.gnupg.options.extra_args.extend(['--force-mdc',
-                                              '--quiet',
-                                              '--no-use-agent'])
+        self.gnupgargs = ['--batch',
+                          '--home', self.tempdir,
+                          '--force-mdc',
+                          '--quiet',
+                          '--no-use-agent']
     
     def __enter__(self):
         return self
     
-    def __exit__ (self, exc_type, exc_value, traceback):
+    def __exit__(self, exc_type, exc_value, traceback):
         self._cleanup()
         return False
     
@@ -172,44 +175,52 @@
     def password_encode(self, password):
         # Passphrase can not be empty and can not contain newlines or
         # NUL bytes.  So we prefix it and hex encode it.
-        return b"mandos" + binascii.hexlify(password)
+        encoded = b"mandos" + binascii.hexlify(password)
+        if len(encoded) > 2048:
+            # GnuPG can't handle long passwords, so encode differently
+            encoded = (b"mandos" + password.replace(b"\\", b"\\\\")
+                       .replace(b"\n", b"\\n")
+                       .replace(b"\0", b"\\x00"))
+        return encoded
     
     def encrypt(self, data, password):
-        self.gnupg.passphrase = self.password_encode(password)
-        with open(os.devnull, "w") as devnull:
-            try:
-                proc = self.gnupg.run(['--symmetric'],
-                                      create_fhs=['stdin', 'stdout'],
-                                      attach_fhs={'stderr': devnull})
-                with contextlib.closing(proc.handles['stdin']) as f:
-                    f.write(data)
-                with contextlib.closing(proc.handles['stdout']) as f:
-                    ciphertext = f.read()
-                proc.wait()
-            except IOError as e:
-                raise PGPError(e)
-        self.gnupg.passphrase = None
+        passphrase = self.password_encode(password)
+        with tempfile.NamedTemporaryFile(dir=self.tempdir
+                                         ) as passfile:
+            passfile.write(passphrase)
+            passfile.flush()
+            proc = subprocess.Popen(['gpg', '--symmetric',
+                                     '--passphrase-file',
+                                     passfile.name]
+                                    + self.gnupgargs,
+                                    stdin = subprocess.PIPE,
+                                    stdout = subprocess.PIPE,
+                                    stderr = subprocess.PIPE)
+            ciphertext, err = proc.communicate(input = data)
+        if proc.returncode != 0:
+            raise PGPError(err)
         return ciphertext
     
     def decrypt(self, data, password):
-        self.gnupg.passphrase = self.password_encode(password)
-        with open(os.devnull, "w") as devnull:
-            try:
-                proc = self.gnupg.run(['--decrypt'],
-                                      create_fhs=['stdin', 'stdout'],
-                                      attach_fhs={'stderr': devnull})
-                with contextlib.closing(proc.handles['stdin']) as f:
-                    f.write(data)
-                with contextlib.closing(proc.handles['stdout']) as f:
-                    decrypted_plaintext = f.read()
-                proc.wait()
-            except IOError as e:
-                raise PGPError(e)
-        self.gnupg.passphrase = None
+        passphrase = self.password_encode(password)
+        with tempfile.NamedTemporaryFile(dir = self.tempdir
+                                         ) as passfile:
+            passfile.write(passphrase)
+            passfile.flush()
+            proc = subprocess.Popen(['gpg', '--decrypt',
+                                     '--passphrase-file',
+                                     passfile.name]
+                                    + self.gnupgargs,
+                                    stdin = subprocess.PIPE,
+                                    stdout = subprocess.PIPE,
+                                    stderr = subprocess.PIPE)
+            decrypted_plaintext, err = proc.communicate(input
+                                                        = data)
+        if proc.returncode != 0:
+            raise PGPError(err)
         return decrypted_plaintext
 
 
-
 class AvahiError(Exception):
     def __init__(self, value, *args, **kwargs):
         self.value = value
@@ -232,7 +243,7 @@
                Used to optionally bind to the specified interface.
     name: string; Example: 'Mandos'
     type: string; Example: '_mandos._tcp'.
-                  See <http://www.dns-sd.org/ServiceTypes.html>
+     See <https://www.iana.org/assignments/service-names-port-numbers>
     port: integer; what port to announce
     TXT: list of strings; TXT record for the service
     domain: string; Domain to publish on, default to .local if empty.
@@ -244,6 +255,7 @@
     server: D-Bus Server
     bus: dbus.SystemBus()
     """
+    
     def __init__(self, interface = avahi.IF_UNSPEC, name = None,
                  servicetype = None, port = None, TXT = None,
                  domain = "", host = "", max_renames = 32768,
@@ -262,6 +274,7 @@
         self.server = None
         self.bus = bus
         self.entry_group_state_changed_match = None
+    
     def rename(self):
         """Derived from the Avahi example code"""
         if self.rename_count >= self.max_renames:
@@ -277,10 +290,11 @@
         try:
             self.add()
         except dbus.exceptions.DBusException as error:
-            logger.critical("DBusException: %s", error)
+            logger.critical("D-Bus Exception", exc_info=error)
             self.cleanup()
             os._exit(1)
         self.rename_count += 1
+    
     def remove(self):
         """Derived from the Avahi example code"""
         if self.entry_group_state_changed_match is not None:
@@ -288,6 +302,7 @@
             self.entry_group_state_changed_match = None
         if self.group is not None:
             self.group.Reset()
+    
     def add(self):
         """Derived from the Avahi example code"""
         self.remove()
@@ -310,6 +325,7 @@
             dbus.UInt16(self.port),
             avahi.string_array_to_txt_array(self.TXT))
         self.group.Commit()
+    
     def entry_group_state_changed(self, state, error):
         """Derived from the Avahi example code"""
         logger.debug("Avahi entry group state change: %i", state)
@@ -322,8 +338,9 @@
         elif state == avahi.ENTRY_GROUP_FAILURE:
             logger.critical("Avahi: Error in group state changed %s",
                             unicode(error))
-            raise AvahiGroupError("State changed: %s"
-                                  % unicode(error))
+            raise AvahiGroupError("State changed: {0!s}"
+                                  .format(error))
+    
     def cleanup(self):
         """Derived from the Avahi example code"""
         if self.group is not None:
@@ -334,6 +351,7 @@
                 pass
             self.group = None
         self.remove()
+    
     def server_state_changed(self, state, error=None):
         """Derived from the Avahi example code"""
         logger.debug("Avahi server state change: %i", state)
@@ -358,6 +376,7 @@
                 logger.debug("Unknown state: %r", state)
             else:
                 logger.debug("Unknown state: %r: %r", state, error)
+    
     def activate(self):
         """Derived from the Avahi example code"""
         if self.server is None:
@@ -370,22 +389,25 @@
                                  self.server_state_changed)
         self.server_state_changed(self.server.GetState())
 
+
 class AvahiServiceToSyslog(AvahiService):
     def rename(self):
         """Add the new name to the syslog messages"""
         ret = AvahiService.rename(self)
         syslogger.setFormatter(logging.Formatter
-                               ('Mandos (%s) [%%(process)d]:'
-                                ' %%(levelname)s: %%(message)s'
-                                % self.name))
+                               ('Mandos ({0}) [%(process)d]:'
+                                ' %(levelname)s: %(message)s'
+                                .format(self.name)))
         return ret
 
+
 def timedelta_to_milliseconds(td):
     "Convert a datetime.timedelta() to milliseconds"
     return ((td.days * 24 * 60 * 60 * 1000)
             + (td.seconds * 1000)
             + (td.microseconds // 1000))
-        
+
+
 class Client(object):
     """A representation of a client host served by this server.
     
@@ -427,19 +449,21 @@
     runtime_expansions: Allowed attributes for runtime expansion.
     expires:    datetime.datetime(); time (UTC) when a client will be
                 disabled, or None
+    server_settings: The server_settings dict from main()
     """
     
     runtime_expansions = ("approval_delay", "approval_duration",
-                          "created", "enabled", "fingerprint",
-                          "host", "interval", "last_checked_ok",
+                          "created", "enabled", "expires",
+                          "fingerprint", "host", "interval",
+                          "last_approval_request", "last_checked_ok",
                           "last_enabled", "name", "timeout")
-    client_defaults = { "timeout": "5m",
-                        "extended_timeout": "15m",
-                        "interval": "2m",
+    client_defaults = { "timeout": "PT5M",
+                        "extended_timeout": "PT15M",
+                        "interval": "PT2M",
                         "checker": "fping -q -- %%(host)s",
                         "host": "",
-                        "approval_delay": "0s",
-                        "approval_duration": "1s",
+                        "approval_delay": "PT0S",
+                        "approval_duration": "PT1S",
                         "approved_by_default": "True",
                         "enabled": "True",
                         }
@@ -458,7 +482,7 @@
     
     def approval_delay_milliseconds(self):
         return timedelta_to_milliseconds(self.approval_delay)
-
+    
     @staticmethod
     def config_parser(config):
         """Construct a new dict of client settings of this form:
@@ -489,8 +513,8 @@
                           "rb") as secfile:
                     client["secret"] = secfile.read()
             else:
-                raise TypeError("No secret or secfile for section %s"
-                                % section)
+                raise TypeError("No secret or secfile for section {0}"
+                                .format(section))
             client["timeout"] = string_to_delta(section["timeout"])
             client["extended_timeout"] = string_to_delta(
                 section["extended_timeout"])
@@ -505,13 +529,12 @@
             client["last_checker_status"] = -2
         
         return settings
-        
-        
-    def __init__(self, settings, name = None):
-        """Note: the 'checker' key in 'config' sets the
-        'checker_command' attribute and *not* the 'checker'
-        attribute."""
+    
+    def __init__(self, settings, name = None, server_settings=None):
         self.name = name
+        if server_settings is None:
+            server_settings = {}
+        self.server_settings = server_settings
         # adding all client settings
         for setting, value in settings.iteritems():
             setattr(self, setting, value)
@@ -525,7 +548,7 @@
         else:
             self.last_enabled = None
             self.expires = None
-       
+        
         logger.debug("Creating client %r", self.name)
         # Uppercase and remove spaces from fingerprint for later
         # comparison purposes with return value from the fingerprint()
@@ -533,7 +556,7 @@
         logger.debug("  Fingerprint: %s", self.fingerprint)
         self.created = settings.get("created",
                                     datetime.datetime.utcnow())
-
+        
         # attributes specific for this server instance
         self.checker = None
         self.checker_initiator_tag = None
@@ -567,29 +590,29 @@
         if getattr(self, "enabled", False):
             # Already enabled
             return
-        self.send_changedstate()
         self.expires = datetime.datetime.utcnow() + self.timeout
         self.enabled = True
         self.last_enabled = datetime.datetime.utcnow()
         self.init_checker()
+        self.send_changedstate()
     
     def disable(self, quiet=True):
         """Disable this client."""
         if not getattr(self, "enabled", False):
             return False
         if not quiet:
-            self.send_changedstate()
-        if not quiet:
             logger.info("Disabling client %s", self.name)
-        if getattr(self, "disable_initiator_tag", False):
+        if getattr(self, "disable_initiator_tag", None) is not None:
             gobject.source_remove(self.disable_initiator_tag)
             self.disable_initiator_tag = None
         self.expires = None
-        if getattr(self, "checker_initiator_tag", False):
+        if getattr(self, "checker_initiator_tag", None) is not None:
             gobject.source_remove(self.checker_initiator_tag)
             self.checker_initiator_tag = None
         self.stop_checker()
         self.enabled = False
+        if not quiet:
+            self.send_changedstate()
         # Do not run this again if called by a gobject.timeout_add
         return False
     
@@ -599,10 +622,14 @@
     def init_checker(self):
         # Schedule a new checker to be started an 'interval' from now,
         # and every interval from then on.
+        if self.checker_initiator_tag is not None:
+            gobject.source_remove(self.checker_initiator_tag)
         self.checker_initiator_tag = (gobject.timeout_add
                                       (self.interval_milliseconds(),
                                        self.start_checker))
         # Schedule a disable() when 'timeout' has passed
+        if self.disable_initiator_tag is not None:
+            gobject.source_remove(self.disable_initiator_tag)
         self.disable_initiator_tag = (gobject.timeout_add
                                    (self.timeout_milliseconds(),
                                     self.disable))
@@ -639,6 +666,7 @@
             timeout = self.timeout
         if self.disable_initiator_tag is not None:
             gobject.source_remove(self.disable_initiator_tag)
+            self.disable_initiator_tag = None
         if getattr(self, "enabled", False):
             self.disable_initiator_tag = (gobject.timeout_add
                                           (timedelta_to_milliseconds
@@ -654,10 +682,10 @@
         If a checker already exists, leave it running and do
         nothing."""
         # The reason for not killing a running checker is that if we
-        # did that, then if a checker (for some reason) started
-        # running slowly and taking more than 'interval' time, the
-        # client would inevitably timeout, since no checker would get
-        # a chance to run to completion.  If we instead leave running
+        # did that, and if a checker (for some reason) started running
+        # slowly and taking more than 'interval' time, then the client
+        # would inevitably timeout, since no checker would get a
+        # chance to run to completion.  If we instead leave running
         # checkers alone, the checker would have to take more time
         # than 'timeout' for the client to be disabled, which is as it
         # should be.
@@ -665,10 +693,11 @@
         # If a checker exists, make sure it is not a zombie
         try:
             pid, status = os.waitpid(self.checker.pid, os.WNOHANG)
-        except (AttributeError, OSError) as error:
-            if (isinstance(error, OSError)
-                and error.errno != errno.ECHILD):
-                raise error
+        except AttributeError:
+            pass
+        except OSError as error:
+            if error.errno != errno.ECHILD:
+                raise
         else:
             if pid:
                 logger.warning("Checker was a zombie")
@@ -677,25 +706,17 @@
                                       self.current_checker_command)
         # Start a new checker if needed
         if self.checker is None:
+            # Escape attributes for the shell
+            escaped_attrs = dict(
+                (attr, re.escape(unicode(getattr(self, attr))))
+                for attr in
+                self.runtime_expansions)
             try:
-                # In case checker_command has exactly one % operator
-                command = self.checker_command % self.host
-            except TypeError:
-                # Escape attributes for the shell
-                escaped_attrs = dict(
-                    (attr,
-                     re.escape(unicode(str(getattr(self, attr, "")),
-                                       errors=
-                                       'replace')))
-                    for attr in
-                    self.runtime_expansions)
-                
-                try:
-                    command = self.checker_command % escaped_attrs
-                except TypeError as error:
-                    logger.error('Could not format string "%s":'
-                                 ' %s', self.checker_command, error)
-                    return True # Try again later
+                command = self.checker_command % escaped_attrs
+            except TypeError as error:
+                logger.error('Could not format string "%s"',
+                             self.checker_command, exc_info=error)
+                return True # Try again later
             self.current_checker_command = command
             try:
                 logger.info("Starting checker %r for %s",
@@ -704,22 +725,40 @@
                 # in normal mode, that is already done by daemon(),
                 # and in debug mode we don't want to.  (Stdin is
                 # always replaced by /dev/null.)
+                # The exception is when not debugging but nevertheless
+                # running in the foreground; use the previously
+                # created wnull.
+                popen_args = {}
+                if (not self.server_settings["debug"]
+                    and self.server_settings["foreground"]):
+                    popen_args.update({"stdout": wnull,
+                                       "stderr": wnull })
                 self.checker = subprocess.Popen(command,
                                                 close_fds=True,
-                                                shell=True, cwd="/")
-                self.checker_callback_tag = (gobject.child_watch_add
-                                             (self.checker.pid,
-                                              self.checker_callback,
-                                              data=command))
-                # The checker may have completed before the gobject
-                # watch was added.  Check for this.
+                                                shell=True, cwd="/",
+                                                **popen_args)
+            except OSError as error:
+                logger.error("Failed to start subprocess",
+                             exc_info=error)
+                return True
+            self.checker_callback_tag = (gobject.child_watch_add
+                                         (self.checker.pid,
+                                          self.checker_callback,
+                                          data=command))
+            # The checker may have completed before the gobject
+            # watch was added.  Check for this.
+            try:
                 pid, status = os.waitpid(self.checker.pid, os.WNOHANG)
-                if pid:
-                    gobject.source_remove(self.checker_callback_tag)
-                    self.checker_callback(pid, status, command)
             except OSError as error:
-                logger.error("Failed to start subprocess: %s",
-                             error)
+                if error.errno == errno.ECHILD:
+                    # This should never happen
+                    logger.error("Child process vanished",
+                                 exc_info=error)
+                    return True
+                raise
+            if pid:
+                gobject.source_remove(self.checker_callback_tag)
+                self.checker_callback(pid, status, command)
         # Re-run this periodically if run by gobject.timeout_add
         return True
     
@@ -732,10 +771,10 @@
             return
         logger.debug("Stopping checker for %(name)s", vars(self))
         try:
-            os.kill(self.checker.pid, signal.SIGTERM)
+            self.checker.terminate()
             #time.sleep(0.5)
             #if self.checker.poll() is None:
-            #    os.kill(self.checker.pid, signal.SIGKILL)
+            #    self.checker.kill()
         except OSError as error:
             if error.errno != errno.ESRCH: # No such process
                 raise
@@ -758,7 +797,7 @@
     # "Set" method, so we fail early here:
     if byte_arrays and signature != "ay":
         raise ValueError("Byte arrays not supported for non-'ay'"
-                         " signature %r" % signature)
+                         " signature {0!r}".format(signature))
     def decorator(func):
         func._dbus_is_property = True
         func._dbus_interface = dbus_interface
@@ -773,7 +812,7 @@
 
 
 def dbus_interface_annotations(dbus_interface):
-    """Decorator for marking functions returning interface annotations.
+    """Decorator for marking functions returning interface annotations
     
     Usage:
     
@@ -791,6 +830,24 @@
     return decorator
 
 
+def dbus_annotations(annotations):
+    """Decorator to annotate D-Bus methods, signals or properties
+    Usage:
+    
+    @dbus_service_property("org.example.Interface", signature="b",
+                           access="r")
+    @dbus_annotations({{"org.freedesktop.DBus.Deprecated": "true",
+                        "org.freedesktop.DBus.Property."
+                        "EmitsChangedSignal": "false"})
+    def Property_dbus_property(self):
+        return dbus.Boolean(False)
+    """
+    def decorator(func):
+        func._dbus_annotations = annotations
+        return func
+    return decorator
+
+
 class DBusPropertyException(dbus.exceptions.DBusException):
     """A base class for D-Bus property-related exceptions
     """
@@ -831,7 +888,8 @@
     def _get_all_dbus_things(self, thing):
         """Returns a generator of (name, attribute) pairs
         """
-        return ((athing.__get__(self)._dbus_name,
+        return ((getattr(athing.__get__(self), "_dbus_name",
+                         name),
                  athing.__get__(self))
                 for cls in self.__class__.__mro__
                 for name, athing in
@@ -879,7 +937,9 @@
             # The byte_arrays option is not supported yet on
             # signatures other than "ay".
             if prop._dbus_signature != "ay":
-                raise ValueError
+                raise ValueError("Byte arrays not supported for non-"
+                                 "'ay' signature {0!r}"
+                                 .format(prop._dbus_signature))
             value = dbus.ByteArray(b''.join(chr(byte)
                                             for byte in value))
         prop(value)
@@ -936,18 +996,37 @@
                             if prop._dbus_interface
                             == if_tag.getAttribute("name")):
                     if_tag.appendChild(tag)
+                # Add annotation tags
+                for typ in ("method", "signal", "property"):
+                    for tag in if_tag.getElementsByTagName(typ):
+                        annots = dict()
+                        for name, prop in (self.
+                                           _get_all_dbus_things(typ)):
+                            if (name == tag.getAttribute("name")
+                                and prop._dbus_interface
+                                == if_tag.getAttribute("name")):
+                                annots.update(getattr
+                                              (prop,
+                                               "_dbus_annotations",
+                                               {}))
+                        for name, value in annots.iteritems():
+                            ann_tag = document.createElement(
+                                "annotation")
+                            ann_tag.setAttribute("name", name)
+                            ann_tag.setAttribute("value", value)
+                            tag.appendChild(ann_tag)
                 # Add interface annotation tags
                 for annotation, value in dict(
-                    itertools.chain(
-                        *(annotations().iteritems()
-                          for name, annotations in
-                          self._get_all_dbus_things("interface")
-                          if name == if_tag.getAttribute("name")
-                          ))).iteritems():
-                    attr_tag = document.createElement("annotation")
-                    attr_tag.setAttribute("name", annotation)
-                    attr_tag.setAttribute("value", value)
-                    if_tag.appendChild(attr_tag)
+                    itertools.chain.from_iterable(
+                        annotations().iteritems()
+                        for name, annotations in
+                        self._get_all_dbus_things("interface")
+                        if name == if_tag.getAttribute("name")
+                        )).iteritems():
+                    ann_tag = document.createElement("annotation")
+                    ann_tag.setAttribute("name", annotation)
+                    ann_tag.setAttribute("value", value)
+                    if_tag.appendChild(ann_tag)
                 # Add the names to the return values for the
                 # "org.freedesktop.DBus.Properties" methods
                 if (if_tag.getAttribute("name")
@@ -968,11 +1047,11 @@
         except (AttributeError, xml.dom.DOMException,
                 xml.parsers.expat.ExpatError) as error:
             logger.error("Failed to override Introspection method",
-                         error)
+                         exc_info=error)
         return xmlstring
 
 
-def datetime_to_dbus (dt, variant_level=0):
+def datetime_to_dbus(dt, variant_level=0):
     """Convert a UTC datetime.datetime() to a D-Bus type."""
     if dt is None:
         return dbus.String("", variant_level = variant_level)
@@ -980,36 +1059,52 @@
                        variant_level=variant_level)
 
 
-class AlternateDBusNamesMetaclass(DBusObjectWithProperties
-                                  .__metaclass__):
-    """Applied to an empty subclass of a D-Bus object, this metaclass
-    will add additional D-Bus attributes matching a certain pattern.
+def alternate_dbus_interfaces(alt_interface_names, deprecate=True):
+    """A class decorator; applied to a subclass of
+    dbus.service.Object, it will add alternate D-Bus attributes with
+    interface names according to the "alt_interface_names" mapping.
+    Usage:
+    
+    @alternate_dbus_interfaces({"org.example.Interface":
+                                    "net.example.AlternateInterface"})
+    class SampleDBusObject(dbus.service.Object):
+        @dbus.service.method("org.example.Interface")
+        def SampleDBusMethod():
+            pass
+    
+    The above "SampleDBusMethod" on "SampleDBusObject" will be
+    reachable via two interfaces: "org.example.Interface" and
+    "net.example.AlternateInterface", the latter of which will have
+    its D-Bus annotation "org.freedesktop.DBus.Deprecated" set to
+    "true", unless "deprecate" is passed with a False value.
+    
+    This works for methods and signals, and also for D-Bus properties
+    (from DBusObjectWithProperties) and interfaces (from the
+    dbus_interface_annotations decorator).
     """
-    def __new__(mcs, name, bases, attr):
-        # Go through all the base classes which could have D-Bus
-        # methods, signals, or properties in them
-        old_interface_names = []
-        for base in (b for b in bases
-                     if issubclass(b, dbus.service.Object)):
-            # Go though all attributes of the base class
-            for attrname, attribute in inspect.getmembers(base):
+    def wrapper(cls):
+        for orig_interface_name, alt_interface_name in (
+            alt_interface_names.iteritems()):
+            attr = {}
+            interface_names = set()
+            # Go though all attributes of the class
+            for attrname, attribute in inspect.getmembers(cls):
                 # Ignore non-D-Bus attributes, and D-Bus attributes
                 # with the wrong interface name
                 if (not hasattr(attribute, "_dbus_interface")
                     or not attribute._dbus_interface
-                    .startswith("se.recompile.Mandos")):
+                    .startswith(orig_interface_name)):
                     continue
                 # Create an alternate D-Bus interface name based on
                 # the current name
                 alt_interface = (attribute._dbus_interface
-                                 .replace("se.recompile.Mandos",
-                                          "se.bsnet.fukt.Mandos"))
-                if alt_interface != attribute._dbus_interface:
-                    old_interface_names.append(alt_interface)
+                                 .replace(orig_interface_name,
+                                          alt_interface_name))
+                interface_names.add(alt_interface)
                 # Is this a D-Bus signal?
                 if getattr(attribute, "_dbus_is_signal", False):
-                    # Extract the original non-method function by
-                    # black magic
+                    # Extract the original non-method undecorated
+                    # function by black magic
                     nonmethod_func = (dict(
                             zip(attribute.func_code.co_freevars,
                                 attribute.__closure__))["func"]
@@ -1026,9 +1121,16 @@
                                 nonmethod_func.func_name,
                                 nonmethod_func.func_defaults,
                                 nonmethod_func.func_closure)))
+                    # Copy annotations, if any
+                    try:
+                        new_function._dbus_annotations = (
+                            dict(attribute._dbus_annotations))
+                    except AttributeError:
+                        pass
                     # Define a creator of a function to call both the
-                    # old and new functions, so both the old and new
-                    # signals gets sent when the function is called
+                    # original and alternate functions, so both the
+                    # original and alternate signals gets sent when
+                    # the function is called
                     def fixscope(func1, func2):
                         """This function is a scope container to pass
                         func1 and func2 to the "call_both" function
@@ -1041,8 +1143,7 @@
                         return call_both
                     # Create the "call_both" function and add it to
                     # the class
-                    attr[attrname] = fixscope(attribute,
-                                              new_function)
+                    attr[attrname] = fixscope(attribute, new_function)
                 # Is this a D-Bus method?
                 elif getattr(attribute, "_dbus_is_method", False):
                     # Create a new, but exactly alike, function
@@ -1059,6 +1160,12 @@
                                         attribute.func_name,
                                         attribute.func_defaults,
                                         attribute.func_closure)))
+                    # Copy annotations, if any
+                    try:
+                        attr[attrname]._dbus_annotations = (
+                            dict(attribute._dbus_annotations))
+                    except AttributeError:
+                        pass
                 # Is this a D-Bus property?
                 elif getattr(attribute, "_dbus_is_property", False):
                     # Create a new, but exactly alike, function
@@ -1078,6 +1185,12 @@
                                         attribute.func_name,
                                         attribute.func_defaults,
                                         attribute.func_closure)))
+                    # Copy annotations, if any
+                    try:
+                        attr[attrname]._dbus_annotations = (
+                            dict(attribute._dbus_annotations))
+                    except AttributeError:
+                        pass
                 # Is this a D-Bus interface?
                 elif getattr(attribute, "_dbus_is_interface", False):
                     # Create a new, but exactly alike, function
@@ -1092,21 +1205,31 @@
                                         attribute.func_name,
                                         attribute.func_defaults,
                                         attribute.func_closure)))
-        # Deprecate all old interfaces
-        basename="_AlternateDBusNamesMetaclass_interface_annotation{0}"
-        for old_interface_name in old_interface_names:
-            @dbus_interface_annotations(old_interface_name)
-            def func(self):
-                return { "org.freedesktop.DBus.Deprecated": "true" }
-            # Find an unused name
-            for aname in (basename.format(i) for i in
-                          itertools.count()):
-                if aname not in attr:
-                    attr[aname] = func
-                    break
-        return type.__new__(mcs, name, bases, attr)
-
-
+            if deprecate:
+                # Deprecate all alternate interfaces
+                iname="_AlternateDBusNames_interface_annotation{0}"
+                for interface_name in interface_names:
+                    @dbus_interface_annotations(interface_name)
+                    def func(self):
+                        return { "org.freedesktop.DBus.Deprecated":
+                                     "true" }
+                    # Find an unused name
+                    for aname in (iname.format(i)
+                                  for i in itertools.count()):
+                        if aname not in attr:
+                            attr[aname] = func
+                            break
+            if interface_names:
+                # Replace the class with a new subclass of it with
+                # methods, signals, etc. as created above.
+                cls = type(b"{0}Alternate".format(cls.__name__),
+                           (cls,), attr)
+        return cls
+    return wrapper
+
+
+@alternate_dbus_interfaces({"se.recompile.Mandos":
+                                "se.bsnet.fukt.Mandos"})
 class ClientDBus(Client, DBusObjectWithProperties):
     """A Client class using D-Bus
     
@@ -1132,7 +1255,7 @@
                                  ("/clients/" + client_object_name))
         DBusObjectWithProperties.__init__(self, self.bus,
                                           self.dbus_object_path)
-        
+    
     def notifychangeproperty(transform_func,
                              dbus_name, type_func=lambda x: x,
                              variant_level=1):
@@ -1161,7 +1284,6 @@
         
         return property(lambda self: getattr(self, attrname), setter)
     
-    
     expires = notifychangeproperty(datetime_to_dbus, "Expires")
     approvals_pending = notifychangeproperty(dbus.Boolean,
                                              "ApprovalPending",
@@ -1231,11 +1353,7 @@
                                        *args, **kwargs)
     
     def start_checker(self, *args, **kwargs):
-        old_checker = self.checker
-        if self.checker is not None:
-            old_checker_pid = self.checker.pid
-        else:
-            old_checker_pid = None
+        old_checker_pid = getattr(self.checker, "pid", None)
         r = Client.start_checker(self, *args, **kwargs)
         # Only if new checker process was started
         if (self.checker is not None
@@ -1249,12 +1367,11 @@
         return False
     
     def approve(self, value=True):
-        self.send_changedstate()
         self.approved = value
         gobject.timeout_add(timedelta_to_milliseconds
                             (self.approval_duration),
                             self._reset_approved)
-    
+        self.send_changedstate()
     
     ## D-Bus methods, signals & properties
     _interface = "se.recompile.Mandos.Client"
@@ -1444,26 +1561,24 @@
     def Timeout_dbus_property(self, value=None):
         if value is None:       # get
             return dbus.UInt64(self.timeout_milliseconds())
+        old_timeout = self.timeout
         self.timeout = datetime.timedelta(0, 0, 0, value)
-        # Reschedule timeout
+        # Reschedule disabling
         if self.enabled:
             now = datetime.datetime.utcnow()
-            time_to_die = timedelta_to_milliseconds(
-                (self.last_checked_ok + self.timeout) - now)
-            if time_to_die <= 0:
+            self.expires += self.timeout - old_timeout
+            if self.expires <= now:
                 # The timeout has passed
                 self.disable()
             else:
-                self.expires = (now +
-                                datetime.timedelta(milliseconds =
-                                                   time_to_die))
                 if (getattr(self, "disable_initiator_tag", None)
                     is None):
                     return
                 gobject.source_remove(self.disable_initiator_tag)
-                self.disable_initiator_tag = (gobject.timeout_add
-                                              (time_to_die,
-                                               self.disable))
+                self.disable_initiator_tag = (
+                    gobject.timeout_add(
+                        timedelta_to_milliseconds(self.expires - now),
+                        self.disable))
     
     # ExtendedTimeout - property
     @dbus_service_property(_interface, signature="t",
@@ -1548,10 +1663,6 @@
         self._pipe.send(('setattr', name, value))
 
 
-class ClientDBusTransitional(ClientDBus):
-    __metaclass__ = AlternateDBusNamesMetaclass
-
-
 class ClientHandler(socketserver.BaseRequestHandler, object):
     """A class to handle client connections.
     
@@ -1593,7 +1704,7 @@
             logger.debug("Protocol version: %r", line)
             try:
                 if int(line.strip().split()[0]) > 1:
-                    raise RuntimeError
+                    raise RuntimeError(line)
             except (ValueError, IndexError, RuntimeError) as error:
                 logger.error("Unknown protocol version: %s", error)
                 return
@@ -1661,9 +1772,9 @@
                     #wait until timeout or approved
                     time = datetime.datetime.now()
                     client.changedstate.acquire()
-                    (client.changedstate.wait
-                     (float(client.timedelta_to_milliseconds(delay)
-                            / 1000)))
+                    client.changedstate.wait(
+                        float(timedelta_to_milliseconds(delay)
+                              / 1000))
                     client.changedstate.release()
                     time2 = datetime.datetime.now()
                     if (time2 - time) >= delay:
@@ -1685,7 +1796,8 @@
                     try:
                         sent = session.send(client.secret[sent_size:])
                     except gnutls.errors.GNUTLSError as error:
-                        logger.warning("gnutls send failed")
+                        logger.warning("gnutls send failed",
+                                       exc_info=error)
                         return
                     logger.debug("Sent: %d, remaining: %d",
                                  sent, len(client.secret)
@@ -1705,7 +1817,8 @@
                 try:
                     session.bye()
                 except gnutls.errors.GNUTLSError as error:
-                    logger.warning("GnuTLS bye failed")
+                    logger.warning("GnuTLS bye failed",
+                                   exc_info=error)
     
     @staticmethod
     def peer_certificate(session):
@@ -1783,8 +1896,7 @@
     def process_request(self, request, address):
         """Start a new process to process the request."""
         proc = multiprocessing.Process(target = self.sub_process_main,
-                                       args = (request,
-                                               address))
+                                       args = (request, address))
         proc.start()
         return proc
 
@@ -1805,7 +1917,7 @@
     
     def add_pipe(self, parent_pipe, proc):
         """Dummy function; override as necessary"""
-        raise NotImplementedError
+        raise NotImplementedError()
 
 
 class IPv6_TCPServer(MultiprocessingMixInWithPipe,
@@ -1818,12 +1930,42 @@
         use_ipv6:       Boolean; to use IPv6 or not
     """
     def __init__(self, server_address, RequestHandlerClass,
-                 interface=None, use_ipv6=True):
+                 interface=None, use_ipv6=True, socketfd=None):
+        """If socketfd is set, use that file descriptor instead of
+        creating a new one with socket.socket().
+        """
         self.interface = interface
         if use_ipv6:
             self.address_family = socket.AF_INET6
+        if socketfd is not None:
+            # Save the file descriptor
+            self.socketfd = socketfd
+            # Save the original socket.socket() function
+            self.socket_socket = socket.socket
+            # To implement --socket, we monkey patch socket.socket.
+            # 
+            # (When socketserver.TCPServer is a new-style class, we
+            # could make self.socket into a property instead of monkey
+            # patching socket.socket.)
+            # 
+            # Create a one-time-only replacement for socket.socket()
+            @functools.wraps(socket.socket)
+            def socket_wrapper(*args, **kwargs):
+                # Restore original function so subsequent calls are
+                # not affected.
+                socket.socket = self.socket_socket
+                del self.socket_socket
+                # This time only, return a new socket object from the
+                # saved file descriptor.
+                return socket.fromfd(self.socketfd, *args, **kwargs)
+            # Replace socket.socket() function with wrapper
+            socket.socket = socket_wrapper
+        # The socketserver.TCPServer.__init__ will call
+        # socket.socket(), which might be our replacement,
+        # socket_wrapper(), if socketfd was set.
         socketserver.TCPServer.__init__(self, server_address,
                                         RequestHandlerClass)
+    
     def server_bind(self):
         """This overrides the normal server_bind() function
         to bind to an interface if one was specified, and also NOT to
@@ -1837,17 +1979,18 @@
                 try:
                     self.socket.setsockopt(socket.SOL_SOCKET,
                                            SO_BINDTODEVICE,
-                                           str(self.interface
-                                               + '\0'))
+                                           str(self.interface + '\0'))
                 except socket.error as error:
-                    if error[0] == errno.EPERM:
-                        logger.error("No permission to"
-                                     " bind to interface %s",
-                                     self.interface)
-                    elif error[0] == errno.ENOPROTOOPT:
+                    if error.errno == errno.EPERM:
+                        logger.error("No permission to bind to"
+                                     " interface %s", self.interface)
+                    elif error.errno == errno.ENOPROTOOPT:
                         logger.error("SO_BINDTODEVICE not available;"
                                      " cannot bind to interface %s",
                                      self.interface)
+                    elif error.errno == errno.ENODEV:
+                        logger.error("Interface %s does not exist,"
+                                     " cannot bind", self.interface)
                     else:
                         raise
         # Only bind(2) the socket if we really need to.
@@ -1856,7 +1999,7 @@
                 if self.address_family == socket.AF_INET6:
                     any_address = "::" # in6addr_any
                 else:
-                    any_address = socket.INADDR_ANY
+                    any_address = "0.0.0.0" # INADDR_ANY
                 self.server_address = (any_address,
                                        self.server_address[1])
             elif not self.server_address[1]:
@@ -1883,7 +2026,7 @@
     """
     def __init__(self, server_address, RequestHandlerClass,
                  interface=None, use_ipv6=True, clients=None,
-                 gnutls_priority=None, use_dbus=True):
+                 gnutls_priority=None, use_dbus=True, socketfd=None):
         self.enabled = False
         self.clients = clients
         if self.clients is None:
@@ -1893,7 +2036,8 @@
         IPv6_TCPServer.__init__(self, server_address,
                                 RequestHandlerClass,
                                 interface = interface,
-                                use_ipv6 = use_ipv6)
+                                use_ipv6 = use_ipv6,
+                                socketfd = socketfd)
     def server_activate(self):
         if self.enabled:
             return socketserver.TCPServer.server_activate(self)
@@ -1912,22 +2056,8 @@
     
     def handle_ipc(self, source, condition, parent_pipe=None,
                    proc = None, client_object=None):
-        condition_names = {
-            gobject.IO_IN: "IN",   # There is data to read.
-            gobject.IO_OUT: "OUT", # Data can be written (without
-                                    # blocking).
-            gobject.IO_PRI: "PRI", # There is urgent data to read.
-            gobject.IO_ERR: "ERR", # Error condition.
-            gobject.IO_HUP: "HUP"  # Hung up (the connection has been
-                                    # broken, usually for pipes and
-                                    # sockets).
-            }
-        conditions_string = ' | '.join(name
-                                       for cond, name in
-                                       condition_names.iteritems()
-                                       if cond & condition)
         # error, or the other end of multiprocessing.Pipe has closed
-        if condition & (gobject.IO_ERR | condition & gobject.IO_HUP):
+        if condition & (gobject.IO_ERR | gobject.IO_HUP):
             # Wait for other process to exit
             proc.join()
             return False
@@ -1991,6 +2121,105 @@
         return True
 
 
+def rfc3339_duration_to_delta(duration):
+    """Parse an RFC 3339 "duration" and return a datetime.timedelta
+    
+    >>> rfc3339_duration_to_delta("P7D")
+    datetime.timedelta(7)
+    >>> rfc3339_duration_to_delta("PT60S")
+    datetime.timedelta(0, 60)
+    >>> rfc3339_duration_to_delta("PT60M")
+    datetime.timedelta(0, 3600)
+    >>> rfc3339_duration_to_delta("PT24H")
+    datetime.timedelta(1)
+    >>> rfc3339_duration_to_delta("P1W")
+    datetime.timedelta(7)
+    >>> rfc3339_duration_to_delta("PT5M30S")
+    datetime.timedelta(0, 330)
+    >>> rfc3339_duration_to_delta("P1DT3M20S")
+    datetime.timedelta(1, 200)
+    """
+    
+    # Parsing an RFC 3339 duration with regular expressions is not
+    # possible - there would have to be multiple places for the same
+    # values, like seconds.  The current code, while more esoteric, is
+    # cleaner without depending on a parsing library.  If Python had a
+    # built-in library for parsing we would use it, but we'd like to
+    # avoid excessive use of external libraries.
+    
+    # New type for defining tokens, syntax, and semantics all-in-one
+    Token = collections.namedtuple("Token",
+                                   ("regexp", # To match token; if
+                                              # "value" is not None,
+                                              # must have a "group"
+                                              # containing digits
+                                    "value",  # datetime.timedelta or
+                                              # None
+                                    "followers")) # Tokens valid after
+                                                  # this token
+    # RFC 3339 "duration" tokens, syntax, and semantics; taken from
+    # the "duration" ABNF definition in RFC 3339, Appendix A.
+    token_end = Token(re.compile(r"$"), None, frozenset())
+    token_second = Token(re.compile(r"(\d+)S"),
+                         datetime.timedelta(seconds=1),
+                         frozenset((token_end,)))
+    token_minute = Token(re.compile(r"(\d+)M"),
+                         datetime.timedelta(minutes=1),
+                         frozenset((token_second, token_end)))
+    token_hour = Token(re.compile(r"(\d+)H"),
+                       datetime.timedelta(hours=1),
+                       frozenset((token_minute, token_end)))
+    token_time = Token(re.compile(r"T"),
+                       None,
+                       frozenset((token_hour, token_minute,
+                                  token_second)))
+    token_day = Token(re.compile(r"(\d+)D"),
+                      datetime.timedelta(days=1),
+                      frozenset((token_time, token_end)))
+    token_month = Token(re.compile(r"(\d+)M"),
+                        datetime.timedelta(weeks=4),
+                        frozenset((token_day, token_end)))
+    token_year = Token(re.compile(r"(\d+)Y"),
+                       datetime.timedelta(weeks=52),
+                       frozenset((token_month, token_end)))
+    token_week = Token(re.compile(r"(\d+)W"),
+                       datetime.timedelta(weeks=1),
+                       frozenset((token_end,)))
+    token_duration = Token(re.compile(r"P"), None,
+                           frozenset((token_year, token_month,
+                                      token_day, token_time,
+                                      token_week))),
+    # Define starting values
+    value = datetime.timedelta() # Value so far
+    found_token = None
+    followers = frozenset(token_duration,) # Following valid tokens
+    s = duration                # String left to parse
+    # Loop until end token is found
+    while found_token is not token_end:
+        # Search for any currently valid tokens
+        for token in followers:
+            match = token.regexp.match(s)
+            if match is not None:
+                # Token found
+                if token.value is not None:
+                    # Value found, parse digits
+                    factor = int(match.group(1), 10)
+                    # Add to value so far
+                    value += factor * token.value
+                # Strip token from string
+                s = token.regexp.sub("", s, 1)
+                # Go to found token
+                found_token = token
+                # Set valid next tokens
+                followers = found_token.followers
+                break
+        else:
+            # No currently valid tokens were found
+            raise ValueError("Invalid RFC 3339 duration")
+    # End token found
+    return value
+
+
 def string_to_delta(interval):
     """Parse a string and return a datetime.timedelta
     
@@ -2007,6 +2236,12 @@
     >>> string_to_delta('5m 30s')
     datetime.timedelta(0, 330)
     """
+    
+    try:
+        return rfc3339_duration_to_delta(interval)
+    except ValueError:
+        pass
+    
     timevalue = datetime.timedelta(0)
     for s in interval.split():
         try:
@@ -2023,8 +2258,9 @@
             elif suffix == "w":
                 delta = datetime.timedelta(0, 0, 0, 0, 0, 0, value)
             else:
-                raise ValueError("Unknown suffix %r" % suffix)
-        except (ValueError, IndexError) as e:
+                raise ValueError("Unknown suffix {0!r}"
+                                 .format(suffix))
+        except IndexError as e:
             raise ValueError(*(e.args))
         timevalue += delta
     return timevalue
@@ -2046,8 +2282,8 @@
         null = os.open(os.devnull, os.O_NOCTTY | os.O_RDWR)
         if not stat.S_ISCHR(os.fstat(null).st_mode):
             raise OSError(errno.ENODEV,
-                          "%s not a character device"
-                          % os.devnull)
+                          "{0} not a character device"
+                          .format(os.devnull))
         os.dup2(null, sys.stdin.fileno())
         os.dup2(null, sys.stdout.fileno())
         os.dup2(null, sys.stderr.fileno())
@@ -2062,7 +2298,7 @@
     
     parser = argparse.ArgumentParser()
     parser.add_argument("-v", "--version", action="version",
-                        version = "%%(prog)s %s" % version,
+                        version = "%(prog)s {0}".format(version),
                         help="show version number and exit")
     parser.add_argument("-i", "--interface", metavar="IF",
                         help="Bind to interface IF")
@@ -2074,7 +2310,7 @@
                         help="Run self-test")
     parser.add_argument("--debug", action="store_true",
                         help="Debug mode; run in foreground and log"
-                        " to terminal")
+                        " to terminal", default=None)
     parser.add_argument("--debuglevel", metavar="LEVEL",
                         help="Debug level for stdout output")
     parser.add_argument("--priority", help="GnuTLS"
@@ -2087,21 +2323,30 @@
                         " files")
     parser.add_argument("--no-dbus", action="store_false",
                         dest="use_dbus", help="Do not provide D-Bus"
-                        " system bus interface")
+                        " system bus interface", default=None)
     parser.add_argument("--no-ipv6", action="store_false",
-                        dest="use_ipv6", help="Do not use IPv6")
+                        dest="use_ipv6", help="Do not use IPv6",
+                        default=None)
     parser.add_argument("--no-restore", action="store_false",
                         dest="restore", help="Do not restore stored"
-                        " state")
+                        " state", default=None)
+    parser.add_argument("--socket", type=int,
+                        help="Specify a file descriptor to a network"
+                        " socket to use instead of creating one")
     parser.add_argument("--statedir", metavar="DIR",
                         help="Directory to save/restore state in")
+    parser.add_argument("--foreground", action="store_true",
+                        help="Run in foreground", default=None)
+    parser.add_argument("--no-zeroconf", action="store_false",
+                        dest="zeroconf", help="Do not use Zeroconf",
+                        default=None)
     
     options = parser.parse_args()
     
     if options.check:
         import doctest
-        doctest.testmod()
-        sys.exit()
+        fail_count, test_count = doctest.testmod()
+        sys.exit(os.EX_OK if fail_count == 0 else 1)
     
     # Default values for config file for server-global settings
     server_defaults = { "interface": "",
@@ -2109,13 +2354,16 @@
                         "port": "",
                         "debug": "False",
                         "priority":
-                        "SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP",
+                        "SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP:+SIGN-RSA-SHA224:+SIGN-RSA-RMD160",
                         "servicename": "Mandos",
                         "use_dbus": "True",
                         "use_ipv6": "True",
                         "debuglevel": "",
                         "restore": "True",
-                        "statedir": "/var/lib/mandos"
+                        "socket": "",
+                        "statedir": "/var/lib/mandos",
+                        "foreground": "False",
+                        "zeroconf": "True",
                         }
     
     # Parse config file for server-global settings
@@ -2126,12 +2374,21 @@
     # Convert the SafeConfigParser object to a dict
     server_settings = server_config.defaults()
     # Use the appropriate methods on the non-string config options
-    for option in ("debug", "use_dbus", "use_ipv6"):
+    for option in ("debug", "use_dbus", "use_ipv6", "foreground"):
         server_settings[option] = server_config.getboolean("DEFAULT",
                                                            option)
     if server_settings["port"]:
         server_settings["port"] = server_config.getint("DEFAULT",
                                                        "port")
+    if server_settings["socket"]:
+        server_settings["socket"] = server_config.getint("DEFAULT",
+                                                         "socket")
+        # Later, stdin will, and stdout and stderr might, be dup'ed
+        # over with an opened os.devnull.  But we don't want this to
+        # happen with a supplied network socket.
+        if 0 <= server_settings["socket"] <= 2:
+            server_settings["socket"] = os.dup(server_settings
+                                               ["socket"])
     del server_config
     
     # Override the settings from the config file with command line
@@ -2139,7 +2396,7 @@
     for option in ("interface", "address", "port", "debug",
                    "priority", "servicename", "configdir",
                    "use_dbus", "use_ipv6", "debuglevel", "restore",
-                   "statedir"):
+                   "statedir", "socket", "foreground", "zeroconf"):
         value = getattr(options, option)
         if value is not None:
             server_settings[option] = value
@@ -2148,10 +2405,23 @@
     for option in server_settings.keys():
         if type(server_settings[option]) is str:
             server_settings[option] = unicode(server_settings[option])
+    # Force all boolean options to be boolean
+    for option in ("debug", "use_dbus", "use_ipv6", "restore",
+                   "foreground", "zeroconf"):
+        server_settings[option] = bool(server_settings[option])
+    # Debug implies foreground
+    if server_settings["debug"]:
+        server_settings["foreground"] = True
     # Now we have our good server settings in "server_settings"
     
     ##################################################################
     
+    if (not server_settings["zeroconf"] and
+        not (server_settings["port"]
+             or server_settings["socket"] != "")):
+            parser.error("Needs port or socket to work without"
+                         " Zeroconf")
+    
     # For convenience
     debug = server_settings["debug"]
     debuglevel = server_settings["debuglevel"]
@@ -2159,6 +2429,8 @@
     use_ipv6 = server_settings["use_ipv6"]
     stored_state_path = os.path.join(server_settings["statedir"],
                                      stored_state_file)
+    foreground = server_settings["foreground"]
+    zeroconf = server_settings["zeroconf"]
     
     if debug:
         initlogger(debug, logging.DEBUG)
@@ -2171,9 +2443,10 @@
     
     if server_settings["servicename"] != "Mandos":
         syslogger.setFormatter(logging.Formatter
-                               ('Mandos (%s) [%%(process)d]:'
-                                ' %%(levelname)s: %%(message)s'
-                                % server_settings["servicename"]))
+                               ('Mandos ({0}) [%(process)d]:'
+                                ' %(levelname)s: %(message)s'
+                                .format(server_settings
+                                        ["servicename"])))
     
     # Parse config file with clients
     client_config = configparser.SafeConfigParser(Client
@@ -2184,6 +2457,9 @@
     global mandos_dbus_service
     mandos_dbus_service = None
     
+    socketfd = None
+    if server_settings["socket"] != "":
+        socketfd = server_settings["socket"]
     tcp_server = MandosServer((server_settings["address"],
                                server_settings["port"]),
                               ClientHandler,
@@ -2192,34 +2468,35 @@
                               use_ipv6=use_ipv6,
                               gnutls_priority=
                               server_settings["priority"],
-                              use_dbus=use_dbus)
-    if not debug:
-        pidfilename = "/var/run/mandos.pid"
+                              use_dbus=use_dbus,
+                              socketfd=socketfd)
+    if not foreground:
+        pidfilename = "/run/mandos.pid"
+        if not os.path.isdir("/run/."):
+            pidfilename = "/var/run/mandos.pid"
+        pidfile = None
         try:
             pidfile = open(pidfilename, "w")
-        except IOError:
-            logger.error("Could not open file %r", pidfilename)
+        except IOError as e:
+            logger.error("Could not open file %r", pidfilename,
+                         exc_info=e)
     
-    try:
-        uid = pwd.getpwnam("_mandos").pw_uid
-        gid = pwd.getpwnam("_mandos").pw_gid
-    except KeyError:
+    for name in ("_mandos", "mandos", "nobody"):
         try:
-            uid = pwd.getpwnam("mandos").pw_uid
-            gid = pwd.getpwnam("mandos").pw_gid
+            uid = pwd.getpwnam(name).pw_uid
+            gid = pwd.getpwnam(name).pw_gid
+            break
         except KeyError:
-            try:
-                uid = pwd.getpwnam("nobody").pw_uid
-                gid = pwd.getpwnam("nobody").pw_gid
-            except KeyError:
-                uid = 65534
-                gid = 65534
+            continue
+    else:
+        uid = 65534
+        gid = 65534
     try:
         os.setgid(gid)
         os.setuid(uid)
     except OSError as error:
-        if error[0] != errno.EPERM:
-            raise error
+        if error.errno != errno.EPERM:
+            raise
     
     if debug:
         # Enable all possible GnuTLS debugging
@@ -2242,10 +2519,12 @@
             os.close(null)
     
     # Need to fork before connecting to D-Bus
-    if not debug:
+    if not foreground:
         # Close all input and output, do double fork, etc.
         daemon()
     
+    # multiprocessing will use threads, so before we use gobject we
+    # need to inform gobject that threads will be used.
     gobject.threads_init()
     
     global main_loop
@@ -2262,31 +2541,39 @@
                             ("se.bsnet.fukt.Mandos", bus,
                              do_not_queue=True))
         except dbus.exceptions.NameExistsException as e:
-            logger.error(unicode(e) + ", disabling D-Bus")
+            logger.error("Disabling D-Bus:", exc_info=e)
             use_dbus = False
             server_settings["use_dbus"] = False
             tcp_server.use_dbus = False
-    protocol = avahi.PROTO_INET6 if use_ipv6 else avahi.PROTO_INET
-    service = AvahiServiceToSyslog(name =
-                                   server_settings["servicename"],
-                                   servicetype = "_mandos._tcp",
-                                   protocol = protocol, bus = bus)
-    if server_settings["interface"]:
-        service.interface = (if_nametoindex
-                             (str(server_settings["interface"])))
+    if zeroconf:
+        protocol = avahi.PROTO_INET6 if use_ipv6 else avahi.PROTO_INET
+        service = AvahiServiceToSyslog(name =
+                                       server_settings["servicename"],
+                                       servicetype = "_mandos._tcp",
+                                       protocol = protocol, bus = bus)
+        if server_settings["interface"]:
+            service.interface = (if_nametoindex
+                                 (str(server_settings["interface"])))
     
     global multiprocessing_manager
     multiprocessing_manager = multiprocessing.Manager()
     
     client_class = Client
     if use_dbus:
-        client_class = functools.partial(ClientDBusTransitional,
-                                         bus = bus)
+        client_class = functools.partial(ClientDBus, bus = bus)
     
     client_settings = Client.config_parser(client_config)
     old_client_settings = {}
     clients_data = {}
     
+    # This is used to redirect stdout and stderr for checker processes
+    global wnull
+    wnull = open(os.devnull, "w") # A writable /dev/null
+    # Only used if server is running in foreground but not in debug
+    # mode
+    if debug or not foreground:
+        wnull.close()
+    
     # Get client data and settings from last running state.
     if server_settings["restore"]:
         try:
@@ -2295,16 +2582,23 @@
                                                      (stored_state))
             os.remove(stored_state_path)
         except IOError as e:
-            logger.warning("Could not load persistent state: {0}"
-                           .format(e))
-            if e.errno != errno.ENOENT:
+            if e.errno == errno.ENOENT:
+                logger.warning("Could not load persistent state: {0}"
+                                .format(os.strerror(e.errno)))
+            else:
+                logger.critical("Could not load persistent state:",
+                                exc_info=e)
                 raise
         except EOFError as e:
             logger.warning("Could not load persistent state: "
-                           "EOFError: {0}".format(e))
+                           "EOFError:", exc_info=e)
     
     with PGPEngine() as pgp:
         for client_name, client in clients_data.iteritems():
+            # Skip removed clients
+            if client_name not in client_settings:
+                continue
+            
             # Decide which value to use after restoring saved state.
             # We have three different values: Old config file,
             # new config file, and saved state.
@@ -2360,7 +2654,6 @@
                              .format(client_name))
                 client["secret"] = (
                     client_settings[client_name]["secret"])
-
     
     # Add/remove clients based on new changes made to config
     for client_name in (set(old_client_settings)
@@ -2369,34 +2662,34 @@
     for client_name in (set(client_settings)
                         - set(old_client_settings)):
         clients_data[client_name] = client_settings[client_name]
-
+    
     # Create all client objects
     for client_name, client in clients_data.iteritems():
         tcp_server.clients[client_name] = client_class(
-            name = client_name, settings = client)
+            name = client_name, settings = client,
+            server_settings = server_settings)
     
     if not tcp_server.clients:
         logger.warning("No clients defined")
-        
-    if not debug:
-        try:
-            with pidfile:
-                pid = os.getpid()
-                pidfile.write(str(pid) + "\n".encode("utf-8"))
-            del pidfile
-        except IOError:
-            logger.error("Could not write to file %r with PID %d",
-                         pidfilename, pid)
-        except NameError:
-            # "pidfile" was never created
-            pass
+    
+    if not foreground:
+        if pidfile is not None:
+            try:
+                with pidfile:
+                    pid = os.getpid()
+                    pidfile.write(str(pid) + "\n".encode("utf-8"))
+            except IOError:
+                logger.error("Could not write to file %r with PID %d",
+                             pidfilename, pid)
+        del pidfile
         del pidfilename
-        signal.signal(signal.SIGINT, signal.SIG_IGN)
     
     signal.signal(signal.SIGHUP, lambda signum, frame: sys.exit())
     signal.signal(signal.SIGTERM, lambda signum, frame: sys.exit())
     
     if use_dbus:
+        @alternate_dbus_interfaces({"se.recompile.Mandos":
+                                        "se.bsnet.fukt.Mandos"})
         class MandosDBusService(DBusObjectWithProperties):
             """A D-Bus proxy object"""
             def __init__(self):
@@ -2456,15 +2749,15 @@
             
             del _interface
         
-        class MandosDBusServiceTransitional(MandosDBusService):
-            __metaclass__ = AlternateDBusNamesMetaclass
-        mandos_dbus_service = MandosDBusServiceTransitional()
+        mandos_dbus_service = MandosDBusService()
     
     def cleanup():
         "Cleanup function; run on exit"
-        service.cleanup()
+        if zeroconf:
+            service.cleanup()
         
         multiprocessing.active_children()
+        wnull.close()
         if not (tcp_server.clients or client_settings):
             return
         
@@ -2482,7 +2775,7 @@
                 # A list of attributes that can not be pickled
                 # + secret.
                 exclude = set(("bus", "changedstate", "secret",
-                               "checker"))
+                               "checker", "server_settings"))
                 for name, typ in (inspect.getmembers
                                   (dbus.service.Object)):
                     exclude.add(name)
@@ -2497,24 +2790,26 @@
                 del client_settings[client.name]["secret"]
         
         try:
-            tempfd, tempname = tempfile.mkstemp(suffix=".pickle",
-                                                prefix="clients-",
-                                                dir=os.path.dirname
-                                                (stored_state_path))
-            with os.fdopen(tempfd, "wb") as stored_state:
+            with (tempfile.NamedTemporaryFile
+                  (mode='wb', suffix=".pickle", prefix='clients-',
+                   dir=os.path.dirname(stored_state_path),
+                   delete=False)) as stored_state:
                 pickle.dump((clients, client_settings), stored_state)
+                tempname=stored_state.name
             os.rename(tempname, stored_state_path)
         except (IOError, OSError) as e:
-            logger.warning("Could not save persistent state: {0}"
-                           .format(e))
             if not debug:
                 try:
                     os.remove(tempname)
                 except NameError:
                     pass
-            if e.errno not in set((errno.ENOENT, errno.EACCES,
-                                   errno.EEXIST)):
-                raise e
+            if e.errno in (errno.ENOENT, errno.EACCES, errno.EEXIST):
+                logger.warning("Could not save persistent state: {0}"
+                               .format(os.strerror(e.errno)))
+            else:
+                logger.warning("Could not save persistent state:",
+                               exc_info=e)
+                raise
         
         # Delete all clients, and settings from config
         while tcp_server.clients:
@@ -2544,26 +2839,28 @@
     tcp_server.server_activate()
     
     # Find out what port we got
-    service.port = tcp_server.socket.getsockname()[1]
+    if zeroconf:
+        service.port = tcp_server.socket.getsockname()[1]
     if use_ipv6:
         logger.info("Now listening on address %r, port %d,"
-                    " flowinfo %d, scope_id %d"
-                    % tcp_server.socket.getsockname())
+                    " flowinfo %d, scope_id %d",
+                    *tcp_server.socket.getsockname())
     else:                       # IPv4
-        logger.info("Now listening on address %r, port %d"
-                    % tcp_server.socket.getsockname())
+        logger.info("Now listening on address %r, port %d",
+                    *tcp_server.socket.getsockname())
     
     #service.interface = tcp_server.socket.getsockname()[3]
     
     try:
-        # From the Avahi example code
-        try:
-            service.activate()
-        except dbus.exceptions.DBusException as error:
-            logger.critical("DBusException: %s", error)
-            cleanup()
-            sys.exit(1)
-        # End of Avahi example code
+        if zeroconf:
+            # From the Avahi example code
+            try:
+                service.activate()
+            except dbus.exceptions.DBusException as error:
+                logger.critical("D-Bus Exception", exc_info=error)
+                cleanup()
+                sys.exit(1)
+            # End of Avahi example code
         
         gobject.io_add_watch(tcp_server.fileno(), gobject.IO_IN,
                              lambda *args, **kwargs:
@@ -2573,7 +2870,7 @@
         logger.debug("Starting main loop")
         main_loop.run()
     except AvahiError as error:
-        logger.critical("AvahiError: %s", error)
+        logger.critical("Avahi Error", exc_info=error)
         cleanup()
         sys.exit(1)
     except KeyboardInterrupt:

=== modified file 'mandos-clients.conf.xml'
--- mandos-clients.conf.xml	2012-01-01 04:02:00 +0000
+++ mandos-clients.conf.xml	2014-06-22 02:19:30 +0000
@@ -3,7 +3,7 @@
 	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY CONFNAME "mandos-clients.conf">
 <!ENTITY CONFPATH "<filename>/etc/mandos/clients.conf</filename>">
-<!ENTITY TIMESTAMP "2012-01-01">
+<!ENTITY TIMESTAMP "2014-06-22">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -117,7 +117,7 @@
 	  <para>
 	    How long to wait for external approval before resorting to
 	    use the <option>approved_by_default</option> value.  The
-	    default is <quote>0s</quote>, i.e. not to wait.
+	    default is <quote>PT0S</quote>, i.e. not to wait.
 	  </para>
 	  <para>
 	    The format of <replaceable>TIME</replaceable> is the same
@@ -167,17 +167,23 @@
 	    This option is <emphasis>optional</emphasis>.
 	  </para>
 	  <para>
-	    This option allows you to override the default shell
-	    command that the server will use to check if the client is
-	    still up.  Any output of the command will be ignored, only
-	    the exit code is checked:  If the exit code of the command
-	    is zero, the client is considered up.  The command will be
-	    run using <quote><command><filename>/bin/sh</filename>
+	    This option overrides the default shell command that the
+	    server will use to check if the client is still up.  Any
+	    output of the command will be ignored, only the exit code
+	    is checked:  If the exit code of the command is zero, the
+	    client is considered up.  The command will be run using
+	    <quote><command><filename>/bin/sh</filename>
 	    <option>-c</option></command></quote>, so
 	    <varname>PATH</varname> will be searched.  The default
 	    value for the checker command is <quote><literal
 	    ><command>fping</command> <option>-q</option> <option
-	    >--</option> %%(host)s</literal></quote>.
+	    >--</option> %%(host)s</literal></quote>.  Note that
+	    <command>mandos-keygen</command>, when generating output
+	    to be inserted into this file, normally looks for an SSH
+	    server on the Mandos client, and, if it find one, outputs
+	    a <option>checker</option> option to check for the
+	    client’s key fingerprint – this is more secure against
+	    spoofing.
 	  </para>
 	  <para>
 	    In addition to normal start time expansion, this option
@@ -335,17 +341,12 @@
 	    <option>extended_timeout</option> option.
 	  </para>
 	  <para>
-	    The <replaceable>TIME</replaceable> is specified as a
-	    space-separated number of values, each of which is a
-	    number and a one-character suffix.  The suffix must be one
-	    of <quote>d</quote>, <quote>s</quote>, <quote>m</quote>,
-	    <quote>h</quote>, and <quote>w</quote> for days, seconds,
-	    minutes, hours, and weeks, respectively.  The values are
-	    added together to give the total time value, so all of
-	    <quote><literal>330s</literal></quote>,
-	    <quote><literal>110s 110s 110s</literal></quote>, and
-	    <quote><literal>5m 30s</literal></quote> will give a value
-	    of five minutes and thirty seconds.
+	    The <replaceable>TIME</replaceable> is specified as an RFC
+	    3339 duration; for example
+	    <quote><literal>P1Y2M3DT4H5M6S</literal></quote> meaning
+	    one year, two months, three days, four hours, five
+	    minutes, and six seconds.  Some values can be omitted, see
+	    RFC 3339 Appendix A for details.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -409,6 +410,7 @@
 	<quote><literal>approval_duration</literal></quote>,
 	<quote><literal>created</literal></quote>,
 	<quote><literal>enabled</literal></quote>,
+	<quote><literal>expires</literal></quote>,
 	<quote><literal>fingerprint</literal></quote>,
 	<quote><literal>host</literal></quote>,
 	<quote><literal>interval</literal></quote>,
@@ -464,8 +466,8 @@
     <informalexample>
       <programlisting>
 [DEFAULT]
-timeout = 5m
-interval = 2m
+timeout = PT5M
+interval = PT2M
 checker = fping -q -- %%(host)s
 
 # Client "foo"
@@ -488,15 +490,15 @@
         4T2zw4dxS5NswXWU0sVEXxjs6PYxuIiCTL7vdpx8QjBkrPWDrAbcMyBr2O
         QlnHIvPzEArRQLo=
 host = foo.example.org
-interval = 1m
+interval = PT1M
 
 # Client "bar"
 [bar]
 fingerprint = 3e393aeaefb84c7e89e2f547b3a107558fca3a27
 secfile = /etc/mandos/bar-secret
-timeout = 15m
+timeout = PT15M
 approved_by_default = False
-approval_delay = 30s
+approval_delay = PT30S
       </programlisting>
     </informalexample>
   </refsect1>
@@ -511,8 +513,24 @@
       <citerefentry><refentrytitle>mandos.conf</refentrytitle>
       <manvolnum>5</manvolnum></citerefentry>,
       <citerefentry><refentrytitle>mandos</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry>,
+      <citerefentry><refentrytitle>fping</refentrytitle>
       <manvolnum>8</manvolnum></citerefentry>
     </para>
+    <variablelist>
+      <varlistentry>
+	<term>
+	  RFC 3339: <citetitle>Date and Time on the Internet:
+	  Timestamps</citetitle>
+	</term>
+      <listitem>
+	<para>
+	  The time intervals are in the "duration" format, as
+	  specified in ABNF in Appendix A of RFC 3339.
+	</para>
+      </listitem>
+      </varlistentry>
+    </variablelist>
   </refsect1>
 </refentry>
 <!-- Local Variables: -->

=== modified file 'mandos-ctl'
--- mandos-ctl	2012-01-15 21:07:44 +0000
+++ mandos-ctl	2014-05-11 20:18:48 +0000
@@ -3,8 +3,8 @@
 # 
 # Mandos Monitor - Control and monitor the Mandos server
 # 
-# Copyright © 2008-2012 Teddy Hogeborn
-# Copyright © 2008-2012 Björn Påhlsson
+# Copyright © 2008-2014 Teddy Hogeborn
+# Copyright © 2008-2014 Björn Påhlsson
 # 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,7 +17,8 @@
 #     GNU General Public License for more details.
 # 
 # You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# along with this program.  If not, see
+# <http://www.gnu.org/licenses/>.
 # 
 # Contact the authors at <mandos@recompile.se>.
 # 
@@ -25,13 +26,18 @@
 from __future__ import (division, absolute_import, print_function,
                         unicode_literals)
 
+from future_builtins import *
+
 import sys
-import dbus
 import argparse
 import locale
 import datetime
 import re
 import os
+import collections
+import doctest
+
+import dbus
 
 locale.setlocale(locale.LC_ALL, "")
 
@@ -60,7 +66,7 @@
 server_path = "/"
 server_interface = domain + ".Mandos"
 client_interface = domain + ".Mandos.Client"
-version = "1.5.3"
+version = "1.6.5"
 
 def timedelta_to_milliseconds(td):
     """Convert a datetime.timedelta object to milliseconds"""
@@ -70,12 +76,112 @@
 
 def milliseconds_to_string(ms):
     td = datetime.timedelta(0, 0, 0, ms)
-    return ("%(days)s%(hours)02d:%(minutes)02d:%(seconds)02d"
-            % { "days": "%dT" % td.days if td.days else "",
-                "hours": td.seconds // 3600,
-                "minutes": (td.seconds % 3600) // 60,
-                "seconds": td.seconds % 60,
-                })
+    return ("{days}{hours:02}:{minutes:02}:{seconds:02}"
+            .format(days = "{0}T".format(td.days) if td.days else "",
+                    hours = td.seconds // 3600,
+                    minutes = (td.seconds % 3600) // 60,
+                    seconds = td.seconds % 60,
+                    ))
+
+
+def rfc3339_duration_to_delta(duration):
+    """Parse an RFC 3339 "duration" and return a datetime.timedelta
+    
+    >>> rfc3339_duration_to_delta("P7D")
+    datetime.timedelta(7)
+    >>> rfc3339_duration_to_delta("PT60S")
+    datetime.timedelta(0, 60)
+    >>> rfc3339_duration_to_delta("PT60M")
+    datetime.timedelta(0, 3600)
+    >>> rfc3339_duration_to_delta("PT24H")
+    datetime.timedelta(1)
+    >>> rfc3339_duration_to_delta("P1W")
+    datetime.timedelta(7)
+    >>> rfc3339_duration_to_delta("PT5M30S")
+    datetime.timedelta(0, 330)
+    >>> rfc3339_duration_to_delta("P1DT3M20S")
+    datetime.timedelta(1, 200)
+    """
+    
+    # Parsing an RFC 3339 duration with regular expressions is not
+    # possible - there would have to be multiple places for the same
+    # values, like seconds.  The current code, while more esoteric, is
+    # cleaner without depending on a parsing library.  If Python had a
+    # built-in library for parsing we would use it, but we'd like to
+    # avoid excessive use of external libraries.
+    
+    # New type for defining tokens, syntax, and semantics all-in-one
+    Token = collections.namedtuple("Token",
+                                   ("regexp", # To match token; if
+                                              # "value" is not None,
+                                              # must have a "group"
+                                              # containing digits
+                                    "value",  # datetime.timedelta or
+                                              # None
+                                    "followers")) # Tokens valid after
+                                                  # this token
+    # RFC 3339 "duration" tokens, syntax, and semantics; taken from
+    # the "duration" ABNF definition in RFC 3339, Appendix A.
+    token_end = Token(re.compile(r"$"), None, frozenset())
+    token_second = Token(re.compile(r"(\d+)S"),
+                         datetime.timedelta(seconds=1),
+                         frozenset((token_end,)))
+    token_minute = Token(re.compile(r"(\d+)M"),
+                         datetime.timedelta(minutes=1),
+                         frozenset((token_second, token_end)))
+    token_hour = Token(re.compile(r"(\d+)H"),
+                       datetime.timedelta(hours=1),
+                       frozenset((token_minute, token_end)))
+    token_time = Token(re.compile(r"T"),
+                       None,
+                       frozenset((token_hour, token_minute,
+                                  token_second)))
+    token_day = Token(re.compile(r"(\d+)D"),
+                      datetime.timedelta(days=1),
+                      frozenset((token_time, token_end)))
+    token_month = Token(re.compile(r"(\d+)M"),
+                        datetime.timedelta(weeks=4),
+                        frozenset((token_day, token_end)))
+    token_year = Token(re.compile(r"(\d+)Y"),
+                       datetime.timedelta(weeks=52),
+                       frozenset((token_month, token_end)))
+    token_week = Token(re.compile(r"(\d+)W"),
+                       datetime.timedelta(weeks=1),
+                       frozenset((token_end,)))
+    token_duration = Token(re.compile(r"P"), None,
+                           frozenset((token_year, token_month,
+                                      token_day, token_time,
+                                      token_week))),
+    # Define starting values
+    value = datetime.timedelta() # Value so far
+    found_token = None
+    followers = frozenset(token_duration,) # Following valid tokens
+    s = duration                # String left to parse
+    # Loop until end token is found
+    while found_token is not token_end:
+        # Search for any currently valid tokens
+        for token in followers:
+            match = token.regexp.match(s)
+            if match is not None:
+                # Token found
+                if token.value is not None:
+                    # Value found, parse digits
+                    factor = int(match.group(1), 10)
+                    # Add to value so far
+                    value += factor * token.value
+                # Strip token from string
+                s = token.regexp.sub("", s, 1)
+                # Go to found token
+                found_token = token
+                # Set valid next tokens
+                followers = found_token.followers
+                break
+        else:
+            # No currently valid tokens were found
+            raise ValueError("Invalid RFC 3339 duration")
+    # End token found
+    return value
+
 
 def string_to_delta(interval):
     """Parse a string and return a datetime.timedelta
@@ -93,8 +199,14 @@
     >>> string_to_delta("5m 30s")
     datetime.timedelta(0, 330)
     """
+    
+    try:
+        return rfc3339_duration_to_delta(interval)
+    except ValueError:
+        pass
+    
     value = datetime.timedelta(0)
-    regexp = re.compile("(\d+)([dsmhw]?)")
+    regexp = re.compile(r"(\d+)([dsmhw]?)")
     
     for num, suffix in regexp.findall(interval):
         if suffix == "d":
@@ -116,23 +228,25 @@
         if type(value) is dbus.Boolean:
             return "Yes" if value else "No"
         if keyword in ("Timeout", "Interval", "ApprovalDelay",
-                       "ApprovalDuration"):
+                       "ApprovalDuration", "ExtendedTimeout"):
             return milliseconds_to_string(value)
         return unicode(value)
     
     # Create format string to print table rows
-    format_string = " ".join("%%-%ds" %
-                             max(len(tablewords[key]),
-                                 max(len(valuetostring(client[key],
-                                                       key))
-                                     for client in
-                                     clients))
-                             for key in keywords)
+    format_string = " ".join("{{{key}:{width}}}".format(
+            width = max(len(tablewords[key]),
+                        max(len(valuetostring(client[key],
+                                              key))
+                            for client in
+                            clients)),
+            key = key) for key in keywords)
     # Print header line
-    print(format_string % tuple(tablewords[key] for key in keywords))
+    print(format_string.format(**tablewords))
     for client in clients:
-        print(format_string % tuple(valuetostring(client[key], key)
-                                    for key in keywords))
+        print(format_string.format(**dict((key,
+                                           valuetostring(client[key],
+                                                         key))
+                                          for key in keywords)))
 
 def has_actions(options):
     return any((options.enable,
@@ -157,7 +271,7 @@
 def main():
     parser = argparse.ArgumentParser()
     parser.add_argument("--version", action="version",
-                        version = "%%prog %s" % version,
+                        version = "%(prog)s {0}".format(version),
                         help="show version number and exit")
     parser.add_argument("-a", "--all", action="store_true",
                         help="Select all clients")
@@ -202,16 +316,22 @@
                         help="Approve any current client request")
     parser.add_argument("-D", "--deny", action="store_true",
                         help="Deny any current client request")
+    parser.add_argument("--check", action="store_true",
+                        help="Run self-test")
     parser.add_argument("client", nargs="*", help="Client name")
     options = parser.parse_args()
     
-    if has_actions(options) and not options.client and not options.all:
+    if has_actions(options) and not (options.client or options.all):
         parser.error("Options require clients names or --all.")
     if options.verbose and has_actions(options):
         parser.error("--verbose can only be used alone or with"
                      " --all.")
     if options.all and not has_actions(options):
         parser.error("--all requires an action.")
+
+    if options.check:
+        fail_count, test_count = doctest.testmod()
+        sys.exit(os.EX_OK if fail_count == 0 else 1)
     
     try:
         bus = dbus.SystemBus()
@@ -256,8 +376,8 @@
                     clients[client_objc] = client
                     break
             else:
-                print("Client not found on server: %r" % name,
-                      file=sys.stderr)
+                print("Client not found on server: {0!r}"
+                      .format(name), file=sys.stderr)
                 sys.exit(1)
     
     if not has_actions(options) and clients:
@@ -277,18 +397,28 @@
     else:
         # Process each client in the list by all selected options
         for client in clients:
+            def set_client_prop(prop, value):
+                """Set a Client D-Bus property"""
+                client.Set(client_interface, prop, value,
+                           dbus_interface=dbus.PROPERTIES_IFACE)
+            def set_client_prop_ms(prop, value):
+                """Set a Client D-Bus property, converted
+                from a string to milliseconds."""
+                set_client_prop(prop,
+                                timedelta_to_milliseconds
+                                (string_to_delta(value)))
             if options.remove:
                 mandos_serv.RemoveClient(client.__dbus_object_path__)
             if options.enable:
-                client.Enable(dbus_interface=client_interface)
+                set_client_prop("Enabled", dbus.Boolean(True))
             if options.disable:
-                client.Disable(dbus_interface=client_interface)
+                set_client_prop("Enabled", dbus.Boolean(False))
             if options.bump_timeout:
-                client.CheckedOK(dbus_interface=client_interface)
+                set_client_prop("LastCheckedOK", "")
             if options.start_checker:
-                client.StartChecker(dbus_interface=client_interface)
+                set_client_prop("CheckerRunning", dbus.Boolean(True))
             if options.stop_checker:
-                client.StopChecker(dbus_interface=client_interface)
+                set_client_prop("CheckerRunning", dbus.Boolean(False))
             if options.is_enabled:
                 sys.exit(0 if client.Get(client_interface,
                                          "Enabled",
@@ -296,48 +426,29 @@
                                          dbus.PROPERTIES_IFACE)
                          else 1)
             if options.checker is not None:
-                client.Set(client_interface, "Checker",
-                           options.checker,
-                           dbus_interface=dbus.PROPERTIES_IFACE)
+                set_client_prop("Checker", options.checker)
             if options.host is not None:
-                client.Set(client_interface, "Host", options.host,
-                           dbus_interface=dbus.PROPERTIES_IFACE)
+                set_client_prop("Host", options.host)
             if options.interval is not None:
-                client.Set(client_interface, "Interval",
-                           timedelta_to_milliseconds
-                           (string_to_delta(options.interval)),
-                           dbus_interface=dbus.PROPERTIES_IFACE)
+                set_client_prop_ms("Interval", options.interval)
             if options.approval_delay is not None:
-                client.Set(client_interface, "ApprovalDelay",
-                           timedelta_to_milliseconds
-                           (string_to_delta(options.
-                                            approval_delay)),
-                           dbus_interface=dbus.PROPERTIES_IFACE)
+                set_client_prop_ms("ApprovalDelay",
+                                   options.approval_delay)
             if options.approval_duration is not None:
-                client.Set(client_interface, "ApprovalDuration",
-                           timedelta_to_milliseconds
-                           (string_to_delta(options.
-                                            approval_duration)),
-                           dbus_interface=dbus.PROPERTIES_IFACE)
+                set_client_prop_ms("ApprovalDuration",
+                                   options.approval_duration)
             if options.timeout is not None:
-                client.Set(client_interface, "Timeout",
-                           timedelta_to_milliseconds
-                           (string_to_delta(options.timeout)),
-                           dbus_interface=dbus.PROPERTIES_IFACE)
+                set_client_prop_ms("Timeout", options.timeout)
             if options.extended_timeout is not None:
-                client.Set(client_interface, "ExtendedTimeout",
-                           timedelta_to_milliseconds
-                           (string_to_delta(options.extended_timeout)),
-                           dbus_interface=dbus.PROPERTIES_IFACE)
+                set_client_prop_ms("ExtendedTimeout",
+                                   options.extended_timeout)
             if options.secret is not None:
-                client.Set(client_interface, "Secret",
-                           dbus.ByteArray(options.secret.read()),
-                           dbus_interface=dbus.PROPERTIES_IFACE)
+                set_client_prop("Secret",
+                                dbus.ByteArray(options.secret.read()))
             if options.approved_by_default is not None:
-                client.Set(client_interface, "ApprovedByDefault",
-                           dbus.Boolean(options
-                                        .approved_by_default),
-                           dbus_interface=dbus.PROPERTIES_IFACE)
+                set_client_prop("ApprovedByDefault",
+                                dbus.Boolean(options
+                                             .approved_by_default))
             if options.approve:
                 client.Approve(dbus.Boolean(True),
                                dbus_interface=client_interface)

=== modified file 'mandos-ctl.xml'
--- mandos-ctl.xml	2011-12-31 23:05:34 +0000
+++ mandos-ctl.xml	2012-06-22 23:33:56 +0000
@@ -2,7 +2,7 @@
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY COMMANDNAME "mandos-ctl">
-<!ENTITY TIMESTAMP "2012-01-01">
+<!ENTITY TIMESTAMP "2012-06-22">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -195,6 +195,10 @@
 	<arg choice="plain"><option>-v</option></arg>
       </group>
     </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice="plain"><option>--check</option></arg>
+    </cmdsynopsis>
   </refsynopsisdiv>
   
   <refsect1 id="description">
@@ -476,6 +480,15 @@
 	</listitem>
       </varlistentry>
       
+      <varlistentry>
+	<term><option>--check</option></term>
+	<listitem>
+	  <para>
+	    Run self-tests.  This includes any unit tests, etc.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
     </variablelist>
   </refsect1>
   

=== modified file 'mandos-keygen'
--- mandos-keygen	2012-01-15 21:07:44 +0000
+++ mandos-keygen	2014-06-22 02:19:30 +0000
@@ -2,8 +2,8 @@
 # 
 # Mandos key generator - create a new OpenPGP key for a Mandos client
 # 
-# Copyright © 2008-2012 Teddy Hogeborn
-# Copyright © 2008-2012 Björn Påhlsson
+# Copyright © 2008-2014 Teddy Hogeborn
+# Copyright © 2008-2014 Björn Påhlsson
 # 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -21,18 +21,19 @@
 # Contact the authors at <mandos@recompile.se>.
 # 
 
-VERSION="1.5.3"
+VERSION="1.6.5"
 
 KEYDIR="/etc/keys/mandos"
-KEYTYPE=DSA
-KEYLENGTH=2048
-SUBKEYTYPE=ELG-E
-SUBKEYLENGTH=2048
+KEYTYPE=RSA
+KEYLENGTH=4096
+SUBKEYTYPE=RSA
+SUBKEYLENGTH=4096
 KEYNAME="`hostname --fqdn 2>/dev/null || hostname`"
 KEYEMAIL=""
-KEYCOMMENT="Mandos client key"
+KEYCOMMENT=""
 KEYEXPIRE=0
 FORCE=no
+SSH=yes
 KEYCOMMENT_ORIG="$KEYCOMMENT"
 mode=keygen
 
@@ -41,8 +42,8 @@
 fi
 
 # Parse options
-TEMP=`getopt --options vhpF:d:t:l:s:L:n:e:c:x:f \
-    --longoptions version,help,password,passfile:,dir:,type:,length:,subtype:,sublength:,name:,email:,comment:,expire:,force \
+TEMP=`getopt --options vhpF:d:t:l:s:L:n:e:c:x:fS \
+    --longoptions version,help,password,passfile:,dir:,type:,length:,subtype:,sublength:,name:,email:,comment:,expire:,force,no-ssh \
     --name "$0" -- "$@"`
 
 help(){
@@ -60,19 +61,18 @@
   -v, --version         Show program's version number and exit
   -h, --help            Show this help message and exit
   -d DIR, --dir DIR     Target directory for key files
-  -t TYPE, --type TYPE  Key type.  Default is DSA.
+  -t TYPE, --type TYPE  Key type.  Default is RSA.
   -l BITS, --length BITS
-                        Key length in bits.  Default is 2048.
+                        Key length in bits.  Default is 4096.
   -s TYPE, --subtype TYPE
-                        Subkey type.  Default is ELG-E.
+                        Subkey type.  Default is RSA.
   -L BITS, --sublength BITS
-                        Subkey length in bits.  Default is 2048.
+                        Subkey length in bits.  Default is 4096.
   -n NAME, --name NAME  Name of key.  Default is the FQDN.
   -e ADDRESS, --email ADDRESS
                         Email address of key.  Default is empty.
   -c TEXT, --comment TEXT
-                        Comment field for key.  The default value is
-                        "Mandos client key".
+                        Comment field for key.  The default is empty.
   -x TIME, --expire TIME
                         Key expire time.  Default is no expiration.
                         See gpg(1) for syntax.
@@ -86,6 +86,7 @@
                         Encrypt a password from FILE using the key in
                         the key directory.  All options other than
                         --dir and --name are ignored.
+  -S, --no-ssh          Don't get SSH key or set "checker" option.
 EOF
 }
 
@@ -104,6 +105,7 @@
 	-c|--comment) KEYCOMMENT="$2"; shift 2;;
 	-x|--expire) KEYEXPIRE="$2"; shift 2;;
 	-f|--force) FORCE=yes; shift;;
+	-S|--no-ssh) SSH=no; shift;;
 	-v|--version) echo "$0 $VERSION"; exit;;
 	-h|--help) help; exit;;
 	--) shift; break;;
@@ -189,7 +191,7 @@
 trap "
 set +e; \
 test -n \"$SECFILE\" && shred --remove \"$SECFILE\"; \
-shred --remove \"$RINGDIR\"/sec*;
+shred --remove \"$RINGDIR\"/sec* 2>/dev/null;
 test -n \"$BATCHFILE\" && rm --force \"$BATCHFILE\"; \
 rm --recursive --force \"$RINGDIR\";
 tty --quiet && stty echo; \
@@ -204,10 +206,10 @@
     cat >"$BATCHFILE" <<-EOF
 	Key-Type: $KEYTYPE
 	Key-Length: $KEYLENGTH
-	#Key-Usage: encrypt,sign,auth
+	Key-Usage: sign,auth
 	Subkey-Type: $SUBKEYTYPE
 	Subkey-Length: $SUBKEYLENGTH
-	#Subkey-Usage: encrypt,sign,auth
+	Subkey-Usage: encrypt
 	Name-Real: $KEYNAME
 	$KEYCOMMENTLINE
 	$KEYEMAILLINE
@@ -229,6 +231,11 @@
 	date
     fi
     
+    # Make sure trustdb.gpg exists;
+    # this is a workaround for Debian bug #737128
+    gpg --quiet --batch --no-tty --no-options --enable-dsa2 \
+	--homedir "$RINGDIR" \
+	--import-ownertrust < /dev/null
     # Generate a new key in the key rings
     gpg --quiet --batch --no-tty --no-options --enable-dsa2 \
 	--homedir "$RINGDIR" --trust-model always \
@@ -270,6 +277,23 @@
 fi
 
 if [ "$mode" = password ]; then
+    
+    # Make SSH be 0 or 1
+    case "$SSH" in
+	[Yy][Ee][Ss]|[Tt][Rr][Uu][Ee]) SSH=1;;
+	[Nn][Oo]|[Ff][Aa][Ll][Ss][Ee]|*) SSH=0;;
+    esac
+    
+    if [ $SSH -eq 1 ]; then
+	set +e
+	ssh_fingerprint="`ssh-keyscan localhost 2>/dev/null`"
+	if [ $? -ne 0 ]; then
+	    ssh_fingerprint=""
+	fi
+	set -e
+	ssh_fingerprint="${ssh_fingerprint#localhost }"
+    fi
+    
     # Import key into temporary key rings
     gpg --quiet --batch --no-tty --no-options --enable-dsa2 \
 	--homedir "$RINGDIR" --trust-model always --armor \
@@ -294,9 +318,11 @@
 	    cat "$PASSFILE"
 	else
 	    tty --quiet && stty -echo
-	    read -p "Enter passphrase: " first
+	    echo -n "Enter passphrase: " >&2
+	    read first
 	    tty --quiet && echo >&2
-	    read -p "Repeat passphrase: " second
+	    echo -n "Repeat passphrase: " >&2
+	    read second
 	    if tty --quiet; then
 		echo >&2
 		stty echo
@@ -336,6 +362,10 @@
 		/^[^-]/s/^/    /p
 	    }
 	}' < "$SECFILE"
+    if [ -n "$ssh_fingerprint" ]; then
+	echo 'checker = ssh-keyscan %%(host)s 2>/dev/null | grep --fixed-strings --line-regexp --quiet --regexp="%%(host)s %(ssh_fingerprint)s"'
+	echo "ssh_fingerprint = ${ssh_fingerprint}"
+    fi
 fi
 
 trap - EXIT
@@ -346,5 +376,5 @@
     shred --remove "$SECFILE"
 fi
 # Remove the key rings
-shred --remove "$RINGDIR"/sec*
+shred --remove "$RINGDIR"/sec* 2>/dev/null
 rm --recursive --force "$RINGDIR"

=== modified file 'mandos-keygen.xml'
--- mandos-keygen.xml	2011-12-31 23:05:34 +0000
+++ mandos-keygen.xml	2014-06-22 02:19:30 +0000
@@ -2,7 +2,7 @@
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY COMMANDNAME "mandos-keygen">
-<!ENTITY TIMESTAMP "2012-01-01">
+<!ENTITY TIMESTAMP "2014-06-22">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -119,7 +119,10 @@
 	<replaceable>TIME</replaceable></option></arg>
       </group>
       <sbr/>
-      <arg><option>--force</option></arg>
+      <group>
+	<arg choice="plain"><option>--force</option></arg>
+	<arg choice="plain"><option>-f</option></arg>
+      </group>
     </cmdsynopsis>
     <cmdsynopsis>
       <command>&COMMANDNAME;</command>
@@ -145,6 +148,10 @@
 	<arg choice="plain"><option>-n
 	<replaceable>NAME</replaceable></option></arg>
       </group>
+      <group>
+	<arg choice="plain"><option>--no-ssh</option></arg>
+	<arg choice="plain"><option>-S</option></arg>
+      </group>
     </cmdsynopsis>
     <cmdsynopsis>
       <command>&COMMANDNAME;</command>
@@ -227,7 +234,7 @@
 	<replaceable>TYPE</replaceable></option></term>
 	<listitem>
 	  <para>
-	    Key type.  Default is <quote>DSA</quote>.
+	    Key type.  Default is <quote>RSA</quote>.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -239,7 +246,7 @@
 	<replaceable>BITS</replaceable></option></term>
 	<listitem>
 	  <para>
-	    Key length in bits.  Default is 2048.
+	    Key length in bits.  Default is 4096.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -251,7 +258,7 @@
 	<replaceable>KEYTYPE</replaceable></option></term>
 	<listitem>
 	  <para>
-	    Subkey type.  Default is <quote>ELG-E</quote> (Elgamal
+	    Subkey type.  Default is <quote>RSA</quote> (Elgamal
 	    encryption-only).
 	  </para>
 	</listitem>
@@ -264,7 +271,7 @@
 	<replaceable>BITS</replaceable></option></term>
 	<listitem>
 	  <para>
-	    Subkey length in bits.  Default is 2048.
+	    Subkey length in bits.  Default is 4096.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -288,8 +295,7 @@
 	<replaceable>TEXT</replaceable></option></term>
 	<listitem>
 	  <para>
-	    Comment field for key.  The default value is
-	    <quote><literal>Mandos client key</literal></quote>.
+	    Comment field for key.  Default is empty.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -347,6 +353,22 @@
 	  </para>
 	</listitem>
       </varlistentry>
+      <varlistentry>
+	<term><option>--no-ssh</option></term>
+	<term><option>-S</option></term>
+	<listitem>
+	  <para>
+	    When <option>--password</option> or
+	    <option>--passfile</option> is given, this option will
+	    prevent <command>&COMMANDNAME;</command> from calling
+	    <command>ssh-keyscan</command> to get an SSH fingerprint
+	    for this host and, if successful, output suitable config
+	    options to use this fingerprint as a
+	    <option>checker</option> option in the output.  This is
+	    otherwise the default behavior.
+	  </para>
+	</listitem>
+      </varlistentry>
     </variablelist>
   </refsect1>
   
@@ -503,7 +525,9 @@
       <citerefentry><refentrytitle>mandos</refentrytitle>
       <manvolnum>8</manvolnum></citerefentry>,
       <citerefentry><refentrytitle>mandos-client</refentrytitle>
-      <manvolnum>8mandos</manvolnum></citerefentry>
+      <manvolnum>8mandos</manvolnum></citerefentry>,
+      <citerefentry><refentrytitle>ssh-keyscan</refentrytitle>
+      <manvolnum>1</manvolnum></citerefentry>
     </para>
   </refsect1>
   

=== modified file 'mandos-monitor'
--- mandos-monitor	2012-01-15 21:07:44 +0000
+++ mandos-monitor	2014-05-11 20:18:48 +0000
@@ -3,8 +3,8 @@
 # 
 # Mandos Monitor - Control and monitor the Mandos server
 # 
-# Copyright © 2009-2012 Teddy Hogeborn
-# Copyright © 2009-2012 Björn Påhlsson
+# Copyright © 2009-2014 Teddy Hogeborn
+# Copyright © 2009-2014 Björn Påhlsson
 # 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -17,17 +17,21 @@
 #     GNU General Public License for more details.
 # 
 # You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# along with this program.  If not, see
+# <http://www.gnu.org/licenses/>.
 # 
 # Contact the authors at <mandos@recompile.se>.
 # 
 
 from __future__ import (division, absolute_import, print_function,
                         unicode_literals)
+try:
+    from future_builtins import *
+except ImportError:
+    pass
 
 import sys
 import os
-import signal
 
 import datetime
 
@@ -35,14 +39,18 @@
 import urwid
 
 from dbus.mainloop.glib import DBusGMainLoop
-import gobject
+try:
+    import gobject
+except ImportError:
+    from gi.repository import GObject as gobject
 
 import dbus
 
-import UserList
-
 import locale
 
+if sys.version_info[0] == 2:
+    str = unicode
+
 locale.setlocale(locale.LC_ALL, '')
 
 import logging
@@ -52,15 +60,7 @@
 domain = 'se.recompile'
 server_interface = domain + '.Mandos'
 client_interface = domain + '.Mandos.Client'
-version = "1.5.3"
-
-# Always run in monochrome mode
-urwid.curses_display.curses.has_colors = lambda : False
-
-# Urwid doesn't support blinking, but we want it.  Since we have no
-# use for underline on its own, we make underline also always blink.
-urwid.curses_display.curses.A_UNDERLINE |= (
-    urwid.curses_display.curses.A_BLINK)
+version = "1.6.5"
 
 def isoformat_to_datetime(iso):
     "Parse an ISO 8601 date string to a datetime.datetime()"
@@ -83,23 +83,26 @@
     properties and calls a hook function when any of them are
     changed.
     """
-    def __init__(self, proxy_object=None, *args, **kwargs):
+    def __init__(self, proxy_object=None, properties=None, **kwargs):
         self.proxy = proxy_object # Mandos Client proxy object
-        
-        self.properties = dict()
+        self.properties = dict() if properties is None else properties
         self.property_changed_match = (
             self.proxy.connect_to_signal("PropertyChanged",
-                                         self.property_changed,
+                                         self._property_changed,
                                          client_interface,
                                          byte_arrays=True))
         
-        self.properties.update(
-            self.proxy.GetAll(client_interface,
-                              dbus_interface = dbus.PROPERTIES_IFACE))
-
-        #XXX This breaks good super behaviour
-#        super(MandosClientPropertyCache, self).__init__(
-#            *args, **kwargs)
+        if properties is None:
+            self.properties.update(
+                self.proxy.GetAll(client_interface,
+                                  dbus_interface
+                                  = dbus.PROPERTIES_IFACE))
+        
+        super(MandosClientPropertyCache, self).__init__(**kwargs)
+    
+    def _property_changed(self, property, value):
+        """Helper which takes positional arguments"""
+        return self.property_changed(property=property, value=value)
     
     def property_changed(self, property=None, value=None):
         """This is called whenever we get a PropertyChanged signal
@@ -108,10 +111,8 @@
         # Update properties dict with new value
         self.properties[property] = value
     
-    def delete(self, *args, **kwargs):
+    def delete(self):
         self.property_changed_match.remove()
-        super(MandosClientPropertyCache, self).__init__(
-            *args, **kwargs)
 
 
 class MandosClientWidget(urwid.FlowWidget, MandosClientPropertyCache):
@@ -119,7 +120,7 @@
     """
     
     def __init__(self, server_proxy_object=None, update_hook=None,
-                 delete_hook=None, logger=None, *args, **kwargs):
+                 delete_hook=None, logger=None, **kwargs):
         # Called on update
         self.update_hook = update_hook
         # Called on delete
@@ -130,27 +131,15 @@
         self.logger = logger
         
         self._update_timer_callback_tag = None
-        self._update_timer_callback_lock = 0
         
         # The widget shown normally
         self._text_widget = urwid.Text("")
         # The widget shown when we have focus
         self._focus_text_widget = urwid.Text("")
-        super(MandosClientWidget, self).__init__(
-            update_hook=update_hook, delete_hook=delete_hook,
-            *args, **kwargs)
+        super(MandosClientWidget, self).__init__(**kwargs)
         self.update()
         self.opened = False
         
-        last_checked_ok = isoformat_to_datetime(self.properties
-                                                ["LastCheckedOK"])
-        
-        if self.properties ["LastCheckerStatus"] != 0:
-            self.using_timer(True)
-        
-        if self.need_approval:
-            self.using_timer(True)
-        
         self.match_objects = (
             self.proxy.connect_to_signal("CheckerCompleted",
                                          self.checker_completed,
@@ -172,34 +161,19 @@
                                          self.rejected,
                                          client_interface,
                                          byte_arrays=True))
-        #self.logger('Created client %s' % (self.properties["Name"]))
-    
-    def property_changed(self, property=None, value=None):
-        super(self, MandosClientWidget).property_changed(property,
-                                                         value)
-        if property == "ApprovalPending":
-            using_timer(bool(value))
-        if property == "LastCheckerStatus":
-            using_timer(value != 0)
-            #self.logger('Checker for client %s (command "%s")'
-            #            ' was successful'
-            #            % (self.properties["Name"], command))
+        #self.logger('Created client {0}'
+        #            .format(self.properties["Name"]))
     
     def using_timer(self, flag):
         """Call this method with True or False when timer should be
         activated or deactivated.
         """
-        old = self._update_timer_callback_lock
-        if flag:
-            self._update_timer_callback_lock += 1
-        else:
-            self._update_timer_callback_lock -= 1
-        if old == 0 and self._update_timer_callback_lock:
+        if flag and self._update_timer_callback_tag is None:
             # Will update the shown timer value every second
             self._update_timer_callback_tag = (gobject.timeout_add
                                                (1000,
                                                 self.update_timer))
-        elif old and self._update_timer_callback_lock == 0:
+        elif not (flag or self._update_timer_callback_tag is None):
             gobject.source_remove(self._update_timer_callback_tag)
             self._update_timer_callback_tag = None
     
@@ -209,47 +183,48 @@
             return
         # Checker failed
         if os.WIFEXITED(condition):
-            self.logger('Checker for client %s (command "%s")'
-                        ' failed with exit code %s'
-                        % (self.properties["Name"], command,
-                           os.WEXITSTATUS(condition)))
+            self.logger('Checker for client {0} (command "{1}")'
+                        ' failed with exit code {2}'
+                        .format(self.properties["Name"], command,
+                                os.WEXITSTATUS(condition)))
         elif os.WIFSIGNALED(condition):
-            self.logger('Checker for client %s (command "%s")'
-                        ' was killed by signal %s'
-                        % (self.properties["Name"], command,
-                           os.WTERMSIG(condition)))
+            self.logger('Checker for client {0} (command "{1}") was'
+                        ' killed by signal {2}'
+                        .format(self.properties["Name"], command,
+                                os.WTERMSIG(condition)))
         elif os.WCOREDUMP(condition):
-            self.logger('Checker for client %s (command "%s")'
+            self.logger('Checker for client {0} (command "{1}")'
                         ' dumped core'
-                        % (self.properties["Name"], command))
+                        .format(self.properties["Name"], command))
         else:
-            self.logger('Checker for client %s completed'
-                        ' mysteriously')
+            self.logger('Checker for client {0} completed'
+                        ' mysteriously'
+                        .format(self.properties["Name"]))
         self.update()
     
     def checker_started(self, command):
         """Server signals that a checker started. This could be useful
            to log in the future. """
-        #self.logger('Client %s started checker "%s"'
-        #            % (self.properties["Name"], unicode(command)))
+        #self.logger('Client {0} started checker "{1}"'
+        #            .format(self.properties["Name"],
+        #                    str(command)))
         pass
     
     def got_secret(self):
-        self.logger('Client %s received its secret'
-                    % self.properties["Name"])
+        self.logger('Client {0} received its secret'
+                    .format(self.properties["Name"]))
     
     def need_approval(self, timeout, default):
         if not default:
-            message = 'Client %s needs approval within %s seconds'
+            message = 'Client {0} needs approval within {1} seconds'
         else:
-            message = 'Client %s will get its secret in %s seconds'
-        self.logger(message
-                    % (self.properties["Name"], timeout/1000))
-        self.using_timer(True)
+            message = 'Client {0} will get its secret in {1} seconds'
+        self.logger(message.format(self.properties["Name"],
+                                   timeout/1000))
     
     def rejected(self, reason):
-        self.logger('Client %s was rejected; reason: %s'
-                    % (self.properties["Name"], reason))
+        self.logger('Client {0} was rejected; reason: {1}'
+                    .format(self.properties["Name"], reason))
     
     def selectable(self):
         """Make this a "selectable" widget.
@@ -277,14 +252,14 @@
                           "bold-underline-blink":
                               "bold-underline-blink-standout",
                           }
-
+        
         # Rebuild focus and non-focus widgets using current properties
-
+        
         # Base part of a client. Name!
-        base = ('%(name)s: '
-                      % {"name": self.properties["Name"]})
+        base = '{name}: '.format(name=self.properties["Name"])
         if not self.properties["Enabled"]:
             message = "DISABLED"
+            self.using_timer(False)
         elif self.properties["ApprovalPending"]:
             timeout = datetime.timedelta(milliseconds
                                          = self.properties
@@ -292,31 +267,36 @@
             last_approval_request = isoformat_to_datetime(
                 self.properties["LastApprovalRequest"])
             if last_approval_request is not None:
-                timer = timeout - (datetime.datetime.utcnow()
-                                   - last_approval_request)
+                timer = max(timeout - (datetime.datetime.utcnow()
+                                       - last_approval_request),
+                            datetime.timedelta())
             else:
                 timer = datetime.timedelta()
             if self.properties["ApprovedByDefault"]:
-                message = "Approval in %s. (d)eny?"
+                message = "Approval in {0}. (d)eny?"
             else:
-                message = "Denial in %s. (a)pprove?"
-            message = message % unicode(timer).rsplit(".", 1)[0]
+                message = "Denial in {0}. (a)pprove?"
+            message = message.format(str(timer).rsplit(".", 1)[0])
+            self.using_timer(True)
         elif self.properties["LastCheckerStatus"] != 0:
-            # When checker has failed, print a timer until client expires
+            # When checker has failed, show timer until client expires
             expires = self.properties["Expires"]
             if expires == "":
                 timer = datetime.timedelta(0)
             else:
-                expires = datetime.datetime.strptime(expires,
-                                                     '%Y-%m-%dT%H:%M:%S.%f')
-                timer = expires - datetime.datetime.utcnow()
+                expires = (datetime.datetime.strptime
+                           (expires, '%Y-%m-%dT%H:%M:%S.%f'))
+                timer = max(expires - datetime.datetime.utcnow(),
+                            datetime.timedelta())
             message = ('A checker has failed! Time until client'
-                       ' gets disabled: %s'
-                           % unicode(timer).rsplit(".", 1)[0])
+                       ' gets disabled: {0}'
+                       .format(str(timer).rsplit(".", 1)[0]))
+            self.using_timer(True)
         else:
             message = "enabled"
-        self._text = "%s%s" % (base, message)
-            
+            self.using_timer(False)
+        self._text = "{0}{1}".format(base, message)
+        
         if not urwid.supports_unicode():
             self._text = self._text.encode("ascii", "replace")
         textlist = [("normal", self._text)]
@@ -339,7 +319,7 @@
         self.update()
         return True             # Keep calling this
     
-    def delete(self, *args, **kwargs):
+    def delete(self, **kwargs):
         if self._update_timer_callback_tag is not None:
             gobject.source_remove(self._update_timer_callback_tag)
             self._update_timer_callback_tag = None
@@ -348,7 +328,7 @@
         self.match_objects = ()
         if self.delete_hook is not None:
             self.delete_hook(self)
-        return super(MandosClientWidget, self).delete(*args, **kwargs)
+        return super(MandosClientWidget, self).delete(**kwargs)
     
     def render(self, maxcolrow, focus=False):
         """Render differently if we have focus.
@@ -396,14 +376,13 @@
         else:
             return key
     
-    def property_changed(self, property=None, value=None,
-                         *args, **kwargs):
+    def property_changed(self, property=None, **kwargs):
         """Call self.update() if old value is not new value.
         This overrides the method from MandosClientPropertyCache"""
-        property_name = unicode(property)
+        property_name = str(property)
         old_value = self.properties.get(property_name)
         super(MandosClientWidget, self).property_changed(
-            property=property, value=value, *args, **kwargs)
+            property=property, **kwargs)
         if self.properties.get(property_name) != old_value:
             self.update()
 
@@ -413,8 +392,8 @@
     "down" key presses, thus not allowing any containing widgets to
     use them as an excuse to shift focus away from this widget.
     """
-    def keypress(self, maxcolrow, key):
-        ret = super(ConstrainedListBox, self).keypress(maxcolrow, key)
+    def keypress(self, *args, **kwargs):
+        ret = super(ConstrainedListBox, self).keypress(*args, **kwargs)
         if ret in ("up", "down"):
             return
         return ret
@@ -433,20 +412,21 @@
                 ("normal",
                  "default", "default", None),
                 ("bold",
-                 "default", "default", "bold"),
+                 "bold", "default", "bold"),
                 ("underline-blink",
-                 "default", "default", "underline"),
+                 "underline,blink", "default", "underline,blink"),
                 ("standout",
-                 "default", "default", "standout"),
+                 "standout", "default", "standout"),
                 ("bold-underline-blink",
-                 "default", "default", ("bold", "underline")),
+                 "bold,underline,blink", "default", "bold,underline,blink"),
                 ("bold-standout",
-                 "default", "default", ("bold", "standout")),
+                 "bold,standout", "default", "bold,standout"),
                 ("underline-blink-standout",
-                 "default", "default", ("underline", "standout")),
+                 "underline,blink,standout", "default",
+                 "underline,blink,standout"),
                 ("bold-underline-blink-standout",
-                 "default", "default", ("bold", "underline",
-                                          "standout")),
+                 "bold,underline,blink,standout", "default",
+                 "bold,underline,blink,standout"),
                 ))
         
         if urwid.supports_unicode():
@@ -486,9 +466,9 @@
         self.main_loop = gobject.MainLoop()
     
     def client_not_found(self, fingerprint, address):
-        self.log_message(("Client with address %s and fingerprint %s"
-                          " could not be found" % (address,
-                                                    fingerprint)))
+        self.log_message("Client with address {0} and fingerprint"
+                         " {1} could not be found"
+                         .format(address, fingerprint))
     
     def rebuild(self):
         """This rebuilds the User Interface.
@@ -507,6 +487,7 @@
         self.topwidget = urwid.Pile(self.uilist)
     
     def log_message(self, message):
+        """Log message formatted with timestamp"""
         timestamp = datetime.datetime.now().isoformat()
         self.log_message_raw(timestamp + ": " + message)
     
@@ -525,7 +506,7 @@
         self.log_visible = not self.log_visible
         self.rebuild()
         #self.log_message("Log visibility changed to: "
-        #                 + unicode(self.log_visible))
+        #                 + str(self.log_visible))
     
     def change_log_display(self):
         """Change type of log display.
@@ -547,8 +528,8 @@
             client = self.clients_dict[path]
         except KeyError:
             # not found?
-            self.log_message("Unknown client %r (%r) removed", name,
-                             path)
+            self.log_message("Unknown client {0!r} ({1!r}) removed"
+                             .format(name, path))
             return
         client.delete()
     
@@ -571,7 +552,7 @@
         if path is None:
             path = client.proxy.object_path
         self.clients_dict[path] = client
-        self.clients.sort(None, lambda c: c.properties["Name"])
+        self.clients.sort(key=lambda c: c.properties["Name"])
         self.refresh()
     
     def remove_client(self, client, path=None):
@@ -579,11 +560,6 @@
         if path is None:
             path = client.proxy.object_path
         del self.clients_dict[path]
-        if not self.clients_dict:
-            # Work around bug in Urwid 0.9.8.3 - if a SimpleListWalker
-            # is completely emptied, we need to recreate it.
-            self.clients = urwid.SimpleListWalker([])
-            self.rebuild()
         self.refresh()
     
     def refresh(self):
@@ -602,7 +578,10 @@
         try:
             mandos_clients = (self.mandos_serv
                               .GetAllClientsWithProperties())
+            if not mandos_clients:
+                self.log_message_raw(("bold", "Note: Server has no clients."))
         except dbus.exceptions.DBusException:
+            self.log_message_raw(("bold", "Note: No Mandos server running."))
             mandos_clients = dbus.Dictionary()
         
         (self.mandos_serv
@@ -620,7 +599,7 @@
                             self.client_not_found,
                             dbus_interface=server_interface,
                             byte_arrays=True))
-        for path, client in mandos_clients.iteritems():
+        for path, client in mandos_clients.items():
             client_proxy_object = self.bus.get_object(self.busname,
                                                       path)
             self.add_client(MandosClientWidget(server_proxy_object
@@ -635,7 +614,7 @@
                                                logger
                                                =self.log_message),
                             path=path)
-
+        
         self.refresh()
         self._input_callback_tag = (gobject.io_add_watch
                                     (sys.stdin.fileno(),
@@ -738,7 +717,7 @@
     ui.run()
 except KeyboardInterrupt:
     ui.screen.stop()
-except Exception, e:
-    ui.log_message(unicode(e))
+except Exception as e:
+    ui.log_message(str(e))
     ui.screen.stop()
     raise

=== modified file 'mandos-options.xml'
--- mandos-options.xml	2012-01-01 04:02:00 +0000
+++ mandos-options.xml	2014-06-15 02:48:49 +0000
@@ -46,11 +46,24 @@
     <emphasis>not</emphasis> run in debug mode.
   </para>
   
+  <para id="priority_compat">
+    GnuTLS priority string for the <acronym>TLS</acronym> handshake.
+    The default is <quote><literal
+    >SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP:+SIGN-RSA-SHA224:</literal>
+    <literal>+SIGN-RSA-RMD160</literal></quote>.
+    See <citerefentry><refentrytitle
+    >gnutls_priority_init</refentrytitle>
+    <manvolnum>3</manvolnum></citerefentry> for the syntax.
+    <emphasis>Warning</emphasis>: changing this may make the
+    <acronym>TLS</acronym> handshake fail, making server-client
+    communication impossible.
+  </para>
+  
   <para id="priority">
     GnuTLS priority string for the <acronym>TLS</acronym> handshake.
     The default is <quote><literal
     >SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP</literal></quote>.  See
-    <citerefentry><refentrytitle>gnutls_priority_init</refentrytitle>
+    <citerefentry><refentrytitle >gnutls_priority_init</refentrytitle>
     <manvolnum>3</manvolnum></citerefentry> for the syntax.
     <emphasis>Warning</emphasis>: changing this may make the
     <acronym>TLS</acronym> handshake fail, making server-client
@@ -97,4 +110,24 @@
     class="directory">/var/lib/mandos</filename></quote>.
   </para>
   
+  <para id="socket">
+    If this option is used, the server will not create a new network
+    socket, but will instead use the supplied file descriptor.  By
+    default, the server will create a new network socket.
+  </para>
+  
+  <para id="foreground">
+    This option will make the server run in the foreground and not
+    write a PID file.  The default is to <emphasis>not</emphasis> run
+    in the foreground, except in <option>debug</option> mode, which
+    implies this option.
+  </para>
+  
+  <para id="zeroconf">
+    This option controls whether the server will announce its
+    existence using Zeroconf.  Default is to use Zeroconf.  If
+    Zeroconf is not used, a <option>port</option> number or a
+    <option>socket</option> is required.
+  </para>
+  
 </section>

=== modified file 'mandos.conf'
--- mandos.conf	2011-11-26 22:22:20 +0000
+++ mandos.conf	2014-06-15 02:48:49 +0000
@@ -23,7 +23,7 @@
 ;debug = False
 
 # GnuTLS priority for the TLS handshake.  See gnutls_priority_init(3).
-;priority = SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP
+;priority = SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP:+SIGN-RSA-SHA224:+SIGN-RSA-RMD160
 
 # Zeroconf service name.  You need to change this if you for some
 # reason want to run more than one server on the same *host*.
@@ -42,3 +42,12 @@
 
 # The directory where state is saved
 ;statedir = /var/lib/mandos
+
+# Whether to run in the foreground
+;foreground = False
+
+# File descriptor number to use for network socket
+;socket = 
+
+# Whether to use ZeroConf; if false, requires port or socket
+;zeroconf = True

=== modified file 'mandos.conf.xml'
--- mandos.conf.xml	2011-12-31 23:05:34 +0000
+++ mandos.conf.xml	2013-10-24 20:21:45 +0000
@@ -3,7 +3,7 @@
 	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY CONFNAME "mandos.conf">
 <!ENTITY CONFPATH "<filename>/etc/mandos/mandos.conf</filename>">
-<!ENTITY TIMESTAMP "2012-01-01">
+<!ENTITY TIMESTAMP "2013-10-23">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -36,6 +36,7 @@
       <year>2009</year>
       <year>2011</year>
       <year>2012</year>
+      <year>2013</year>
       <holder>Teddy Hogeborn</holder>
       <holder>Björn Påhlsson</holder>
     </copyright>
@@ -120,7 +121,8 @@
 	<term><option>priority<literal> = </literal><replaceable
 	>STRING</replaceable></option></term>
 	<listitem>
-	  <xi:include href="mandos-options.xml" xpointer="priority"/>
+	  <xi:include href="mandos-options.xml"
+                      xpointer="priority_compat"/>
 	</listitem>
       </varlistentry>
       
@@ -174,6 +176,14 @@
 	</listitem>
       </varlistentry>
       
+      <varlistentry>
+	<term><option>socket<literal> = </literal><replaceable
+	>NUMBER</replaceable></option></term>
+	<listitem>
+	  <xi:include href="mandos-options.xml" xpointer="socket"/>
+	</listitem>
+      </varlistentry>
+      
     </variablelist>
   </refsect1>
   

=== modified file 'mandos.lsm'
--- mandos.lsm	2012-01-15 21:07:44 +0000
+++ mandos.lsm	2014-05-11 20:18:48 +0000
@@ -1,22 +1,23 @@
 Begin4
 Title:		Mandos
-Version:	1.5.3
-Entered-date:	2012-01-15
+Version:	1.6.5
+Entered-date:	2014-05-11
 Description:	The Mandos system allows computers to have encrypted
-root file systems and at the same time be capable of remote and/or
-unattended reboots.
+		root file systems and at the same time be capable of
+		remote and/or unattended reboots.
 Keywords:	boot, encryption, luks, cryptsetup, network, openpgp,
-tls, dm-crypt
+		tls, dm-crypt
 Author:		teddy@recompile.se (Teddy Hogeborn),
 		belorn@recompile.se (Björn Påhlsson)
 Maintained-by:	teddy@recompile.se (Teddy Hogeborn),
 		belorn@recompile.se (Björn Påhlsson)
 Primary-site: http://www.recompile.se/mandos
-	      147K mandos_1.5.3.orig.tar.gz
+	      159K mandos_1.6.5.orig.tar.gz
 Alternate-site: ftp://ftp.recompile.se/pub/mandos
-		147K mandos_1.5.3.orig.tar.gz
+		159K mandos_1.6.5.orig.tar.gz
 Platforms:	Requires GCC, GNU libC, Avahi, GnuPG, Python 2.6, and
-various other libraries.  While made for Debian GNU/Linux, it is
-probably portable to other distributions, but not other Unixes.
+		various other libraries.  While made for Debian
+		GNU/Linux, it is probably portable to other
+		distributions, but not other Unixes. 
 Copying-policy: GNU General Public License version 3.0 or later
 End

=== added file 'mandos.service'
--- mandos.service	1970-01-01 00:00:00 +0000
+++ mandos.service	2013-12-31 16:02:18 +0000
@@ -0,0 +1,21 @@
+[Unit]
+Description=Server of encrypted passwords to Mandos clients
+
+[Service]
+Type=simple
+## Type=dbus is not appropriate, because Mandos also needs to announce
+## its ZeroConf service and be reachable on the network.
+#Type=dbus
+BusName=se.recompile.Mandos
+# If you add --no-dbus, also comment out BusName above, and vice versa
+ExecStart=/usr/sbin/mandos --foreground
+Restart=always
+KillMode=process
+## Using socket activation won't work either, because systemd always
+## does bind() on the socket, and also won't announce the ZeroConf
+## service.
+#ExecStart=/usr/sbin/mandos --foreground --socket=0
+#StandardInput=socket
+
+[Install]
+WantedBy=multi-user.target

=== modified file 'mandos.xml'
--- mandos.xml	2012-01-15 21:01:13 +0000
+++ mandos.xml	2014-06-15 02:48:49 +0000
@@ -2,7 +2,7 @@
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 "http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY COMMANDNAME "mandos">
-<!ENTITY TIMESTAMP "2012-01-15">
+<!ENTITY TIMESTAMP "2014-06-15">
 <!ENTITY % common SYSTEM "common.ent">
 %common;
 ]>
@@ -36,6 +36,7 @@
       <year>2010</year>
       <year>2011</year>
       <year>2012</year>
+      <year>2013</year>
       <holder>Teddy Hogeborn</holder>
       <holder>Björn Påhlsson</holder>
     </copyright>
@@ -100,6 +101,13 @@
       <sbr/>
       <arg><option>--statedir
       <replaceable>DIRECTORY</replaceable></option></arg>
+      <sbr/>
+      <arg><option>--socket
+      <replaceable>FD</replaceable></option></arg>
+      <sbr/>
+      <arg><option>--foreground</option></arg>
+      <sbr/>
+      <arg><option>--no-zeroconf</option></arg>
     </cmdsynopsis>
     <cmdsynopsis>
       <command>&COMMANDNAME;</command>
@@ -228,7 +236,8 @@
 	<term><option>--priority <replaceable>
 	PRIORITY</replaceable></option></term>
 	<listitem>
-	  <xi:include href="mandos-options.xml" xpointer="priority"/>
+	  <xi:include href="mandos-options.xml"
+                      xpointer="priority_compat"/>
 	</listitem>
       </varlistentry>
       
@@ -299,6 +308,30 @@
 	  <xi:include href="mandos-options.xml" xpointer="statedir"/>
 	</listitem>
       </varlistentry>
+      
+      <varlistentry>
+	<term><option>--socket
+	<replaceable>FD</replaceable></option></term>
+	<listitem>
+	  <xi:include href="mandos-options.xml" xpointer="socket"/>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--foreground</option></term>
+	<listitem>
+	  <xi:include href="mandos-options.xml"
+		      xpointer="foreground"/>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><option>--no-zeroconf</option></term>
+	<listitem>
+	  <xi:include href="mandos-options.xml" xpointer="zeroconf"/>
+	</listitem>
+      </varlistentry>
+      
     </variablelist>
   </refsect1>
   
@@ -494,11 +527,15 @@
 	</listitem>
       </varlistentry>
       <varlistentry>
-	<term><filename>/var/run/mandos.pid</filename></term>
+	<term><filename>/run/mandos.pid</filename></term>
 	<listitem>
 	  <para>
 	    The file containing the process id of the
 	    <command>&COMMANDNAME;</command> process started last.
+	    <emphasis >Note:</emphasis> If the <filename
+	    class="directory">/run</filename> directory does not
+	    exist, <filename>/var/run/mandos.pid</filename> will be
+	    used instead.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -549,9 +586,6 @@
       There is no fine-grained control over logging and debug output.
     </para>
     <para>
-      Debug mode is conflated with running in the foreground.
-    </para>
-    <para>
       This server does not check the expire time of clients’ OpenPGP
       keys.
     </para>

=== modified file 'network-hooks.d/bridge'
--- network-hooks.d/bridge	2012-01-15 16:10:09 +0000
+++ network-hooks.d/bridge	2012-06-13 22:06:57 +0000
@@ -6,8 +6,8 @@
 # configuration file(s) should be copied into the
 # /etc/mandos/network-hooks.d directory.
 # 
-# Copyright © 2011 Teddy Hogeborn
-# Copyright © 2011 Björn Påhlsson
+# Copyright © 2012 Teddy Hogeborn
+# Copyright © 2012 Björn Påhlsson
 # 
 # Copying and distribution of this file, with or without modification,
 # are permitted in any medium without royalty provided the copyright
@@ -33,8 +33,11 @@
     exit
 fi
 
-if [ -n "$DEVICE" -a "$DEVICE" != "$BRIDGE" ]; then
-    exit
+if [ -n "$DEVICE" ]; then
+    case "$DEVICE" in
+	*,"$BRIDGE"|*,"$BRIDGE",*|"$BRIDGE",*|"$BRIDGE") :;;
+	*) exit;;
+    esac
 fi
 
 brctl="/sbin/brctl"
@@ -45,35 +48,40 @@
     fi
 done
 
-case "$1" in
-    start)
-	"$brctl" addbr "$BRIDGE"
-	for address in $PORT_ADDRESSES; do
-	    interface=`addrtoif "$address"`
-	    "$brctl" addif "$BRIDGE" "$interface"
-	    ip link set dev "$interface" up
-	done
-	ip link set dev "$BRIDGE" up
-	sleep "${DELAY%%.*}"
-	if [ -n "$IPADDRS" ]; then
-            for ipaddr in $IPADDRS; do
-		ip addr add "$ipaddr" dev "$BRIDGE"
-	    done
-	fi
-	if [ -n "$ROUTES" ]; then
-            for route in $ROUTES; do
-		ip route add "$route" dev "$BRIDGE"
-	    done
-	fi
-	;;
-    stop)
-	ip link set dev "$BRIDGE" down
-	for address in $PORT_ADDRESSES; do
-	    interface=`addrtoif "$address"`
-	    ip link set dev "$interface" down
-	    "$brctl" delif "$BRIDGE" "$interface"
-	done
-	"$brctl" delbr "$BRIDGE"
+do_start(){
+    "$brctl" addbr "$BRIDGE"
+    for address in $PORT_ADDRESSES; do
+	interface=`addrtoif "$address"`
+	"$brctl" addif "$BRIDGE" "$interface"
+	ip link set dev "$interface" up
+    done
+    ip link set dev "$BRIDGE" up
+    sleep "${DELAY%%.*}"
+    if [ -n "$IPADDRS" ]; then
+        for ipaddr in $IPADDRS; do
+	    ip addr add "$ipaddr" dev "$BRIDGE"
+	done
+    fi
+    if [ -n "$ROUTES" ]; then
+        for route in $ROUTES; do
+	    ip route add "$route" dev "$BRIDGE"
+	done
+    fi
+}
+
+do_stop(){
+    ip link set dev "$BRIDGE" down
+    for address in $PORT_ADDRESSES; do
+	interface=`addrtoif "$address"`
+	ip link set dev "$interface" down
+	"$brctl" delif "$BRIDGE" "$interface"
+    done
+    "$brctl" delbr "$BRIDGE"
+}
+
+case "${MODE:-$1}" in
+    start|stop)
+	do_"${MODE:-$1}"
 	;;
     files)
 	echo /bin/ip

=== modified file 'network-hooks.d/openvpn'
--- network-hooks.d/openvpn	2012-01-01 17:38:33 +0000
+++ network-hooks.d/openvpn	2012-06-13 22:06:57 +0000
@@ -31,24 +31,31 @@
 fi
 
 # Exit if DEVICE is set and it doesn't match the VPN interface
-if [ -n "$DEVICE" -a "$DEVICE" = "${DEVICE#$VPNDEVICE}" ]; then
-    exit
+if [ -n "$DEVICE" ]; then
+    case "$DEVICE" in
+	*,"$VPNDEVICE"*|"$VPNDEVICE"*) :;;
+	*) exit;;
+    esac
 fi
 
 openvpn=/usr/sbin/openvpn
 
-case "$1" in
-    start)
-	"$openvpn" --cd "$MANDOSNETHOOKDIR" \
-	    --daemon 'openvpn(Mandos)' --writepid "$PIDFILE" \
-	    --config "$CONFIG"
-	sleep "$DELAY"
-	;;
-    stop)
-	PID="`cat \"$PIDFILE\"`"
-	if [ "$PID" -gt 0 ]; then
-	    kill "$PID"
-	fi
+do_start(){
+    "$openvpn" --cd "$MANDOSNETHOOKDIR" --daemon 'openvpn(Mandos)' \
+	--writepid "$PIDFILE" --config "$CONFIG"
+     sleep "$DELAY"
+}
+
+do_stop(){
+    PID="`cat \"$PIDFILE\"`"
+    if [ "$PID" -gt 0 ]; then
+	kill "$PID"
+    fi
+}
+
+case "${MODE:-$1}" in
+    start|stop)
+	do_"${MODE:-$1}"
 	;;
     files)
 	echo "$openvpn"

=== modified file 'network-hooks.d/wireless'
--- network-hooks.d/wireless	2012-01-01 17:38:33 +0000
+++ network-hooks.d/wireless	2012-06-13 22:06:57 +0000
@@ -35,8 +35,7 @@
     exit
 fi
 
-ifkeys=`env | sed -n -e 's/^ADDRESS_\([^=]*\)=.*/\1/p' "$CONFIG" \
-    | sort -u`
+ifkeys=`sed -n -e 's/^ADDRESS_\([^=]*\)=.*/\1/p' "$CONFIG" | sort -u`
 
 # Exit if DEVICE is set and is not any of the wireless interfaces
 if [ -n "$DEVICE" ]; then
@@ -44,9 +43,11 @@
 	for KEY in $ifkeys; do
 	    ADDRESS=`eval 'echo "$ADDRESS_'"$KEY"\"`
 	    INTERFACE=`addrtoif "$ADDRESS"`
-	    if [ "$INTERFACE" = "$DEVICE" ]; then
-		break 2
-	    fi
+	    
+	    case "$DEVICE" in
+		*,"$INTERFACE"|*,"$INTERFACE",*|"$INTERFACE",*|"$INTERFACE")
+		    break 2;;
+	    esac
 	done
 	exit
     done
@@ -73,77 +74,81 @@
     WPAS_OPTIONS="-P$PIDFILE $WPAS_OPTIONS"
 fi
 
-case "${MODE:-$1}" in
-    start)
-	mkdir -m u=rwx,go= -p "$CTRLDIR"
-	"$wpa_supplicant" -B -g "$CTRL" -p "$CTRLDIR" $WPAS_OPTIONS
-        for KEY in $ifkeys; do
-	    ADDRESS=`eval 'echo "$ADDRESS_'"$KEY"\"`
-	    INTERFACE=`addrtoif "$ADDRESS"`
-	    DRIVER=`eval 'echo "$WPA_DRIVER_'"$KEY"\"`
-	    IFDELAY=`eval 'echo "$DELAY_'"$KEY"\"`
-	    "$wpa_cli" -g "$CTRL" interface_add "$INTERFACE" "" \
-		"${DRIVER:-wext}" "$CTRLDIR" > /dev/null \
-		| sed -e '/^OK$/d'
-	    NETWORK=`"$wpa_cli" -p "$CTRLDIR" -i "$INTERFACE" \
-		add_network`
-	    eval wpa_interface_"$KEY"
-	    "$wpa_cli" -p "$CTRLDIR" -i "$INTERFACE" enable_network \
-		"$NETWORK" | sed -e '/^OK$/d'
-	    sleep "${IFDELAY:-$DELAY}" &
-	    sleep=$!
-	    while :; do
-		kill -0 $sleep 2>/dev/null || break
-		STATE=`"$wpa_cli" -p "$CTRLDIR" -i "$INTERFACE" \
-		    status | sed -n -e 's/^wpa_state=//p'`
-		if [ "$STATE" = COMPLETED ]; then
-		    while :; do
-			kill -0 $sleep 2>/dev/null || break 2
-			UP=`cat /sys/class/net/"$INTERFACE"/operstate`
-			if [ "$UP" = up ]; then
-			    kill $sleep 2>/dev/null
-			    break 2
-			fi
-			sleep 1
-		    done
-		fi
-		sleep 1
-	    done &
-	    wait $sleep || :
-	    IPADDRS=`eval 'echo "$IPADDRS_'"$KEY"\"`
-	    if [ -n "$IPADDRS" ]; then
-		if [ "$IPADDRS" = dhcp ]; then
-		    ipconfig -c dhcp -d "$INTERFACE" || :
-		    #dhclient "$INTERFACE"
-		else
-		    for ipaddr in $IPADDRS; do
-			"$ip" addr add "$ipaddr" dev "$INTERFACE"
-		    done
-		fi
-	    fi
-	    ROUTES=`eval 'echo "$ROUTES_'"$KEY"\"`
-	    if [ -n "$ROUTES" ]; then
-		for route in $ROUTES; do
-		    "$ip" route add "$route" dev "$BRIDGE"
-		done
-	    fi
-        done
-	;;
-    stop)
-	"$wpa_cli" -g "$CTRL" terminate 2>&1 | sed -e '/^OK$/d'
-        for KEY in $ifkeys; do
-	    ADDRESS=`eval 'echo "$ADDRESS_'"$KEY"\"`
-	    INTERFACE=`addrtoif "$ADDRESS"`
-	    "$ip" addr show scope global permanent dev "$INTERFACE" \
-		| while read type addr rest; do
+do_start(){
+    mkdir -m u=rwx,go= -p "$CTRLDIR"
+    "$wpa_supplicant" -B -g "$CTRL" -p "$CTRLDIR" $WPAS_OPTIONS
+    for KEY in $ifkeys; do
+	ADDRESS=`eval 'echo "$ADDRESS_'"$KEY"\"`
+	INTERFACE=`addrtoif "$ADDRESS"`
+	DRIVER=`eval 'echo "$WPA_DRIVER_'"$KEY"\"`
+	IFDELAY=`eval 'echo "$DELAY_'"$KEY"\"`
+	"$wpa_cli" -g "$CTRL" interface_add "$INTERFACE" "" \
+	    "${DRIVER:-wext}" "$CTRLDIR" > /dev/null \
+	    | sed -e '/^OK$/d'
+        NETWORK=`"$wpa_cli" -p "$CTRLDIR" -i "$INTERFACE" add_network`
+	eval wpa_interface_"$KEY"
+	"$wpa_cli" -p "$CTRLDIR" -i "$INTERFACE" enable_network \
+	    "$NETWORK" | sed -e '/^OK$/d'
+	sleep "${IFDELAY:-$DELAY}" &
+	sleep=$!
+	while :; do
+	    kill -0 $sleep 2>/dev/null || break
+	    STATE=`"$wpa_cli" -p "$CTRLDIR" -i "$INTERFACE" status \
+		| sed -n -e 's/^wpa_state=//p'`
+	    if [ "$STATE" = COMPLETED ]; then
+		while :; do
+		    kill -0 $sleep 2>/dev/null || break 2
+		    UP=`cat /sys/class/net/"$INTERFACE"/operstate`
+		    if [ "$UP" = up ]; then
+			kill $sleep 2>/dev/null
+			break 2
+		    fi
+		    sleep 1
+		done
+	    fi
+	    sleep 1
+	done &
+	wait $sleep || :
+	IPADDRS=`eval 'echo "$IPADDRS_'"$KEY"\"`
+	if [ -n "$IPADDRS" ]; then
+	    if [ "$IPADDRS" = dhcp ]; then
+		ipconfig -c dhcp -d "$INTERFACE" || :
+		#dhclient "$INTERFACE"
+	    else
+		for ipaddr in $IPADDRS; do
+		    "$ip" addr add "$ipaddr" dev "$INTERFACE"
+		done
+	    fi
+	fi
+	ROUTES=`eval 'echo "$ROUTES_'"$KEY"\"`
+	if [ -n "$ROUTES" ]; then
+	    for route in $ROUTES; do
+		"$ip" route add "$route" dev "$INTERFACE"
+	    done
+	fi
+    done
+}
+
+do_stop(){
+    "$wpa_cli" -g "$CTRL" terminate 2>&1 | sed -e '/^OK$/d'
+    for KEY in $ifkeys; do
+	ADDRESS=`eval 'echo "$ADDRESS_'"$KEY"\"`
+	INTERFACE=`addrtoif "$ADDRESS"`
+	"$ip" addr show scope global permanent dev "$INTERFACE" \
+	    | while read type addr rest; do
 		case "$type" in
 		    inet|inet6)
 			"$ip" addr del "$addr" dev "$INTERFACE"
 			;;
 		esac
 	    done
-	    "$ip" link set dev "$INTERFACE" down
-	done
+	"$ip" link set dev "$INTERFACE" down
+    done
+}
+
+case "${MODE:-$1}" in
+    start|stop)
+	do_"${MODE:-$1}"
 	;;
     files)
 	echo "$wpa_supplicant"

=== modified file 'plugin-runner.c'
--- plugin-runner.c	2011-12-31 23:05:34 +0000
+++ plugin-runner.c	2014-06-14 23:43:07 +0000
@@ -2,8 +2,8 @@
 /*
  * Mandos plugin runner - Run Mandos plugins
  *
- * Copyright © 2008-2012 Teddy Hogeborn
- * Copyright © 2008-2012 Björn Påhlsson
+ * Copyright © 2008-2014 Teddy Hogeborn
+ * Copyright © 2008-2014 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -23,38 +23,37 @@
  */
 
 #define _GNU_SOURCE		/* TEMP_FAILURE_RETRY(), getline(),
-				   asprintf(), O_CLOEXEC */
+				   O_CLOEXEC, pipe2() */
 #include <stddef.h>		/* size_t, NULL */
 #include <stdlib.h>		/* malloc(), exit(), EXIT_SUCCESS,
 				   realloc() */
 #include <stdbool.h>		/* bool, true, false */
 #include <stdio.h>		/* fileno(), fprintf(),
-				   stderr, STDOUT_FILENO */
-#include <sys/types.h>	        /* DIR, fdopendir(), stat(), struct
-				   stat, waitpid(), WIFEXITED(),
-				   WEXITSTATUS(), wait(), pid_t,
-				   uid_t, gid_t, getuid(), getgid(),
-				   dirfd() */
+				   stderr, STDOUT_FILENO, fclose() */
+#include <sys/types.h>	        /* fstat(), struct stat, waitpid(),
+				   WIFEXITED(), WEXITSTATUS(), wait(),
+				   pid_t, uid_t, gid_t, getuid(),
+				   getgid() */
 #include <sys/select.h>		/* fd_set, select(), FD_ZERO(),
 				   FD_SET(), FD_ISSET(), FD_CLR */
 #include <sys/wait.h>		/* wait(), waitpid(), WIFEXITED(),
 				   WEXITSTATUS(), WTERMSIG(),
 				   WCOREDUMP() */
-#include <sys/stat.h>		/* struct stat, stat(), S_ISREG() */
+#include <sys/stat.h>		/* struct stat, fstat(), S_ISREG() */
 #include <iso646.h>		/* and, or, not */
-#include <dirent.h>		/* DIR, struct dirent, fdopendir(),
-				   readdir(), closedir(), dirfd() */
-#include <unistd.h>		/* struct stat, stat(), S_ISREG(),
-				   fcntl(), setuid(), setgid(),
-				   F_GETFD, F_SETFD, FD_CLOEXEC,
-				   access(), pipe(), fork(), close()
-				   dup2(), STDOUT_FILENO, _exit(),
-				   execv(), write(), read(),
-				   close() */
+#include <dirent.h>		/* struct dirent, scandirat() */
+#include <unistd.h>		/* fcntl(), F_GETFD, F_SETFD,
+				   FD_CLOEXEC, write(), STDOUT_FILENO,
+				   struct stat, fstat(), close(),
+				   setgid(), setuid(), S_ISREG(),
+				   faccessat() pipe2(), fork(),
+				   _exit(), dup2(), fexecve(), read()
+				*/
 #include <fcntl.h>		/* fcntl(), F_GETFD, F_SETFD,
-				   FD_CLOEXEC */
-#include <string.h>		/* strsep, strlen(), asprintf(),
-				   strsignal(), strcmp(), strncmp() */
+				   FD_CLOEXEC, openat(), scandirat(),
+				   pipe2() */
+#include <string.h>		/* strsep, strlen(), strsignal(),
+				   strcmp(), strncmp() */
 #include <errno.h>		/* errno */
 #include <argp.h>		/* struct argp_option, struct
 				   argp_state, struct argp,
@@ -72,6 +71,7 @@
 				   EX_CONFIG, EX_UNAVAILABLE, EX_OK */
 #include <errno.h> 		/* errno */
 #include <error.h>		/* error() */
+#include <fnmatch.h>		/* fnmatch() */
 
 #define BUFFER_SIZE 256
 
@@ -105,6 +105,7 @@
 
 /* Gets an existing plugin based on name,
    or if none is found, creates a new one */
+__attribute__((warn_unused_result))
 static plugin *getplugin(char *name){
   /* Check for existing plugin with that name */
   for(plugin *p = plugin_list; p != NULL; p = p->next){
@@ -171,18 +172,20 @@
 }
 
 /* Helper function for add_argument and add_environment */
-__attribute__((nonnull))
+__attribute__((nonnull, warn_unused_result))
 static bool add_to_char_array(const char *new, char ***array,
 			      int *len){
   /* Resize the pointed-to array to hold one more pointer */
+  char **new_array = NULL;
   do {
-    *array = realloc(*array, sizeof(char *)
-		     * (size_t) ((*len) + 2));
-  } while(*array == NULL and errno == EINTR);
+    new_array = realloc(*array, sizeof(char *)
+			* (size_t) ((*len) + 2));
+  } while(new_array == NULL and errno == EINTR);
   /* Malloc check */
-  if(*array == NULL){
+  if(new_array == NULL){
     return false;
   }
+  *array = new_array;
   /* Make a copy of the new string */
   char *copy;
   do {
@@ -200,7 +203,7 @@
 }
 
 /* Add to a plugin's argument vector */
-__attribute__((nonnull(2)))
+__attribute__((nonnull(2), warn_unused_result))
 static bool add_argument(plugin *p, const char *arg){
   if(p == NULL){
     return false;
@@ -209,7 +212,7 @@
 }
 
 /* Add to a plugin's environment */
-__attribute__((nonnull(2)))
+__attribute__((nonnull(2), warn_unused_result))
 static bool add_environment(plugin *p, const char *def, bool replace){
   if(p == NULL){
     return false;
@@ -217,19 +220,19 @@
   /* namelen = length of name of environment variable */
   size_t namelen = (size_t)(strchrnul(def, '=') - def);
   /* Search for this environment variable */
-  for(char **e = p->environ; *e != NULL; e++){
-    if(strncmp(*e, def, namelen + 1) == 0){
+  for(char **envdef = p->environ; *envdef != NULL; envdef++){
+    if(strncmp(*envdef, def, namelen + 1) == 0){
       /* It already exists */
       if(replace){
-	char *new;
+	char *new_envdef;
 	do {
-	  new = realloc(*e, strlen(def) + 1);
-	} while(new == NULL and errno == EINTR);
-	if(new == NULL){
+	  new_envdef = realloc(*envdef, strlen(def) + 1);
+	} while(new_envdef == NULL and errno == EINTR);
+	if(new_envdef == NULL){
 	  return false;
 	}
-	*e = new;
-	strcpy(*e, def);
+	*envdef = new_envdef;
+	strcpy(*envdef, def);
       }
       return true;
     }
@@ -237,11 +240,13 @@
   return add_to_char_array(def, &(p->environ), &(p->envc));
 }
 
+#ifndef O_CLOEXEC
 /*
  * Based on the example in the GNU LibC manual chapter 13.13 "File
  * Descriptor Flags".
  | [[info:libc:Descriptor%20Flags][File Descriptor Flags]] |
  */
+__attribute__((warn_unused_result))
 static int set_cloexec_flag(int fd){
   int ret = (int)TEMP_FAILURE_RETRY(fcntl(fd, F_GETFD, 0));
   /* If reading the flags failed, return error indication now. */
@@ -252,6 +257,7 @@
   return (int)TEMP_FAILURE_RETRY(fcntl(fd, F_SETFD,
 				       ret | FD_CLOEXEC));
 }
+#endif	/* not O_CLOEXEC */
 
 
 /* Mark processes as completed when they exit, and save their exit
@@ -289,7 +295,7 @@
 }
 
 /* Prints out a password to stdout */
-__attribute__((nonnull))
+__attribute__((nonnull, warn_unused_result))
 static bool print_out_password(const char *buffer, size_t length){
   ssize_t ret;
   for(size_t written = 0; written < length; written += (size_t)ret){
@@ -343,9 +349,7 @@
   char *plugindir = NULL;
   char *argfile = NULL;
   FILE *conffp;
-  size_t d_name_len;
-  DIR *dir = NULL;
-  struct dirent *dirst;
+  struct dirent **direntries = NULL;
   struct stat st;
   fd_set rfds_all;
   int ret, maxfd = 0;
@@ -359,6 +363,7 @@
 				      .sa_flags = SA_NOCLDSTOP };
   char **custom_argv = NULL;
   int custom_argc = 0;
+  int dir_fd = -1;
   
   /* Establish a signal handler */
   sigemptyset(&sigchld_action.sa_mask);
@@ -435,10 +440,13 @@
 	    break;
 	  }
 	}
+	errno = 0;
       }
       break;
     case 'G':			/* --global-env */
-      add_environment(getplugin(NULL), arg, true);
+      if(add_environment(getplugin(NULL), arg, true)){
+	errno = 0;
+      }
       break;
     case 'o':			/* --options-for */
       {
@@ -461,6 +469,7 @@
 	    break;
 	  }
 	}
+	errno = 0;
       }
       break;
     case 'E':			/* --env-for */
@@ -478,7 +487,9 @@
 	  errno = EINVAL;
 	  break;
 	}
-	add_environment(getplugin(arg), envdef, true);
+	if(add_environment(getplugin(arg), envdef, true)){
+	  errno = 0;
+	}
       }
       break;
     case 'd':			/* --disable */
@@ -486,6 +497,7 @@
 	plugin *p = getplugin(arg);
 	if(p != NULL){
 	  p->disabled = true;
+	  errno = 0;
 	}
       }
       break;
@@ -494,12 +506,16 @@
 	plugin *p = getplugin(arg);
 	if(p != NULL){
 	  p->disabled = false;
+	  errno = 0;
 	}
       }
       break;
     case 128:			/* --plugin-dir */
       free(plugindir);
       plugindir = strdup(arg);
+      if(plugindir != NULL){
+	errno = 0;
+      }
       break;
     case 129:			/* --config-file */
       /* This is already done by parse_opt_config_file() */
@@ -513,6 +529,7 @@
 	break;
       }
       uid = (uid_t)tmp_id;
+      errno = 0;
       break;
     case 131:			/* --groupid */
       tmp_id = strtoimax(arg, &tmp, 10);
@@ -523,6 +540,7 @@
 	break;
       }
       gid = (gid_t)tmp_id;
+      errno = 0;
       break;
     case 132:			/* --debug */
       debug = true;
@@ -576,6 +594,9 @@
     case 129:			/* --config-file */
       free(argfile);
       argfile = strdup(arg);
+      if(argfile != NULL){
+	errno = 0;
+      }
       break;
     case 130:			/* --userid */
     case 131:			/* --groupid */
@@ -664,13 +685,19 @@
 	}
 	
 	custom_argc += 1;
-	custom_argv = realloc(custom_argv, sizeof(char *)
-			      * ((unsigned int) custom_argc + 1));
-	if(custom_argv == NULL){
-	  error(0, errno, "realloc");
-	  exitstatus = EX_OSERR;
-	  free(org_line);
-	  goto fallback;
+	{
+	  char **new_argv = realloc(custom_argv, sizeof(char *)
+				    * ((unsigned int)
+				       custom_argc + 1));
+	  if(new_argv == NULL){
+	    error(0, errno, "realloc");
+	    exitstatus = EX_OSERR;
+	    free(new_arg);
+	    free(org_line);
+	    goto fallback;
+	  } else {
+	    custom_argv = new_argv;
+	  }
 	}
 	custom_argv[custom_argc-1] = new_arg;
 	custom_argv[custom_argc] = NULL;
@@ -753,7 +780,9 @@
        <http://bugs.debian.org/633582> */
     int plugindir_fd = open(/* plugindir or */ PDIR, O_RDONLY);
     if(plugindir_fd == -1){
-      error(0, errno, "open");
+      if(errno != ENOENT){
+	error(0, errno, "open(\"" PDIR "\")");
+      }
     } else {
       ret = (int)TEMP_FAILURE_RETRY(fstat(plugindir_fd, &st));
       if(ret == -1){
@@ -771,7 +800,7 @@
   }
   
   /* Lower permissions */
-  setgid(gid);
+  ret = setgid(gid);
   if(ret == -1){
     error(0, errno, "setgid");
   }
@@ -782,24 +811,13 @@
   
   /* Open plugin directory with close_on_exec flag */
   {
-    int dir_fd = -1;
-    if(plugindir == NULL){
-      dir_fd = open(PDIR, O_RDONLY |
-#ifdef O_CLOEXEC
-		    O_CLOEXEC
-#else  /* not O_CLOEXEC */
-		    0
-#endif	/* not O_CLOEXEC */
-		    );
-    } else {
-      dir_fd = open(plugindir, O_RDONLY |
-#ifdef O_CLOEXEC
-		    O_CLOEXEC
-#else  /* not O_CLOEXEC */
-		    0
-#endif	/* not O_CLOEXEC */
-		    );
-    }
+    dir_fd = open(plugindir != NULL ? plugindir : PDIR, O_RDONLY |
+#ifdef O_CLOEXEC
+		  O_CLOEXEC
+#else  /* not O_CLOEXEC */
+		  0
+#endif	/* not O_CLOEXEC */
+		  );
     if(dir_fd == -1){
       error(0, errno, "Could not open plugin dir");
       exitstatus = EX_UNAVAILABLE;
@@ -811,129 +829,99 @@
     ret = set_cloexec_flag(dir_fd);
     if(ret < 0){
       error(0, errno, "set_cloexec_flag");
-      TEMP_FAILURE_RETRY(close(dir_fd));
       exitstatus = EX_OSERR;
       goto fallback;
     }
 #endif	/* O_CLOEXEC */
-    
-    dir = fdopendir(dir_fd);
-    if(dir == NULL){
-      error(0, errno, "Could not open plugin dir");
-      TEMP_FAILURE_RETRY(close(dir_fd));
-      exitstatus = EX_OSERR;
-      goto fallback;
+  }
+  
+  int good_name(const struct dirent * const dirent){
+    const char * const patterns[] = { ".*", "#*#", "*~", "*.dpkg-new",
+				      "*.dpkg-old", "*.dpkg-bak",
+				      "*.dpkg-divert", NULL };
+#ifdef __GNUC__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wcast-qual"
+#endif
+    for(const char **pat = (const char **)patterns;
+	*pat != NULL; pat++){
+#ifdef __GNUC__
+#pragma GCC diagnostic pop
+#endif
+      if(fnmatch(*pat, dirent->d_name, FNM_FILE_NAME | FNM_PERIOD)
+	 != FNM_NOMATCH){
+	if(debug){
+	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
+		    " matching pattern %s\n", dirent->d_name, *pat);
+	}
+	return 0;
+      }
     }
+    return 1;
+  }
+  
+#ifdef __GLIBC__
+#if __GLIBC_PREREQ(2, 15)
+  int numplugins = scandirat(dir_fd, ".", &direntries, good_name,
+			     alphasort);
+#else  /* not __GLIBC_PREREQ(2, 15) */
+  int numplugins = scandir(plugindir != NULL ? plugindir : PDIR,
+			   &direntries, good_name, alphasort);
+#endif	/* not __GLIBC_PREREQ(2, 15) */
+#else	/* not __GLIBC__ */
+  int numplugins = scandir(plugindir != NULL ? plugindir : PDIR,
+			   &direntries, good_name, alphasort);
+#endif	/* not __GLIBC__ */
+  if(numplugins == -1){
+    error(0, errno, "Could not scan plugin dir");
+    direntries = NULL;
+    exitstatus = EX_OSERR;
+    goto fallback;
   }
   
   FD_ZERO(&rfds_all);
   
   /* Read and execute any executable in the plugin directory*/
-  while(true){
-    do {
-      dirst = readdir(dir);
-    } while(dirst == NULL and errno == EINTR);
-    
-    /* All directory entries have been processed */
-    if(dirst == NULL){
-      if(errno == EBADF){
-	error(0, errno, "readdir");
-	exitstatus = EX_IOERR;
-	goto fallback;
-      }
-      break;
-    }
-    
-    d_name_len = strlen(dirst->d_name);
-    
-    /* Ignore dotfiles, backup files and other junk */
-    {
-      bool bad_name = false;
-      
-      const char const *bad_prefixes[] = { ".", "#", NULL };
-      
-      const char const *bad_suffixes[] = { "~", "#", ".dpkg-new",
-					   ".dpkg-old",
-					   ".dpkg-bak",
-					   ".dpkg-divert", NULL };
-      for(const char **pre = bad_prefixes; *pre != NULL; pre++){
-	size_t pre_len = strlen(*pre);
-	if((d_name_len >= pre_len)
-	   and strncmp((dirst->d_name), *pre, pre_len) == 0){
-	  if(debug){
-	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
-		    " with bad prefix %s\n", dirst->d_name, *pre);
-	  }
-	  bad_name = true;
-	  break;
-	}
-      }
-      if(bad_name){
-	continue;
-      }
-      for(const char **suf = bad_suffixes; *suf != NULL; suf++){
-	size_t suf_len = strlen(*suf);
-	if((d_name_len >= suf_len)
-	   and (strcmp((dirst->d_name) + d_name_len-suf_len, *suf)
-		== 0)){
-	  if(debug){
-	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
-		    " with bad suffix %s\n", dirst->d_name, *suf);
-	  }
-	  bad_name = true;
-	  break;
-	}
-      }
-      
-      if(bad_name){
-	continue;
-      }
-    }
-    
-    char *filename;
-    if(plugindir == NULL){
-      ret = (int)TEMP_FAILURE_RETRY(asprintf(&filename, PDIR "/%s",
-					     dirst->d_name));
-    } else {
-      ret = (int)TEMP_FAILURE_RETRY(asprintf(&filename, "%s/%s",
-					     plugindir,
-					     dirst->d_name));
-    }
-    if(ret < 0){
-      error(0, errno, "asprintf");
+  for(int i = 0; i < numplugins; i++){
+    
+    int plugin_fd = openat(dir_fd, direntries[i]->d_name, O_RDONLY);
+    if(plugin_fd == -1){
+      error(0, errno, "Could not open plugin");
       continue;
     }
-    
-    ret = (int)TEMP_FAILURE_RETRY(stat(filename, &st));
+    ret = (int)TEMP_FAILURE_RETRY(fstat(plugin_fd, &st));
     if(ret == -1){
       error(0, errno, "stat");
-      free(filename);
+      TEMP_FAILURE_RETRY(close(plugin_fd));
       continue;
     }
     
     /* Ignore non-executable files */
     if(not S_ISREG(st.st_mode)
-       or (TEMP_FAILURE_RETRY(access(filename, X_OK)) != 0)){
+       or (TEMP_FAILURE_RETRY(faccessat(dir_fd, direntries[i]->d_name,
+					X_OK, 0)) != 0)){
       if(debug){
-	fprintf(stderr, "Ignoring plugin dir entry \"%s\""
-		" with bad type or mode\n", filename);
+	fprintf(stderr, "Ignoring plugin dir entry \"%s/%s\""
+		" with bad type or mode\n",
+		plugindir != NULL ? plugindir : PDIR,
+		direntries[i]->d_name);
       }
-      free(filename);
+      TEMP_FAILURE_RETRY(close(plugin_fd));
       continue;
     }
     
-    plugin *p = getplugin(dirst->d_name);
+    plugin *p = getplugin(direntries[i]->d_name);
     if(p == NULL){
       error(0, errno, "getplugin");
-      free(filename);
+      TEMP_FAILURE_RETRY(close(plugin_fd));
       continue;
     }
     if(p->disabled){
       if(debug){
 	fprintf(stderr, "Ignoring disabled plugin \"%s\"\n",
-		dirst->d_name);
+		direntries[i]->d_name);
       }
-      free(filename);
+      TEMP_FAILURE_RETRY(close(plugin_fd));
       continue;
     }
     {
@@ -953,9 +941,8 @@
 	}
       }
     }
-    /* If this plugin has any environment variables, we will call
-       using execve and need to duplicate the environment from this
-       process, too. */
+    /* If this plugin has any environment variables, we need to
+       duplicate the environment from this process, too. */
     if(p->environ[0] != NULL){
       for(char **e = environ; *e != NULL; e++){
 	if(not add_environment(p, *e, false)){
@@ -965,25 +952,43 @@
     }
     
     int pipefd[2];
+#ifndef O_CLOEXEC
     ret = (int)TEMP_FAILURE_RETRY(pipe(pipefd));
+#else  /* O_CLOEXEC */
+    ret = (int)TEMP_FAILURE_RETRY(pipe2(pipefd, O_CLOEXEC));
+#endif	/* O_CLOEXEC */
     if(ret == -1){
       error(0, errno, "pipe");
       exitstatus = EX_OSERR;
       goto fallback;
     }
+    if(pipefd[0] >= FD_SETSIZE){
+      fprintf(stderr, "pipe()[0] (%d) >= FD_SETSIZE (%d)", pipefd[0],
+	      FD_SETSIZE);
+      TEMP_FAILURE_RETRY(close(pipefd[0]));
+      TEMP_FAILURE_RETRY(close(pipefd[1]));
+      exitstatus = EX_OSERR;
+      goto fallback;
+    }
+#ifndef O_CLOEXEC
     /* Ask OS to automatic close the pipe on exec */
     ret = set_cloexec_flag(pipefd[0]);
     if(ret < 0){
       error(0, errno, "set_cloexec_flag");
+      TEMP_FAILURE_RETRY(close(pipefd[0]));
+      TEMP_FAILURE_RETRY(close(pipefd[1]));
       exitstatus = EX_OSERR;
       goto fallback;
     }
     ret = set_cloexec_flag(pipefd[1]);
     if(ret < 0){
       error(0, errno, "set_cloexec_flag");
+      TEMP_FAILURE_RETRY(close(pipefd[0]));
+      TEMP_FAILURE_RETRY(close(pipefd[1]));
       exitstatus = EX_OSERR;
       goto fallback;
     }
+#endif	/* not O_CLOEXEC */
     /* Block SIGCHLD until process is safely in process list */
     ret = (int)TEMP_FAILURE_RETRY(sigprocmask(SIG_BLOCK,
 					      &sigchld_action.sa_mask,
@@ -1000,6 +1005,10 @@
     } while(pid == -1 and errno == EINTR);
     if(pid == -1){
       error(0, errno, "fork");
+      TEMP_FAILURE_RETRY(sigprocmask(SIG_UNBLOCK,
+				     &sigchld_action.sa_mask, NULL));
+      TEMP_FAILURE_RETRY(close(pipefd[0]));
+      TEMP_FAILURE_RETRY(close(pipefd[1]));
       exitstatus = EX_OSERR;
       goto fallback;
     }
@@ -1022,29 +1031,20 @@
 	_exit(EX_OSERR);
       }
       
-      if(dirfd(dir) < 0){
-	/* If dir has no file descriptor, we could not set FD_CLOEXEC
-	   above and must now close it manually here. */
-	closedir(dir);
-      }
-      if(p->environ[0] == NULL){
-	if(execv(filename, p->argv) < 0){
-	  error(0, errno, "execv for %s", filename);
-	  _exit(EX_OSERR);
-	}
-      } else {
-	if(execve(filename, p->argv, p->environ) < 0){
-	  error(0, errno, "execve for %s", filename);
-	  _exit(EX_OSERR);
-	}
+      if(fexecve(plugin_fd, p->argv,
+		(p->environ[0] != NULL) ? p->environ : environ) < 0){
+	error(0, errno, "fexecve for %s/%s",
+	      plugindir != NULL ? plugindir : PDIR,
+	      direntries[i]->d_name);
+	_exit(EX_OSERR);
       }
       /* no return */
     }
     /* Parent process */
     TEMP_FAILURE_RETRY(close(pipefd[1])); /* Close unused write end of
 					     pipe */
-    free(filename);
-    plugin *new_plugin = getplugin(dirst->d_name);
+    TEMP_FAILURE_RETRY(close(plugin_fd));
+    plugin *new_plugin = getplugin(direntries[i]->d_name);
     if(new_plugin == NULL){
       error(0, errno, "getplugin");
       ret = (int)(TEMP_FAILURE_RETRY
@@ -1071,16 +1071,30 @@
       goto fallback;
     }
     
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wsign-conversion"
+#endif
+#endif
     FD_SET(new_plugin->fd, &rfds_all); /* Spurious warning from
-					  -Wconversion */
+					  -Wconversion in GNU libc
+					  before 2.16 */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic pop
+#endif
+#endif
     
     if(maxfd < new_plugin->fd){
       maxfd = new_plugin->fd;
     }
   }
   
-  TEMP_FAILURE_RETRY(closedir(dir));
-  dir = NULL;
+  free(direntries);
+  direntries = NULL;
+  TEMP_FAILURE_RETRY(close(dir_fd));
+  dir_fd = -1;
   free_plugin(getplugin(NULL));
   
   for(plugin *p = plugin_list; p != NULL; p = p->next){
@@ -1132,8 +1146,20 @@
 	  }
 	  
 	  /* Remove the plugin */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wsign-conversion"
+#endif
+#endif
 	  FD_CLR(proc->fd, &rfds_all); /* Spurious warning from
-					  -Wconversion */
+					  -Wconversion in GNU libc
+					  before 2.16 */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic pop
+#endif
+#endif
 	  
 	  /* Block signal while modifying process_list */
 	  ret = (int)TEMP_FAILURE_RETRY(sigprocmask
@@ -1179,22 +1205,37 @@
       }
       
       /* This process has not completed.  Does it have any output? */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wsign-conversion"
+#endif
+#endif
       if(proc->eof or not FD_ISSET(proc->fd, &rfds)){ /* Spurious
 							 warning from
-							 -Wconversion */
+							 -Wconversion
+							 in GNU libc
+							 before
+							 2.16 */
+#if defined (__GNUC__) and defined (__GLIBC__)
+#if not __GLIBC_PREREQ(2, 16)
+#pragma GCC diagnostic pop
+#endif
+#endif
 	/* This process had nothing to say at this time */
 	proc = proc->next;
 	continue;
       }
       /* Before reading, make the process' data buffer large enough */
       if(proc->buffer_length + BUFFER_SIZE > proc->buffer_size){
-	proc->buffer = realloc(proc->buffer, proc->buffer_size
-			       + (size_t) BUFFER_SIZE);
-	if(proc->buffer == NULL){
+	char *new_buffer = realloc(proc->buffer, proc->buffer_size
+				   + (size_t) BUFFER_SIZE);
+	if(new_buffer == NULL){
 	  error(0, errno, "malloc");
 	  exitstatus = EX_OSERR;
 	  goto fallback;
 	}
+	proc->buffer = new_buffer;
 	proc->buffer_size += BUFFER_SIZE;
       }
       /* Read from the process */
@@ -1253,8 +1294,10 @@
     free(custom_argv);
   }
   
-  if(dir != NULL){
-    closedir(dir);
+  free(direntries);
+  
+  if(dir_fd != -1){
+    TEMP_FAILURE_RETRY(close(dir_fd));
   }
   
   /* Kill the processes */

=== modified file 'plugins.d/askpass-fifo.c'
--- plugins.d/askpass-fifo.c	2011-12-31 23:05:34 +0000
+++ plugins.d/askpass-fifo.c	2014-03-29 02:38:15 +0000
@@ -2,8 +2,8 @@
 /*
  * Askpass-FIFO - Read a password from a FIFO and output it
  * 
- * Copyright © 2008-2012 Teddy Hogeborn
- * Copyright © 2008-2012 Björn Påhlsson
+ * Copyright © 2008-2014 Teddy Hogeborn
+ * Copyright © 2008-2014 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -55,7 +55,7 @@
   
   va_start(ap, formatstring);
   ret = vasprintf(&text, formatstring, ap);
-  if (ret == -1){
+  if(ret == -1){
     fprintf(stderr, "Mandos plugin %s: ",
 	    program_invocation_short_name);
     vfprintf(stderr, formatstring, ap);

=== modified file 'plugins.d/mandos-client.c'
--- plugins.d/mandos-client.c	2011-12-31 23:05:34 +0000
+++ plugins.d/mandos-client.c	2014-07-12 13:13:28 +0000
@@ -9,8 +9,8 @@
  * "browse_callback", and parts of "main".
  * 
  * Everything else is
- * Copyright © 2008-2012 Teddy Hogeborn
- * Copyright © 2008-2012 Björn Påhlsson
+ * Copyright © 2008-2014 Teddy Hogeborn
+ * Copyright © 2008-2014 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -32,16 +32,16 @@
 /* Needed by GPGME, specifically gpgme_data_seek() */
 #ifndef _LARGEFILE_SOURCE
 #define _LARGEFILE_SOURCE
-#endif
+#endif	/* not _LARGEFILE_SOURCE */
 #ifndef _FILE_OFFSET_BITS
 #define _FILE_OFFSET_BITS 64
-#endif
+#endif	/* not _FILE_OFFSET_BITS */
 
 #define _GNU_SOURCE		/* TEMP_FAILURE_RETRY(), asprintf() */
 
 #include <stdio.h>		/* fprintf(), stderr, fwrite(),
-				   stdout, ferror(), remove() */
-#include <stdint.h> 		/* uint16_t, uint32_t */
+				   stdout, ferror() */
+#include <stdint.h> 		/* uint16_t, uint32_t, intptr_t */
 #include <stddef.h>		/* NULL, size_t, ssize_t */
 #include <stdlib.h> 		/* free(), EXIT_SUCCESS, srand(),
 				   strtof(), abort() */
@@ -55,13 +55,13 @@
 				   opendir(), DIR */
 #include <sys/stat.h>		/* open(), S_ISREG */
 #include <sys/socket.h>		/* socket(), struct sockaddr_in6,
-				   inet_pton(), connect() */
-#include <fcntl.h>		/* open() */
+				   inet_pton(), connect(),
+				   getnameinfo() */
+#include <fcntl.h>		/* open(), unlinkat() */
 #include <dirent.h>		/* opendir(), struct dirent, readdir()
 				 */
 #include <inttypes.h>		/* PRIu16, PRIdMAX, intmax_t,
 				   strtoimax() */
-#include <assert.h>		/* assert() */
 #include <errno.h>		/* perror(), errno,
 				   program_invocation_short_name */
 #include <time.h>		/* nanosleep(), time(), sleep() */
@@ -73,8 +73,9 @@
 				*/
 #include <unistd.h>		/* close(), SEEK_SET, off_t, write(),
 				   getuid(), getgid(), seteuid(),
-				   setgid(), pause(), _exit() */
-#include <arpa/inet.h>		/* inet_pton(), htons, inet_ntop() */
+				   setgid(), pause(), _exit(),
+				   unlinkat() */
+#include <arpa/inet.h>		/* inet_pton(), htons() */
 #include <iso646.h>		/* not, or, and */
 #include <argp.h>		/* struct argp_option, error_t, struct
 				   argp_state, struct argp,
@@ -88,6 +89,12 @@
 #include <sys/wait.h>		/* waitpid(), WIFEXITED(),
 				   WEXITSTATUS(), WTERMSIG() */
 #include <grp.h>		/* setgroups() */
+#include <argz.h>		/* argz_add_sep(), argz_next(),
+				   argz_delete(), argz_append(),
+				   argz_stringify(), argz_add(),
+				   argz_count() */
+#include <netdb.h>		/* getnameinfo(), NI_NUMERICHOST,
+				   EAI_SYSTEM, gai_strerror() */
 
 #ifdef __linux__
 #include <sys/klog.h> 		/* klogctl() */
@@ -135,11 +142,14 @@
 static const char sys_class_net[] = "/sys/class/net";
 char *connect_to = NULL;
 const char *hookdir = HOOKDIR;
+int hookdir_fd = -1;
+uid_t uid = 65534;
+gid_t gid = 65534;
 
 /* Doubly linked list that need to be circularly linked when used */
 typedef struct server{
   const char *ip;
-  uint16_t port;
+  in_port_t port;
   AvahiIfIndex if_index;
   int af;
   struct timespec last_seen;
@@ -149,7 +159,6 @@
 
 /* Used for passing in values through the Avahi callback functions */
 typedef struct {
-  AvahiSimplePoll *simple_poll;
   AvahiServer *server;
   gnutls_certificate_credentials_t cred;
   unsigned int dh_bits;
@@ -157,13 +166,12 @@
   const char *priority;
   gpgme_ctx_t ctx;
   server *current_server;
+  char *interfaces;
+  size_t interfaces_size;
 } mandos_context;
 
-/* global context so signal handler can reach it*/
-mandos_context mc = { .simple_poll = NULL, .server = NULL,
-		      .dh_bits = 1024, .priority = "SECURE256"
-		      ":!CTYPE-X.509:+CTYPE-OPENPGP",
-		      .current_server = NULL };
+/* global so signal handler can reach it*/
+AvahiSimplePoll *simple_poll;
 
 sig_atomic_t quit_now = 0;
 int signal_received = 0;
@@ -177,14 +185,14 @@
   perror(print_text);
 }
 
-__attribute__((format (gnu_printf, 2, 3)))
+__attribute__((format (gnu_printf, 2, 3), nonnull))
 int fprintf_plus(FILE *stream, const char *format, ...){
   va_list ap;
   va_start (ap, format);
   
   TEMP_FAILURE_RETRY(fprintf(stream, "Mandos plugin %s: ",
 			     program_invocation_short_name));
-  return TEMP_FAILURE_RETRY(vfprintf(stream, format, ap));
+  return (int)TEMP_FAILURE_RETRY(vfprintf(stream, format, ap));
 }
 
 /*
@@ -192,21 +200,28 @@
  * bytes. "buffer_capacity" is how much is currently allocated,
  * "buffer_length" is how much is already used.
  */
+__attribute__((nonnull, warn_unused_result))
 size_t incbuffer(char **buffer, size_t buffer_length,
 		 size_t buffer_capacity){
   if(buffer_length + BUFFER_SIZE > buffer_capacity){
-    *buffer = realloc(*buffer, buffer_capacity + BUFFER_SIZE);
-    if(buffer == NULL){
+    char *new_buf = realloc(*buffer, buffer_capacity + BUFFER_SIZE);
+    if(new_buf == NULL){
+      int old_errno = errno;
+      free(*buffer);
+      errno = old_errno;
+      *buffer = NULL;
       return 0;
     }
+    *buffer = new_buf;
     buffer_capacity += BUFFER_SIZE;
   }
   return buffer_capacity;
 }
 
 /* Add server to set of servers to retry periodically */
-bool add_server(const char *ip, uint16_t port, AvahiIfIndex if_index,
-		int af){
+__attribute__((nonnull, warn_unused_result))
+bool add_server(const char *ip, in_port_t port, AvahiIfIndex if_index,
+		int af, server **current_server){
   int ret;
   server *new_server = malloc(sizeof(server));
   if(new_server == NULL){
@@ -219,24 +234,27 @@
 			  .af = af };
   if(new_server->ip == NULL){
     perror_plus("strdup");
+    free(new_server);
+    return false;
+  }
+  ret = clock_gettime(CLOCK_MONOTONIC, &(new_server->last_seen));
+  if(ret == -1){
+    perror_plus("clock_gettime");
+    free(new_server->ip);
+    free(new_server);
     return false;
   }
   /* Special case of first server */
-  if (mc.current_server == NULL){
+  if(*current_server == NULL){
     new_server->next = new_server;
     new_server->prev = new_server;
-    mc.current_server = new_server;
-  /* Place the new server last in the list */
+    *current_server = new_server;
   } else {
-    new_server->next = mc.current_server;
-    new_server->prev = mc.current_server->prev;
+    /* Place the new server last in the list */
+    new_server->next = *current_server;
+    new_server->prev = (*current_server)->prev;
     new_server->prev->next = new_server;
-    mc.current_server->prev = new_server;
-  }
-  ret = clock_gettime(CLOCK_MONOTONIC, &mc.current_server->last_seen);
-  if(ret == -1){
-    perror_plus("clock_gettime");
-    return false;
+    (*current_server)->prev = new_server;
   }
   return true;
 }
@@ -244,16 +262,18 @@
 /* 
  * Initialize GPGME.
  */
-static bool init_gpgme(const char *seckey, const char *pubkey,
-		       const char *tempdir){
+__attribute__((nonnull, warn_unused_result))
+static bool init_gpgme(const char * const seckey,
+		       const char * const pubkey,
+		       const char * const tempdir,
+		       mandos_context *mc){
   gpgme_error_t rc;
   gpgme_engine_info_t engine_info;
   
-  
   /*
    * Helper function to insert pub and seckey to the engine keyring.
    */
-  bool import_key(const char *filename){
+  bool import_key(const char * const filename){
     int ret;
     int fd;
     gpgme_data_t pgp_data;
@@ -271,7 +291,7 @@
       return false;
     }
     
-    rc = gpgme_op_import(mc.ctx, pgp_data);
+    rc = gpgme_op_import(mc->ctx, pgp_data);
     if(rc != GPG_ERR_NO_ERROR){
       fprintf_plus(stderr, "bad gpgme_op_import: %s: %s\n",
 		   gpgme_strsource(rc), gpgme_strerror(rc));
@@ -321,7 +341,7 @@
   }
   
   /* Create new GPGME "context" */
-  rc = gpgme_new(&(mc.ctx));
+  rc = gpgme_new(&(mc->ctx));
   if(rc != GPG_ERR_NO_ERROR){
     fprintf_plus(stderr, "Mandos plugin mandos-client: "
 		 "bad gpgme_new: %s: %s\n", gpgme_strsource(rc),
@@ -340,9 +360,11 @@
  * Decrypt OpenPGP data.
  * Returns -1 on error
  */
+__attribute__((nonnull, warn_unused_result))
 static ssize_t pgp_packet_decrypt(const char *cryptotext,
 				  size_t crypto_size,
-				  char **plaintext){
+				  char **plaintext,
+				  mandos_context *mc){
   gpgme_data_t dh_crypto, dh_plain;
   gpgme_error_t rc;
   ssize_t ret;
@@ -374,14 +396,14 @@
   
   /* Decrypt data from the cryptotext data buffer to the plaintext
      data buffer */
-  rc = gpgme_op_decrypt(mc.ctx, dh_crypto, dh_plain);
+  rc = gpgme_op_decrypt(mc->ctx, dh_crypto, dh_plain);
   if(rc != GPG_ERR_NO_ERROR){
     fprintf_plus(stderr, "bad gpgme_op_decrypt: %s: %s\n",
 		 gpgme_strsource(rc), gpgme_strerror(rc));
     plaintext_length = -1;
     if(debug){
       gpgme_decrypt_result_t result;
-      result = gpgme_op_decrypt_result(mc.ctx);
+      result = gpgme_op_decrypt_result(mc->ctx);
       if(result == NULL){
 	fprintf_plus(stderr, "gpgme_op_decrypt_result failed\n");
       } else {
@@ -464,22 +486,25 @@
   return plaintext_length;
 }
 
-static const char * safer_gnutls_strerror(int value){
-  const char *ret = gnutls_strerror(value); /* Spurious warning from
-					       -Wunreachable-code */
+__attribute__((warn_unused_result))
+static const char *safer_gnutls_strerror(int value){
+  const char *ret = gnutls_strerror(value);
   if(ret == NULL)
     ret = "(unknown)";
   return ret;
 }
 
 /* GnuTLS log function callback */
+__attribute__((nonnull))
 static void debuggnutls(__attribute__((unused)) int level,
 			const char* string){
   fprintf_plus(stderr, "GnuTLS: %s", string);
 }
 
+__attribute__((nonnull, warn_unused_result))
 static int init_gnutls_global(const char *pubkeyfilename,
-			      const char *seckeyfilename){
+			      const char *seckeyfilename,
+			      mandos_context *mc){
   int ret;
   
   if(debug){
@@ -502,7 +527,7 @@
   }
   
   /* OpenPGP credentials */
-  ret = gnutls_certificate_allocate_credentials(&mc.cred);
+  ret = gnutls_certificate_allocate_credentials(&mc->cred);
   if(ret != GNUTLS_E_SUCCESS){
     fprintf_plus(stderr, "GnuTLS memory error: %s\n",
 		 safer_gnutls_strerror(ret));
@@ -518,7 +543,7 @@
   }
   
   ret = gnutls_certificate_set_openpgp_key_file
-    (mc.cred, pubkeyfilename, seckeyfilename,
+    (mc->cred, pubkeyfilename, seckeyfilename,
      GNUTLS_OPENPGP_FMT_BASE64);
   if(ret != GNUTLS_E_SUCCESS){
     fprintf_plus(stderr,
@@ -530,33 +555,35 @@
   }
   
   /* GnuTLS server initialization */
-  ret = gnutls_dh_params_init(&mc.dh_params);
+  ret = gnutls_dh_params_init(&mc->dh_params);
   if(ret != GNUTLS_E_SUCCESS){
     fprintf_plus(stderr, "Error in GnuTLS DH parameter"
 		 " initialization: %s\n",
 		 safer_gnutls_strerror(ret));
     goto globalfail;
   }
-  ret = gnutls_dh_params_generate2(mc.dh_params, mc.dh_bits);
+  ret = gnutls_dh_params_generate2(mc->dh_params, mc->dh_bits);
   if(ret != GNUTLS_E_SUCCESS){
     fprintf_plus(stderr, "Error in GnuTLS prime generation: %s\n",
 		 safer_gnutls_strerror(ret));
     goto globalfail;
   }
   
-  gnutls_certificate_set_dh_params(mc.cred, mc.dh_params);
+  gnutls_certificate_set_dh_params(mc->cred, mc->dh_params);
   
   return 0;
   
  globalfail:
   
-  gnutls_certificate_free_credentials(mc.cred);
+  gnutls_certificate_free_credentials(mc->cred);
   gnutls_global_deinit();
-  gnutls_dh_params_deinit(mc.dh_params);
+  gnutls_dh_params_deinit(mc->dh_params);
   return -1;
 }
 
-static int init_gnutls_session(gnutls_session_t *session){
+__attribute__((nonnull, warn_unused_result))
+static int init_gnutls_session(gnutls_session_t *session,
+			       mandos_context *mc){
   int ret;
   /* GnuTLS session creation */
   do {
@@ -574,7 +601,7 @@
   {
     const char *err;
     do {
-      ret = gnutls_priority_set_direct(*session, mc.priority, &err);
+      ret = gnutls_priority_set_direct(*session, mc->priority, &err);
       if(quit_now){
 	gnutls_deinit(*session);
 	return -1;
@@ -591,7 +618,7 @@
   
   do {
     ret = gnutls_credentials_set(*session, GNUTLS_CRD_CERTIFICATE,
-				 mc.cred);
+				 mc->cred);
     if(quit_now){
       gnutls_deinit(*session);
       return -1;
@@ -607,7 +634,7 @@
   /* ignore client certificate if any. */
   gnutls_certificate_server_set_request(*session, GNUTLS_CERT_IGNORE);
   
-  gnutls_dh_set_prime_bits(*session, mc.dh_bits);
+  gnutls_dh_set_prime_bits(*session, mc->dh_bits);
   
   return 0;
 }
@@ -617,15 +644,13 @@
 		      __attribute__((unused)) const char *txt){}
 
 /* Called when a Mandos server is found */
-static int start_mandos_communication(const char *ip, uint16_t port,
+__attribute__((nonnull, warn_unused_result))
+static int start_mandos_communication(const char *ip, in_port_t port,
 				      AvahiIfIndex if_index,
-				      int af){
+				      int af, mandos_context *mc){
   int ret, tcp_sd = -1;
   ssize_t sret;
-  union {
-    struct sockaddr_in in;
-    struct sockaddr_in6 in6;
-  } to;
+  struct sockaddr_storage to;
   char *buffer = NULL;
   char *decrypted_buffer = NULL;
   size_t buffer_length = 0;
@@ -655,14 +680,46 @@
     return -1;
   }
   
-  ret = init_gnutls_session(&session);
+  /* If the interface is specified and we have a list of interfaces */
+  if(if_index != AVAHI_IF_UNSPEC and mc->interfaces != NULL){
+    /* Check if the interface is one of the interfaces we are using */
+    bool match = false;
+    {
+      char *interface = NULL;
+      while((interface=argz_next(mc->interfaces, mc->interfaces_size,
+				 interface))){
+	if(if_nametoindex(interface) == (unsigned int)if_index){
+	  match = true;
+	  break;
+	}
+      }
+    }
+    if(not match){
+      /* This interface does not match any in the list, so we don't
+	 connect to the server */
+      if(debug){
+	char interface[IF_NAMESIZE];
+	if(if_indextoname((unsigned int)if_index, interface) == NULL){
+	  perror_plus("if_indextoname");
+	} else {
+	  fprintf_plus(stderr, "Skipping server on non-used interface"
+		       " \"%s\"\n",
+		       if_indextoname((unsigned int)if_index,
+				      interface));
+	}
+      }
+      return -1;
+    }
+  }
+  
+  ret = init_gnutls_session(&session, mc);
   if(ret != 0){
     return -1;
   }
   
   if(debug){
     fprintf_plus(stderr, "Setting up a TCP connection to %s, port %"
-		 PRIu16 "\n", ip, port);
+		 PRIuMAX "\n", ip, (uintmax_t)port);
   }
   
   tcp_sd = socket(pf, SOCK_STREAM, 0);
@@ -680,11 +737,11 @@
   
   memset(&to, 0, sizeof(to));
   if(af == AF_INET6){
-    to.in6.sin6_family = (sa_family_t)af;
-    ret = inet_pton(af, ip, &to.in6.sin6_addr);
+    ((struct sockaddr_in6 *)&to)->sin6_family = (sa_family_t)af;
+    ret = inet_pton(af, ip, &((struct sockaddr_in6 *)&to)->sin6_addr);
   } else {			/* IPv4 */
-    to.in.sin_family = (sa_family_t)af;
-    ret = inet_pton(af, ip, &to.in.sin_addr);
+    ((struct sockaddr_in *)&to)->sin_family = (sa_family_t)af;
+    ret = inet_pton(af, ip, &((struct sockaddr_in *)&to)->sin_addr);
   }
   if(ret < 0 ){
     int e = errno;
@@ -699,13 +756,9 @@
     goto mandos_end;
   }
   if(af == AF_INET6){
-    to.in6.sin6_port = htons(port); /* Spurious warnings from
-				       -Wconversion and
-				       -Wunreachable-code */
-    
-    if(IN6_IS_ADDR_LINKLOCAL /* Spurious warnings from */
-       (&to.in6.sin6_addr)){ /* -Wstrict-aliasing=2 or lower and
-				-Wunreachable-code*/
+    ((struct sockaddr_in6 *)&to)->sin6_port = htons(port);
+    if(IN6_IS_ADDR_LINKLOCAL
+       (&((struct sockaddr_in6 *)&to)->sin6_addr)){
       if(if_index == AVAHI_IF_UNSPEC){
 	fprintf_plus(stderr, "An IPv6 link-local address is"
 		     " incomplete without a network interface\n");
@@ -713,12 +766,10 @@
 	goto mandos_end;
       }
       /* Set the network interface number as scope */
-      to.in6.sin6_scope_id = (uint32_t)if_index;
+      ((struct sockaddr_in6 *)&to)->sin6_scope_id = (uint32_t)if_index;
     }
   } else {
-    to.in.sin_port = htons(port); /* Spurious warnings from
-				     -Wconversion and
-				     -Wunreachable-code */
+    ((struct sockaddr_in *)&to)->sin_port = htons(port);
   }
   
   if(quit_now){
@@ -732,29 +783,32 @@
       if(if_indextoname((unsigned int)if_index, interface) == NULL){
 	perror_plus("if_indextoname");
       } else {
-	fprintf_plus(stderr, "Connection to: %s%%%s, port %" PRIu16
-		     "\n", ip, interface, port);
+	fprintf_plus(stderr, "Connection to: %s%%%s, port %" PRIuMAX
+		     "\n", ip, interface, (uintmax_t)port);
       }
     } else {
-      fprintf_plus(stderr, "Connection to: %s, port %" PRIu16 "\n",
-		   ip, port);
+      fprintf_plus(stderr, "Connection to: %s, port %" PRIuMAX "\n",
+		   ip, (uintmax_t)port);
     }
     char addrstr[(INET_ADDRSTRLEN > INET6_ADDRSTRLEN) ?
 		 INET_ADDRSTRLEN : INET6_ADDRSTRLEN] = "";
-    const char *pcret;
     if(af == AF_INET6){
-      pcret = inet_ntop(af, &(to.in6.sin6_addr), addrstr,
-			sizeof(addrstr));
+      ret = getnameinfo((struct sockaddr *)&to,
+			sizeof(struct sockaddr_in6),
+			addrstr, sizeof(addrstr), NULL, 0,
+			NI_NUMERICHOST);
     } else {
-      pcret = inet_ntop(af, &(to.in.sin_addr), addrstr,
-			sizeof(addrstr));
+      ret = getnameinfo((struct sockaddr *)&to,
+			sizeof(struct sockaddr_in),
+			addrstr, sizeof(addrstr), NULL, 0,
+			NI_NUMERICHOST);
     }
-    if(pcret == NULL){
-      perror_plus("inet_ntop");
-    } else {
-      if(strcmp(addrstr, ip) != 0){
-	fprintf_plus(stderr, "Canonical address form: %s\n", addrstr);
-      }
+    if(ret == EAI_SYSTEM){
+      perror_plus("getnameinfo");
+    } else if(ret != 0) {
+      fprintf_plus(stderr, "getnameinfo: %s", gai_strerror(ret));
+    } else if(strcmp(addrstr, ip) != 0){
+      fprintf_plus(stderr, "Canonical address form: %s\n", addrstr);
     }
   }
   
@@ -764,12 +818,14 @@
   }
   
   if(af == AF_INET6){
-    ret = connect(tcp_sd, &to.in6, sizeof(to));
+    ret = connect(tcp_sd, (struct sockaddr *)&to,
+		  sizeof(struct sockaddr_in6));
   } else {
-    ret = connect(tcp_sd, &to.in, sizeof(to)); /* IPv4 */
+    ret = connect(tcp_sd, (struct sockaddr *)&to, /* IPv4 */
+		  sizeof(struct sockaddr_in));
   }
   if(ret < 0){
-    if ((errno != ECONNREFUSED and errno != ENETUNREACH) or debug){
+    if((errno != ECONNREFUSED and errno != ENETUNREACH) or debug){
       int e = errno;
       perror_plus("connect");
       errno = e;
@@ -821,8 +877,11 @@
     goto mandos_end;
   }
   
-  /* Spurious warning from -Wint-to-pointer-cast */
-  gnutls_transport_set_ptr(session, (gnutls_transport_ptr_t) tcp_sd);
+  /* This casting via intptr_t is to eliminate warning about casting
+     an int to a pointer type.  This is exactly how the GnuTLS Guile
+     function "set-session-transport-fd!" does it. */
+  gnutls_transport_set_ptr(session,
+			   (gnutls_transport_ptr_t)(intptr_t)tcp_sd);
   
   if(quit_now){
     errno = EINTR;
@@ -933,7 +992,7 @@
   if(buffer_length > 0){
     ssize_t decrypted_buffer_size;
     decrypted_buffer_size = pgp_packet_decrypt(buffer, buffer_length,
-					       &decrypted_buffer);
+					       &decrypted_buffer, mc);
     if(decrypted_buffer_size >= 0){
       
       written = 0;
@@ -987,6 +1046,7 @@
   return retval;
 }
 
+__attribute__((nonnull))
 static void resolve_callback(AvahiSServiceResolver *r,
 			     AvahiIfIndex interface,
 			     AvahiProtocol proto,
@@ -1000,8 +1060,10 @@
 			     AVAHI_GCC_UNUSED AvahiStringList *txt,
 			     AVAHI_GCC_UNUSED AvahiLookupResultFlags
 			     flags,
-			     AVAHI_GCC_UNUSED void* userdata){
-  assert(r);
+			     void *mc){
+  if(r == NULL){
+    return;
+  }
   
   /* Called whenever a service has been resolved successfully or
      timed out */
@@ -1016,7 +1078,8 @@
     fprintf_plus(stderr, "(Avahi Resolver) Failed to resolve service "
 		 "'%s' of type '%s' in domain '%s': %s\n", name, type,
 		 domain,
-		 avahi_strerror(avahi_server_errno(mc.server)));
+		 avahi_strerror(avahi_server_errno
+				(((mandos_context*)mc)->server)));
     break;
     
   case AVAHI_RESOLVER_FOUND:
@@ -1028,13 +1091,16 @@
 		     PRIdMAX ") on port %" PRIu16 "\n", name,
 		     host_name, ip, (intmax_t)interface, port);
       }
-      int ret = start_mandos_communication(ip, port, interface,
-					   avahi_proto_to_af(proto));
+      int ret = start_mandos_communication(ip, (in_port_t)port,
+					   interface,
+					   avahi_proto_to_af(proto),
+					   mc);
       if(ret == 0){
-	avahi_simple_poll_quit(mc.simple_poll);
+	avahi_simple_poll_quit(simple_poll);
       } else {
-	if(not add_server(ip, port, interface,
-			  avahi_proto_to_af(proto))){
+	if(not add_server(ip, (in_port_t)port, interface,
+			  avahi_proto_to_af(proto),
+			  &((mandos_context*)mc)->current_server)){
 	  fprintf_plus(stderr, "Failed to add server \"%s\" to server"
 		       " list\n", name);
 	}
@@ -1053,8 +1119,10 @@
 			    const char *domain,
 			    AVAHI_GCC_UNUSED AvahiLookupResultFlags
 			    flags,
-			    AVAHI_GCC_UNUSED void* userdata){
-  assert(b);
+			    void *mc){
+  if(b == NULL){
+    return;
+  }
   
   /* Called whenever a new services becomes available on the LAN or
      is removed from the LAN */
@@ -1068,8 +1136,9 @@
   case AVAHI_BROWSER_FAILURE:
     
     fprintf_plus(stderr, "(Avahi browser) %s\n",
-		 avahi_strerror(avahi_server_errno(mc.server)));
-    avahi_simple_poll_quit(mc.simple_poll);
+		 avahi_strerror(avahi_server_errno
+				(((mandos_context*)mc)->server)));
+    avahi_simple_poll_quit(simple_poll);
     return;
     
   case AVAHI_BROWSER_NEW:
@@ -1078,12 +1147,14 @@
        the callback function is called the Avahi server will free the
        resolver for us. */
     
-    if(avahi_s_service_resolver_new(mc.server, interface, protocol,
-				    name, type, domain, protocol, 0,
-				    resolve_callback, NULL) == NULL)
+    if(avahi_s_service_resolver_new(((mandos_context*)mc)->server,
+				    interface, protocol, name, type,
+				    domain, protocol, 0,
+				    resolve_callback, mc) == NULL)
       fprintf_plus(stderr, "Avahi: Failed to resolve service '%s':"
 		   " %s\n", name,
-		   avahi_strerror(avahi_server_errno(mc.server)));
+		   avahi_strerror(avahi_server_errno
+				  (((mandos_context*)mc)->server)));
     break;
     
   case AVAHI_BROWSER_REMOVE:
@@ -1108,31 +1179,38 @@
   signal_received = sig;
   int old_errno = errno;
   /* set main loop to exit */
-  if(mc.simple_poll != NULL){
-    avahi_simple_poll_quit(mc.simple_poll);
+  if(simple_poll != NULL){
+    avahi_simple_poll_quit(simple_poll);
   }
   errno = old_errno;
 }
 
+__attribute__((nonnull, warn_unused_result))
 bool get_flags(const char *ifname, struct ifreq *ifr){
   int ret;
+  error_t ret_errno;
   
   int s = socket(PF_INET6, SOCK_DGRAM, IPPROTO_IP);
   if(s < 0){
+    ret_errno = errno;
     perror_plus("socket");
+    errno = ret_errno;
     return false;
   }
   strcpy(ifr->ifr_name, ifname);
   ret = ioctl(s, SIOCGIFFLAGS, ifr);
   if(ret == -1){
     if(debug){
+      ret_errno = errno;
       perror_plus("ioctl SIOCGIFFLAGS");
+      errno = ret_errno;
     }
     return false;
   }
   return true;
 }
 
+__attribute__((nonnull, warn_unused_result))
 bool good_flags(const char *ifname, const struct ifreq *ifr){
   
   /* Reject the loopback device */
@@ -1180,6 +1258,7 @@
  * corresponds to an acceptable network device.
  * (This function is passed to scandir(3) as a filter function.)
  */
+__attribute__((nonnull, warn_unused_result))
 int good_interface(const struct dirent *if_entry){
   if(if_entry->d_name[0] == '.'){
     return 0;
@@ -1201,48 +1280,40 @@
 }
 
 /* 
- * This function determines if a directory entry in /sys/class/net
- * corresponds to an acceptable network device which is up.
- * (This function is passed to scandir(3) as a filter function.)
- */
-int up_interface(const struct dirent *if_entry){
-  if(if_entry->d_name[0] == '.'){
-    return 0;
-  }
-  
-  struct ifreq ifr;
-  if(not get_flags(if_entry->d_name, &ifr)){
-    if(debug){
-      fprintf_plus(stderr, "Failed to get flags for interface "
-		   "\"%s\"\n", if_entry->d_name);
-    }
-    return 0;
-  }
-  
-  /* Reject down interfaces */
-  if(not (ifr.ifr_flags & IFF_UP)){
-    if(debug){
-      fprintf_plus(stderr, "Rejecting down interface \"%s\"\n",
-		   if_entry->d_name);
-    }
-    return 0;
-  }
-  
-  /* Reject non-running interfaces */
-  if(not (ifr.ifr_flags & IFF_RUNNING)){
-    if(debug){
-      fprintf_plus(stderr, "Rejecting non-running interface \"%s\"\n",
-		   if_entry->d_name);
-    }
-    return 0;
-  }
-  
-  if(not good_flags(if_entry->d_name, &ifr)){
-    return 0;
-  }
-  return 1;
-}
-
+ * This function determines if a network interface is up.
+ */
+__attribute__((nonnull, warn_unused_result))
+bool interface_is_up(const char *interface){
+  struct ifreq ifr;
+  if(not get_flags(interface, &ifr)){
+    if(debug){
+      fprintf_plus(stderr, "Failed to get flags for interface "
+		   "\"%s\"\n", interface);
+    }
+    return false;
+  }
+  
+  return (bool)(ifr.ifr_flags & IFF_UP);
+}
+
+/* 
+ * This function determines if a network interface is running
+ */
+__attribute__((nonnull, warn_unused_result))
+bool interface_is_running(const char *interface){
+  struct ifreq ifr;
+  if(not get_flags(interface, &ifr)){
+    if(debug){
+      fprintf_plus(stderr, "Failed to get flags for interface "
+		   "\"%s\"\n", interface);
+    }
+    return false;
+  }
+  
+  return (bool)(ifr.ifr_flags & IFF_RUNNING);
+}
+
+__attribute__((nonnull, pure, warn_unused_result))
 int notdotentries(const struct dirent *direntry){
   /* Skip "." and ".." */
   if(direntry->d_name[0] == '.'
@@ -1255,6 +1326,7 @@
 }
 
 /* Is this directory entry a runnable program? */
+__attribute__((nonnull, warn_unused_result))
 int runnable_hook(const struct dirent *direntry){
   int ret;
   size_t sret;
@@ -1268,7 +1340,7 @@
   sret = strspn(direntry->d_name, "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
 		"abcdefghijklmnopqrstuvwxyz"
 		"0123456789"
-		"_-");
+		"_.-");
   if((direntry->d_name)[sret] != '\0'){
     /* Contains non-allowed characters */
     if(debug){
@@ -1278,14 +1350,7 @@
     return 0;
   }
   
-  char *fullname = NULL;
-  ret = asprintf(&fullname, "%s/%s", hookdir, direntry->d_name);
-  if(ret < 0){
-    perror_plus("asprintf");
-    return 0;
-  }
-  
-  ret = stat(fullname, &st);
+  ret = fstatat(hookdir_fd, direntry->d_name, &st, 0);
   if(ret == -1){
     if(debug){
       perror_plus("Could not stat hook");
@@ -1315,21 +1380,23 @@
   return 1;
 }
 
-int avahi_loop_with_timeout(AvahiSimplePoll *s, int retry_interval){
+__attribute__((nonnull, warn_unused_result))
+int avahi_loop_with_timeout(AvahiSimplePoll *s, int retry_interval,
+			    mandos_context *mc){
   int ret;
   struct timespec now;
   struct timespec waited_time;
   intmax_t block_time;
   
   while(true){
-    if(mc.current_server == NULL){
-      if (debug){
+    if(mc->current_server == NULL){
+      if(debug){
 	fprintf_plus(stderr, "Wait until first server is found."
 		     " No timeout!\n");
       }
       ret = avahi_simple_poll_iterate(s, -1);
     } else {
-      if (debug){
+      if(debug){
 	fprintf_plus(stderr, "Check current_server if we should run"
 		     " it, or wait\n");
       }
@@ -1342,9 +1409,9 @@
       /* Calculating in ms how long time between now and server
 	 who we visted longest time ago. Now - last seen.  */
       waited_time.tv_sec = (now.tv_sec
-			    - mc.current_server->last_seen.tv_sec);
+			    - mc->current_server->last_seen.tv_sec);
       waited_time.tv_nsec = (now.tv_nsec
-			     - mc.current_server->last_seen.tv_nsec);
+			     - mc->current_server->last_seen.tv_nsec);
       /* total time is 10s/10,000ms.
 	 Converting to s from ms by dividing by 1,000,
 	 and ns to ms by dividing by 1,000,000. */
@@ -1352,27 +1419,27 @@
 		     - ((intmax_t)waited_time.tv_sec * 1000))
 		    - ((intmax_t)waited_time.tv_nsec / 1000000));
       
-      if (debug){
+      if(debug){
 	fprintf_plus(stderr, "Blocking for %" PRIdMAX " ms\n",
 		     block_time);
       }
       
       if(block_time <= 0){
-	ret = start_mandos_communication(mc.current_server->ip,
-					 mc.current_server->port,
-					 mc.current_server->if_index,
-					 mc.current_server->af);
+	ret = start_mandos_communication(mc->current_server->ip,
+					 mc->current_server->port,
+					 mc->current_server->if_index,
+					 mc->current_server->af, mc);
 	if(ret == 0){
-	  avahi_simple_poll_quit(mc.simple_poll);
+	  avahi_simple_poll_quit(s);
 	  return 0;
 	}
 	ret = clock_gettime(CLOCK_MONOTONIC,
-			    &mc.current_server->last_seen);
+			    &mc->current_server->last_seen);
 	if(ret == -1){
 	  perror_plus("clock_gettime");
 	  return -1;
 	}
-	mc.current_server = mc.current_server->next;
+	mc->current_server = mc->current_server->next;
 	block_time = 0; 	/* Call avahi to find new Mandos
 				   servers, but don't block */
       }
@@ -1380,167 +1447,460 @@
       ret = avahi_simple_poll_iterate(s, (int)block_time);
     }
     if(ret != 0){
-      if (ret > 0 or errno != EINTR){
+      if(ret > 0 or errno != EINTR){
 	return (ret != 1) ? ret : 0;
       }
     }
   }
 }
 
-bool run_network_hooks(const char *mode, const char *interface,
+/* Set effective uid to 0, return errno */
+__attribute__((warn_unused_result))
+error_t raise_privileges(void){
+  error_t old_errno = errno;
+  error_t ret_errno = 0;
+  if(seteuid(0) == -1){
+    ret_errno = errno;
+  }
+  errno = old_errno;
+  return ret_errno;
+}
+
+/* Set effective and real user ID to 0.  Return errno. */
+__attribute__((warn_unused_result))
+error_t raise_privileges_permanently(void){
+  error_t old_errno = errno;
+  error_t ret_errno = raise_privileges();
+  if(ret_errno != 0){
+    errno = old_errno;
+    return ret_errno;
+  }
+  if(setuid(0) == -1){
+    ret_errno = errno;
+  }
+  errno = old_errno;
+  return ret_errno;
+}
+
+/* Set effective user ID to unprivileged saved user ID */
+__attribute__((warn_unused_result))
+error_t lower_privileges(void){
+  error_t old_errno = errno;
+  error_t ret_errno = 0;
+  if(seteuid(uid) == -1){
+    ret_errno = errno;
+  }
+  errno = old_errno;
+  return ret_errno;
+}
+
+/* Lower privileges permanently */
+__attribute__((warn_unused_result))
+error_t lower_privileges_permanently(void){
+  error_t old_errno = errno;
+  error_t ret_errno = 0;
+  if(setuid(uid) == -1){
+    ret_errno = errno;
+  }
+  errno = old_errno;
+  return ret_errno;
+}
+
+__attribute__((nonnull))
+void run_network_hooks(const char *mode, const char *interface,
 		       const float delay){
-  struct dirent **direntries;
-  struct dirent *direntry;
-  int ret;
-  int numhooks = scandir(hookdir, &direntries, runnable_hook,
-			 alphasort);
+  struct dirent **direntries = NULL;
+  if(hookdir_fd == -1){
+    hookdir_fd = open(hookdir, O_RDONLY);
+    if(hookdir_fd == -1){
+      if(errno == ENOENT){
+	if(debug){
+	  fprintf_plus(stderr, "Network hook directory \"%s\" not"
+		       " found\n", hookdir);
+	}
+      } else {
+	perror_plus("open");
+      }
+      return;
+    }
+  }
+#ifdef __GLIBC__
+#if __GLIBC_PREREQ(2, 15)
+  int numhooks = scandirat(hookdir_fd, ".", &direntries,
+			   runnable_hook, alphasort);
+#else  /* not __GLIBC_PREREQ(2, 15) */
+  int numhooks = scandir(hookdir, &direntries, runnable_hook,
+			 alphasort);
+#endif	/* not __GLIBC_PREREQ(2, 15) */
+#else	/* not __GLIBC__ */
+  int numhooks = scandir(hookdir, &direntries, runnable_hook,
+			 alphasort);
+#endif	/* not __GLIBC__ */
   if(numhooks == -1){
     perror_plus("scandir");
-  } else {
-    int devnull = open("/dev/null", O_RDONLY);
-    for(int i = 0; i < numhooks; i++){
-      direntry = direntries[i];
-      char *fullname = NULL;
-      ret = asprintf(&fullname, "%s/%s", hookdir, direntry->d_name);
-      if(ret < 0){
+    return;
+  }
+  struct dirent *direntry;
+  int ret;
+  int devnull = open("/dev/null", O_RDONLY);
+  for(int i = 0; i < numhooks; i++){
+    direntry = direntries[i];
+    if(debug){
+      fprintf_plus(stderr, "Running network hook \"%s\"\n",
+		   direntry->d_name);
+    }
+    pid_t hook_pid = fork();
+    if(hook_pid == 0){
+      /* Child */
+      /* Raise privileges */
+      errno = raise_privileges_permanently();
+      if(errno != 0){
+	perror_plus("Failed to raise privileges");
+	_exit(EX_NOPERM);
+      }
+      /* Set group */
+      errno = 0;
+      ret = setgid(0);
+      if(ret == -1){
+	perror_plus("setgid");
+	_exit(EX_NOPERM);
+      }
+      /* Reset supplementary groups */
+      errno = 0;
+      ret = setgroups(0, NULL);
+      if(ret == -1){
+	perror_plus("setgroups");
+	_exit(EX_NOPERM);
+      }
+      ret = dup2(devnull, STDIN_FILENO);
+      if(ret == -1){
+	perror_plus("dup2(devnull, STDIN_FILENO)");
+	_exit(EX_OSERR);
+      }
+      ret = close(devnull);
+      if(ret == -1){
+	perror_plus("close");
+	_exit(EX_OSERR);
+      }
+      ret = dup2(STDERR_FILENO, STDOUT_FILENO);
+      if(ret == -1){
+	perror_plus("dup2(STDERR_FILENO, STDOUT_FILENO)");
+	_exit(EX_OSERR);
+      }
+      ret = setenv("MANDOSNETHOOKDIR", hookdir, 1);
+      if(ret == -1){
+	perror_plus("setenv");
+	_exit(EX_OSERR);
+      }
+      ret = setenv("DEVICE", interface, 1);
+      if(ret == -1){
+	perror_plus("setenv");
+	_exit(EX_OSERR);
+      }
+      ret = setenv("VERBOSITY", debug ? "1" : "0", 1);
+      if(ret == -1){
+	perror_plus("setenv");
+	_exit(EX_OSERR);
+      }
+      ret = setenv("MODE", mode, 1);
+      if(ret == -1){
+	perror_plus("setenv");
+	_exit(EX_OSERR);
+      }
+      char *delaystring;
+      ret = asprintf(&delaystring, "%f", (double)delay);
+      if(ret == -1){
 	perror_plus("asprintf");
-	continue;
-      }
-      if(debug){
-	fprintf_plus(stderr, "Running network hook \"%s\"\n",
-		     direntry->d_name);
-      }
-      pid_t hook_pid = fork();
-      if(hook_pid == 0){
-	/* Child */
-	/* Raise privileges */
-	errno = 0;
-	ret = seteuid(0);
-	if(ret == -1){
-	  perror_plus("seteuid");
-	}
-	/* Raise privileges even more */
-	errno = 0;
-	ret = setuid(0);
-	if(ret == -1){
-	  perror_plus("setuid");
-	}
-	/* Set group */
-	errno = 0;
-	ret = setgid(0);
-	if(ret == -1){
-	  perror_plus("setgid");
-	}
-	/* Reset supplementary groups */
-	errno = 0;
-	ret = setgroups(0, NULL);
-	if(ret == -1){
-	  perror_plus("setgroups");
-	}
-	dup2(devnull, STDIN_FILENO);
-	close(devnull);
-	dup2(STDERR_FILENO, STDOUT_FILENO);
-	ret = setenv("MANDOSNETHOOKDIR", hookdir, 1);
-	if(ret == -1){
-	  perror_plus("setenv");
-	  _exit(EX_OSERR);
-	}
-	ret = setenv("DEVICE", interface, 1);
-	if(ret == -1){
-	  perror_plus("setenv");
-	  _exit(EX_OSERR);
-	}
-	ret = setenv("VERBOSITY", debug ? "1" : "0", 1);
-	if(ret == -1){
-	  perror_plus("setenv");
-	  _exit(EX_OSERR);
-	}
-	ret = setenv("MODE", mode, 1);
-	if(ret == -1){
-	  perror_plus("setenv");
-	  _exit(EX_OSERR);
-	}
-	char *delaystring;
-	ret = asprintf(&delaystring, "%f", delay);
-	if(ret == -1){
-	  perror_plus("asprintf");
-	  _exit(EX_OSERR);
-	}
-	ret = setenv("DELAY", delaystring, 1);
-	if(ret == -1){
-	  free(delaystring);
-	  perror_plus("setenv");
-	  _exit(EX_OSERR);
-	}
+	_exit(EX_OSERR);
+      }
+      ret = setenv("DELAY", delaystring, 1);
+      if(ret == -1){
 	free(delaystring);
-	if(connect_to != NULL){
-	  ret = setenv("CONNECT", connect_to, 1);
-	  if(ret == -1){
-	    perror_plus("setenv");
-	    _exit(EX_OSERR);
-	  }
-	}
-	if(execl(fullname, direntry->d_name, mode, NULL) == -1){
-	  perror_plus("execl");
-	  _exit(EXIT_FAILURE);
-	}
-      } else {
-	int status;
-	if(TEMP_FAILURE_RETRY(waitpid(hook_pid, &status, 0)) == -1){
-	  perror_plus("waitpid");
-	  free(fullname);
-	  continue;
-	}
-	if(WIFEXITED(status)){
-	  if(WEXITSTATUS(status) != 0){
-	    fprintf_plus(stderr, "Warning: network hook \"%s\" exited"
-			 " with status %d\n", direntry->d_name,
-			 WEXITSTATUS(status));
-	    free(fullname);
-	    continue;
-	  }
-	} else if(WIFSIGNALED(status)){
-	  fprintf_plus(stderr, "Warning: network hook \"%s\" died by"
-		       " signal %d\n", direntry->d_name,
-		       WTERMSIG(status));
-	  free(fullname);
-	  continue;
-	} else {
-	  fprintf_plus(stderr, "Warning: network hook \"%s\""
-		       " crashed\n", direntry->d_name);
-	  free(fullname);
-	  continue;
-	}
-      }
-      free(fullname);
-      if(debug){
-	fprintf_plus(stderr, "Network hook \"%s\" ran successfully\n",
-		     direntry->d_name);
-      }
-    }
-    close(devnull);
-  }
-  return true;
+	perror_plus("setenv");
+	_exit(EX_OSERR);
+      }
+      free(delaystring);
+      if(connect_to != NULL){
+	ret = setenv("CONNECT", connect_to, 1);
+	if(ret == -1){
+	  perror_plus("setenv");
+	  _exit(EX_OSERR);
+	}
+      }
+      int hook_fd = openat(hookdir_fd, direntry->d_name, O_RDONLY);
+      if(hook_fd == -1){
+	perror_plus("openat");
+	_exit(EXIT_FAILURE);
+      }
+      if((int)TEMP_FAILURE_RETRY(close(hookdir_fd)) == -1){
+	perror_plus("close");
+	_exit(EXIT_FAILURE);
+      }
+      if(fexecve(hook_fd, (char *const []){ direntry->d_name, NULL },
+		 environ) == -1){
+	perror_plus("fexecve");
+	_exit(EXIT_FAILURE);
+      }
+    } else {
+      int status;
+      if(TEMP_FAILURE_RETRY(waitpid(hook_pid, &status, 0)) == -1){
+	perror_plus("waitpid");
+	continue;
+      }
+      if(WIFEXITED(status)){
+	if(WEXITSTATUS(status) != 0){
+	  fprintf_plus(stderr, "Warning: network hook \"%s\" exited"
+		       " with status %d\n", direntry->d_name,
+		       WEXITSTATUS(status));
+	  continue;
+	}
+      } else if(WIFSIGNALED(status)){
+	fprintf_plus(stderr, "Warning: network hook \"%s\" died by"
+		     " signal %d\n", direntry->d_name,
+		     WTERMSIG(status));
+	continue;
+      } else {
+	fprintf_plus(stderr, "Warning: network hook \"%s\""
+		     " crashed\n", direntry->d_name);
+	continue;
+      }
+    }
+    if(debug){
+      fprintf_plus(stderr, "Network hook \"%s\" ran successfully\n",
+		   direntry->d_name);
+    }
+  }
+  free(direntries);
+  if((int)TEMP_FAILURE_RETRY(close(hookdir_fd)) == -1){
+    perror_plus("close");
+  } else {
+    hookdir_fd = -1;
+  }
+  close(devnull);
+}
+
+__attribute__((nonnull, warn_unused_result))
+error_t bring_up_interface(const char *const interface,
+			   const float delay){
+  error_t old_errno = errno;
+  int ret;
+  struct ifreq network;
+  unsigned int if_index = if_nametoindex(interface);
+  if(if_index == 0){
+    fprintf_plus(stderr, "No such interface: \"%s\"\n", interface);
+    errno = old_errno;
+    return ENXIO;
+  }
+  
+  if(quit_now){
+    errno = old_errno;
+    return EINTR;
+  }
+  
+  if(not interface_is_up(interface)){
+    error_t ret_errno = 0, ioctl_errno = 0;
+    if(not get_flags(interface, &network)){
+      ret_errno = errno;
+      fprintf_plus(stderr, "Failed to get flags for interface "
+		   "\"%s\"\n", interface);
+      errno = old_errno;
+      return ret_errno;
+    }
+    network.ifr_flags |= IFF_UP; /* set flag */
+    
+    int sd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_IP);
+    if(sd == -1){
+      ret_errno = errno;
+      perror_plus("socket");
+      errno = old_errno;
+      return ret_errno;
+    }
+    
+    if(quit_now){
+      ret = (int)TEMP_FAILURE_RETRY(close(sd));
+      if(ret == -1){
+	perror_plus("close");
+      }
+      errno = old_errno;
+      return EINTR;
+    }
+    
+    if(debug){
+      fprintf_plus(stderr, "Bringing up interface \"%s\"\n",
+		   interface);
+    }
+    
+    /* Raise privileges */
+    ret_errno = raise_privileges();
+    if(ret_errno != 0){
+      errno = ret_errno;
+      perror_plus("Failed to raise privileges");
+    }
+    
+#ifdef __linux__
+    int ret_linux;
+    bool restore_loglevel = false;
+    if(ret_errno == 0){
+      /* Lower kernel loglevel to KERN_NOTICE to avoid KERN_INFO
+	 messages about the network interface to mess up the prompt */
+      ret_linux = klogctl(8, NULL, 5);
+      if(ret_linux == -1){
+	perror_plus("klogctl");
+      } else {
+	restore_loglevel = true;
+      }
+    }
+#endif	/* __linux__ */
+    int ret_setflags = ioctl(sd, SIOCSIFFLAGS, &network);
+    ioctl_errno = errno;
+#ifdef __linux__
+    if(restore_loglevel){
+      ret_linux = klogctl(7, NULL, 0);
+      if(ret_linux == -1){
+	perror_plus("klogctl");
+      }
+    }
+#endif	/* __linux__ */
+    
+    /* If raise_privileges() succeeded above */
+    if(ret_errno == 0){
+      /* Lower privileges */
+      ret_errno = lower_privileges();
+      if(ret_errno != 0){
+	errno = ret_errno;
+	perror_plus("Failed to lower privileges");
+      }
+    }
+    
+    /* Close the socket */
+    ret = (int)TEMP_FAILURE_RETRY(close(sd));
+    if(ret == -1){
+      perror_plus("close");
+    }
+    
+    if(ret_setflags == -1){
+      errno = ioctl_errno;
+      perror_plus("ioctl SIOCSIFFLAGS +IFF_UP");
+      errno = old_errno;
+      return ioctl_errno;
+    }
+  } else if(debug){
+    fprintf_plus(stderr, "Interface \"%s\" is already up; good\n",
+		 interface);
+  }
+  
+  /* Sleep checking until interface is running.
+     Check every 0.25s, up to total time of delay */
+  for(int i=0; i < delay * 4; i++){
+    if(interface_is_running(interface)){
+      break;
+    }
+    struct timespec sleeptime = { .tv_nsec = 250000000 };
+    ret = nanosleep(&sleeptime, NULL);
+    if(ret == -1 and errno != EINTR){
+      perror_plus("nanosleep");
+    }
+  }
+  
+  errno = old_errno;
+  return 0;
+}
+
+__attribute__((nonnull, warn_unused_result))
+error_t take_down_interface(const char *const interface){
+  error_t old_errno = errno;
+  struct ifreq network;
+  unsigned int if_index = if_nametoindex(interface);
+  if(if_index == 0){
+    fprintf_plus(stderr, "No such interface: \"%s\"\n", interface);
+    errno = old_errno;
+    return ENXIO;
+  }
+  if(interface_is_up(interface)){
+    error_t ret_errno = 0, ioctl_errno = 0;
+    if(not get_flags(interface, &network) and debug){
+      ret_errno = errno;
+      fprintf_plus(stderr, "Failed to get flags for interface "
+		   "\"%s\"\n", interface);
+      errno = old_errno;
+      return ret_errno;
+    }
+    network.ifr_flags &= ~(short)IFF_UP; /* clear flag */
+    
+    int sd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_IP);
+    if(sd == -1){
+      ret_errno = errno;
+      perror_plus("socket");
+      errno = old_errno;
+      return ret_errno;
+    }
+    
+    if(debug){
+      fprintf_plus(stderr, "Taking down interface \"%s\"\n",
+		   interface);
+    }
+    
+    /* Raise privileges */
+    ret_errno = raise_privileges();
+    if(ret_errno != 0){
+      errno = ret_errno;
+      perror_plus("Failed to raise privileges");
+    }
+    
+    int ret_setflags = ioctl(sd, SIOCSIFFLAGS, &network);
+    ioctl_errno = errno;
+    
+    /* If raise_privileges() succeeded above */
+    if(ret_errno == 0){
+      /* Lower privileges */
+      ret_errno = lower_privileges();
+      if(ret_errno != 0){
+	errno = ret_errno;
+	perror_plus("Failed to lower privileges");
+      }
+    }
+    
+    /* Close the socket */
+    int ret = (int)TEMP_FAILURE_RETRY(close(sd));
+    if(ret == -1){
+      perror_plus("close");
+    }
+    
+    if(ret_setflags == -1){
+      errno = ioctl_errno;
+      perror_plus("ioctl SIOCSIFFLAGS -IFF_UP");
+      errno = old_errno;
+      return ioctl_errno;
+    }
+  } else if(debug){
+    fprintf_plus(stderr, "Interface \"%s\" is already down; odd\n",
+		 interface);
+  }
+  
+  errno = old_errno;
+  return 0;
 }
 
 int main(int argc, char *argv[]){
+  mandos_context mc = { .server = NULL, .dh_bits = 1024,
+			.priority = "SECURE256:!CTYPE-X.509:"
+			"+CTYPE-OPENPGP", .current_server = NULL,
+			.interfaces = NULL, .interfaces_size = 0 };
   AvahiSServiceBrowser *sb = NULL;
-  int error;
+  error_t ret_errno;
   int ret;
   intmax_t tmpmax;
   char *tmp;
   int exitcode = EXIT_SUCCESS;
-  const char *interface = "";
-  struct ifreq network;
-  int sd = -1;
-  bool take_down_interface = false;
-  uid_t uid;
-  gid_t gid;
-  char tempdir[] = "/tmp/mandosXXXXXX";
-  bool tempdir_created = false;
+  char *interfaces_to_take_down = NULL;
+  size_t interfaces_to_take_down_size = 0;
+  char run_tempdir[] = "/run/tmp/mandosXXXXXX";
+  char old_tempdir[] = "/tmp/mandosXXXXXX";
+  char *tempdir = NULL;
   AvahiIfIndex if_index = AVAHI_IF_UNSPEC;
   const char *seckey = PATHDIR "/" SECKEY;
   const char *pubkey = PATHDIR "/" PUBKEY;
+  char *interfaces_hooks = NULL;
   
   bool gnutls_initialized = false;
   bool gpgme_initialized = false;
@@ -1637,7 +1997,11 @@
 	connect_to = arg;
 	break;
       case 'i':			/* --interface */
-	interface = arg;
+	ret_errno = argz_add_sep(&mc.interfaces, &mc.interfaces_size,
+				 arg, (int)',');
+	if(ret_errno != 0){
+	  argp_error(state, "%s", strerror(ret_errno));
+	}
 	break;
       case 's':			/* --seckey */
 	seckey = arg;
@@ -1720,11 +2084,11 @@
     /* Work around Debian bug #633582:
        <http://bugs.debian.org/633582> */
     
-    /* Re-raise priviliges */
-    errno = 0;
-    ret = seteuid(0);
-    if(ret == -1){
-      perror_plus("seteuid");
+    /* Re-raise privileges */
+    ret_errno = raise_privileges();
+    if(ret_errno != 0){
+      errno = ret_errno;
+      perror_plus("Failed to raise privileges");
     } else {
       struct stat st;
       
@@ -1771,61 +2135,56 @@
       }
     
       /* Lower privileges */
-      errno = 0;
-      ret = seteuid(uid);
-      if(ret == -1){
-	perror_plus("seteuid");
+      ret_errno = lower_privileges();
+      if(ret_errno != 0){
+	errno = ret_errno;
+	perror_plus("Failed to lower privileges");
+      }
+    }
+  }
+  
+  /* Remove invalid interface names (except "none") */
+  {
+    char *interface = NULL;
+    while((interface = argz_next(mc.interfaces, mc.interfaces_size,
+				 interface))){
+      if(strcmp(interface, "none") != 0
+	 and if_nametoindex(interface) == 0){
+	if(interface[0] != '\0'){
+	  fprintf_plus(stderr, "Not using nonexisting interface"
+		       " \"%s\"\n", interface);
+	}
+	argz_delete(&mc.interfaces, &mc.interfaces_size, interface);
+	interface = NULL;
       }
     }
   }
   
   /* Run network hooks */
-  if(not run_network_hooks("start", interface, delay)){
-    goto end;
+  {
+    if(mc.interfaces != NULL){
+      interfaces_hooks = malloc(mc.interfaces_size);
+      if(interfaces_hooks == NULL){
+	perror_plus("malloc");
+	goto end;
+      }
+      memcpy(interfaces_hooks, mc.interfaces, mc.interfaces_size);
+      argz_stringify(interfaces_hooks, mc.interfaces_size, (int)',');
+    }
+    run_network_hooks("start", interfaces_hooks != NULL ?
+		      interfaces_hooks : "", delay);
   }
   
   if(not debug){
     avahi_set_log_function(empty_log);
   }
   
-  if(interface[0] == '\0'){
-    struct dirent **direntries;
-    /* First look for interfaces that are up */
-    ret = scandir(sys_class_net, &direntries, up_interface,
-		  alphasort);
-    if(ret == 0){
-      /* No up interfaces, look for any good interfaces */
-      free(direntries);
-      ret = scandir(sys_class_net, &direntries, good_interface,
-		    alphasort);
-    }
-    if(ret >= 1){
-      /* Pick the first interface returned */
-      interface = strdup(direntries[0]->d_name);
-      if(debug){
-	fprintf_plus(stderr, "Using interface \"%s\"\n", interface);
-      }
-      if(interface == NULL){
-	perror_plus("malloc");
-	free(direntries);
-	exitcode = EXIT_FAILURE;
-	goto end;
-      }
-      free(direntries);
-    } else {
-      free(direntries);
-      fprintf_plus(stderr, "Could not find a network interface\n");
-      exitcode = EXIT_FAILURE;
-      goto end;
-    }
-  }
-  
   /* Initialize Avahi early so avahi_simple_poll_quit() can be called
      from the signal handler */
   /* Initialize the pseudo-RNG for Avahi */
   srand((unsigned int) time(NULL));
-  mc.simple_poll = avahi_simple_poll_new();
-  if(mc.simple_poll == NULL){
+  simple_poll = avahi_simple_poll_new();
+  if(simple_poll == NULL){
     fprintf_plus(stderr,
 		 "Avahi: Failed to create simple poll object.\n");
     exitcode = EX_UNAVAILABLE;
@@ -1895,148 +2254,93 @@
     }
   }
   
-  /* If the interface is down, bring it up */
-  if(strcmp(interface, "none") != 0){
-    if_index = (AvahiIfIndex) if_nametoindex(interface);
-    if(if_index == 0){
-      fprintf_plus(stderr, "No such interface: \"%s\"\n", interface);
-      exitcode = EX_UNAVAILABLE;
-      goto end;
-    }
-    
-    if(quit_now){
-      goto end;
-    }
-    
-    /* Re-raise priviliges */
-    errno = 0;
-    ret = seteuid(0);
-    if(ret == -1){
-      perror_plus("seteuid");
-    }
-    
-#ifdef __linux__
-    /* Lower kernel loglevel to KERN_NOTICE to avoid KERN_INFO
-       messages about the network interface to mess up the prompt */
-    ret = klogctl(8, NULL, 5);
-    bool restore_loglevel = true;
-    if(ret == -1){
-      restore_loglevel = false;
-      perror_plus("klogctl");
-    }
-#endif	/* __linux__ */
-    
-    sd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_IP);
-    if(sd < 0){
-      perror_plus("socket");
-      exitcode = EX_OSERR;
-#ifdef __linux__
-      if(restore_loglevel){
-	ret = klogctl(7, NULL, 0);
-	if(ret == -1){
-	  perror_plus("klogctl");
-	}
-      }
-#endif	/* __linux__ */
-      /* Lower privileges */
-      errno = 0;
-      ret = seteuid(uid);
-      if(ret == -1){
-	perror_plus("seteuid");
-      }
-      goto end;
-    }
-    strcpy(network.ifr_name, interface);
-    ret = ioctl(sd, SIOCGIFFLAGS, &network);
-    if(ret == -1){
-      perror_plus("ioctl SIOCGIFFLAGS");
-#ifdef __linux__
-      if(restore_loglevel){
-	ret = klogctl(7, NULL, 0);
-	if(ret == -1){
-	  perror_plus("klogctl");
-	}
-      }
-#endif	/* __linux__ */
-      exitcode = EX_OSERR;
-      /* Lower privileges */
-      errno = 0;
-      ret = seteuid(uid);
-      if(ret == -1){
-	perror_plus("seteuid");
-      }
-      goto end;
-    }
-    if((network.ifr_flags & IFF_UP) == 0){
-      network.ifr_flags |= IFF_UP;
-      take_down_interface = true;
-      ret = ioctl(sd, SIOCSIFFLAGS, &network);
-      if(ret == -1){
-	take_down_interface = false;
-	perror_plus("ioctl SIOCSIFFLAGS +IFF_UP");
-	exitcode = EX_OSERR;
-#ifdef __linux__
-	if(restore_loglevel){
-	  ret = klogctl(7, NULL, 0);
-	  if(ret == -1){
-	    perror_plus("klogctl");
+  /* If no interfaces were specified, make a list */
+  if(mc.interfaces == NULL){
+    struct dirent **direntries = NULL;
+    /* Look for any good interfaces */
+    ret = scandir(sys_class_net, &direntries, good_interface,
+		  alphasort);
+    if(ret >= 1){
+      /* Add all found interfaces to interfaces list */
+      for(int i = 0; i < ret; ++i){
+	ret_errno = argz_add(&mc.interfaces, &mc.interfaces_size,
+			     direntries[i]->d_name);
+	if(ret_errno != 0){
+	  errno = ret_errno;
+	  perror_plus("argz_add");
+	  continue;
+	}
+	if(debug){
+	  fprintf_plus(stderr, "Will use interface \"%s\"\n",
+		       direntries[i]->d_name);
+	}
+      }
+      free(direntries);
+    } else {
+      if(ret == 0){
+	free(direntries);
+      }
+      fprintf_plus(stderr, "Could not find a network interface\n");
+      exitcode = EXIT_FAILURE;
+      goto end;
+    }
+  }
+  
+  /* Bring up interfaces which are down, and remove any "none"s */
+  {
+    char *interface = NULL;
+    while((interface = argz_next(mc.interfaces, mc.interfaces_size,
+				 interface))){
+      /* If interface name is "none", stop bringing up interfaces.
+	 Also remove all instances of "none" from the list */
+      if(strcmp(interface, "none") == 0){
+	argz_delete(&mc.interfaces, &mc.interfaces_size,
+		    interface);
+	interface = NULL;
+	while((interface = argz_next(mc.interfaces,
+				     mc.interfaces_size, interface))){
+	  if(strcmp(interface, "none") == 0){
+	    argz_delete(&mc.interfaces, &mc.interfaces_size,
+			interface);
+	    interface = NULL;
 	  }
 	}
-#endif	/* __linux__ */
-	/* Lower privileges */
-	errno = 0;
-	ret = seteuid(uid);
-	if(ret == -1){
-	  perror_plus("seteuid");
-	}
-	goto end;
-      }
-    }
-    /* Sleep checking until interface is running.
-       Check every 0.25s, up to total time of delay */
-    for(int i=0; i < delay * 4; i++){
-      ret = ioctl(sd, SIOCGIFFLAGS, &network);
-      if(ret == -1){
-	perror_plus("ioctl SIOCGIFFLAGS");
-      } else if(network.ifr_flags & IFF_RUNNING){
 	break;
       }
-      struct timespec sleeptime = { .tv_nsec = 250000000 };
-      ret = nanosleep(&sleeptime, NULL);
-      if(ret == -1 and errno != EINTR){
-	perror_plus("nanosleep");
-      }
-    }
-    if(not take_down_interface){
-      /* We won't need the socket anymore */
-      ret = (int)TEMP_FAILURE_RETRY(close(sd));
-      if(ret == -1){
-	perror_plus("close");
-      }
-    }
-#ifdef __linux__
-    if(restore_loglevel){
-      /* Restores kernel loglevel to default */
-      ret = klogctl(7, NULL, 0);
-      if(ret == -1){
-	perror_plus("klogctl");
-      }
-    }
-#endif	/* __linux__ */
-    /* Lower privileges */
-    errno = 0;
-    /* Lower privileges */
-    ret = seteuid(uid);
-    if(ret == -1){
-      perror_plus("seteuid");
-    }
+      bool interface_was_up = interface_is_up(interface);
+      errno = bring_up_interface(interface, delay);
+      if(not interface_was_up){
+	if(errno != 0){
+	  perror_plus("Failed to bring up interface");
+	} else {
+	  errno = argz_add(&interfaces_to_take_down,
+			   &interfaces_to_take_down_size,
+			   interface);
+	  if(errno != 0){
+	    perror_plus("argz_add");
+	  }
+	}
+      }
+    }
+    if(debug and (interfaces_to_take_down == NULL)){
+      fprintf_plus(stderr, "No interfaces were brought up\n");
+    }
+  }
+  
+  /* If we only got one interface, explicitly use only that one */
+  if(argz_count(mc.interfaces, mc.interfaces_size) == 1){
+    if(debug){
+      fprintf_plus(stderr, "Using only interface \"%s\"\n",
+		   mc.interfaces);
+    }
+    if_index = (AvahiIfIndex)if_nametoindex(mc.interfaces);
   }
   
   if(quit_now){
     goto end;
   }
   
-  ret = init_gnutls_global(pubkey, seckey);
+  ret = init_gnutls_global(pubkey, seckey, &mc);
   if(ret == -1){
     fprintf_plus(stderr, "init_gnutls_global failed\n");
     exitcode = EX_UNAVAILABLE;
@@ -2049,17 +2353,25 @@
     goto end;
   }
   
-  if(mkdtemp(tempdir) == NULL){
+  /* Try /run/tmp before /tmp */
+  tempdir = mkdtemp(run_tempdir);
+  if(tempdir == NULL and errno == ENOENT){
+      if(debug){
+	fprintf_plus(stderr, "Tempdir %s did not work, trying %s\n",
+		     run_tempdir, old_tempdir);
+      }
+      tempdir = mkdtemp(old_tempdir);
+  }
+  if(tempdir == NULL){
     perror_plus("mkdtemp");
     goto end;
   }
-  tempdir_created = true;
   
   if(quit_now){
     goto end;
   }
   
-  if(not init_gpgme(pubkey, seckey, tempdir)){
+  if(not init_gpgme(pubkey, seckey, tempdir, &mc)){
     fprintf_plus(stderr, "init_gpgme failed\n");
     exitcode = EX_UNAVAILABLE;
     goto end;
@@ -2075,6 +2387,7 @@
     /* Connect directly, do not use Zeroconf */
     /* (Mainly meant for debugging) */
     char *address = strrchr(connect_to, ':');
+    
     if(address == NULL){
       fprintf_plus(stderr, "No colon in address\n");
       exitcode = EX_USAGE;
@@ -2085,21 +2398,21 @@
       goto end;
     }
     
-    uint16_t port;
+    in_port_t port;
     errno = 0;
     tmpmax = strtoimax(address+1, &tmp, 10);
     if(errno != 0 or tmp == address+1 or *tmp != '\0'
-       or tmpmax != (uint16_t)tmpmax){
+       or tmpmax != (in_port_t)tmpmax){
       fprintf_plus(stderr, "Bad port number\n");
       exitcode = EX_USAGE;
       goto end;
     }
-  
+    
     if(quit_now){
       goto end;
     }
     
-    port = (uint16_t)tmpmax;
+    port = (in_port_t)tmpmax;
     *address = '\0';
     /* Colon in address indicates IPv6 */
     int af;
@@ -2121,7 +2434,8 @@
     }
     
     while(not quit_now){
-      ret = start_mandos_communication(address, port, if_index, af);
+      ret = start_mandos_communication(address, port, if_index, af,
+				       &mc);
       if(quit_now or ret == 0){
 	break;
       }
@@ -2129,10 +2443,10 @@
 	fprintf_plus(stderr, "Retrying in %d seconds\n",
 		     (int)retry_interval);
       }
-      sleep((int)retry_interval);
+      sleep((unsigned int)retry_interval);
     }
     
-    if (not quit_now){
+    if(not quit_now){
       exitcode = EXIT_SUCCESS;
     }
     
@@ -2153,9 +2467,8 @@
     config.publish_domain = 0;
     
     /* Allocate a new server */
-    mc.server = avahi_server_new(avahi_simple_poll_get
-				 (mc.simple_poll), &config, NULL,
-				 NULL, &error);
+    mc.server = avahi_server_new(avahi_simple_poll_get(simple_poll),
+				 &config, NULL, NULL, &ret_errno);
     
     /* Free the Avahi configuration data */
     avahi_server_config_free(&config);
@@ -2164,7 +2477,7 @@
   /* Check if creating the Avahi server object succeeded */
   if(mc.server == NULL){
     fprintf_plus(stderr, "Failed to create Avahi server: %s\n",
-		 avahi_strerror(error));
+		 avahi_strerror(ret_errno));
     exitcode = EX_UNAVAILABLE;
     goto end;
   }
@@ -2176,7 +2489,8 @@
   /* Create the Avahi service browser */
   sb = avahi_s_service_browser_new(mc.server, if_index,
 				   AVAHI_PROTO_UNSPEC, "_mandos._tcp",
-				   NULL, 0, browse_callback, NULL);
+				   NULL, 0, browse_callback,
+				   (void *)&mc);
   if(sb == NULL){
     fprintf_plus(stderr, "Failed to create service browser: %s\n",
 		 avahi_strerror(avahi_server_errno(mc.server)));
@@ -2193,9 +2507,9 @@
   if(debug){
     fprintf_plus(stderr, "Starting Avahi loop search\n");
   }
-
-  ret = avahi_loop_with_timeout(mc.simple_poll,
-				(int)(retry_interval * 1000));
+  
+  ret = avahi_loop_with_timeout(simple_poll,
+				(int)(retry_interval * 1000), &mc);
   if(debug){
     fprintf_plus(stderr, "avahi_loop_with_timeout exited %s\n",
 		 (ret == 0) ? "successfully" : "with error");
@@ -2208,14 +2522,16 @@
   }
   
   /* Cleanup things */
+  free(mc.interfaces);
+  
   if(sb != NULL)
     avahi_s_service_browser_free(sb);
   
   if(mc.server != NULL)
     avahi_server_free(mc.server);
   
-  if(mc.simple_poll != NULL)
-    avahi_simple_poll_free(mc.simple_poll);
+  if(simple_poll != NULL)
+    avahi_simple_poll_free(simple_poll);
   
   if(gnutls_initialized){
     gnutls_certificate_free_credentials(mc.cred);
@@ -2238,75 +2554,88 @@
     }
   }
   
-  /* Run network hooks */
-  run_network_hooks("stop", interface, delay);
-  
-  /* Re-raise priviliges */
+  /* Re-raise privileges */
   {
-    errno = 0;
-    ret = seteuid(0);
-    if(ret == -1){
-      perror_plus("seteuid");
+    ret_errno = raise_privileges();
+    if(ret_errno != 0){
+      errno = ret_errno;
+      perror_plus("Failed to raise privileges");
+    } else {
+      
+      /* Run network hooks */
+      run_network_hooks("stop", interfaces_hooks != NULL ?
+			interfaces_hooks : "", delay);
+      
+      /* Take down the network interfaces which were brought up */
+      {
+	char *interface = NULL;
+	while((interface=argz_next(interfaces_to_take_down,
+				   interfaces_to_take_down_size,
+				   interface))){
+	  ret_errno = take_down_interface(interface);
+	  if(ret_errno != 0){
+	    errno = ret_errno;
+	    perror_plus("Failed to take down interface");
+	  }
+	}
+	if(debug and (interfaces_to_take_down == NULL)){
+	  fprintf_plus(stderr, "No interfaces needed to be taken"
+		       " down\n");
+	}
+      }
     }
     
-    /* Take down the network interface */
-    if(take_down_interface and geteuid() == 0){
-      ret = ioctl(sd, SIOCGIFFLAGS, &network);
-      if(ret == -1){
-	perror_plus("ioctl SIOCGIFFLAGS");
-      } else if(network.ifr_flags & IFF_UP){
-	network.ifr_flags &= ~(short)IFF_UP; /* clear flag */
-	ret = ioctl(sd, SIOCSIFFLAGS, &network);
-	if(ret == -1){
-	  perror_plus("ioctl SIOCSIFFLAGS -IFF_UP");
-	}
-      }
-      ret = (int)TEMP_FAILURE_RETRY(close(sd));
-      if(ret == -1){
-	perror_plus("close");
-      }
+    ret_errno = lower_privileges_permanently();
+    if(ret_errno != 0){
+      errno = ret_errno;
+      perror_plus("Failed to lower privileges permanently");
     }
   }
-  /* Lower privileges permanently */
-  errno = 0;
-  ret = setuid(uid);
-  if(ret == -1){
-    perror_plus("setuid");
-  }
+  
+  free(interfaces_to_take_down);
+  free(interfaces_hooks);
   
   /* Removes the GPGME temp directory and all files inside */
-  if(tempdir_created){
+  if(tempdir != NULL){
     struct dirent **direntries = NULL;
-    struct dirent *direntry = NULL;
-    int numentries = scandir(tempdir, &direntries, notdotentries,
-			     alphasort);
-    if (numentries > 0){
-      for(int i = 0; i < numentries; i++){
-	direntry = direntries[i];
-	char *fullname = NULL;
-	ret = asprintf(&fullname, "%s/%s", tempdir,
-		       direntry->d_name);
-	if(ret < 0){
-	  perror_plus("asprintf");
-	  continue;
-	}
-	ret = remove(fullname);
-	if(ret == -1){
-	  fprintf_plus(stderr, "remove(\"%s\"): %s\n", fullname,
-		       strerror(errno));
-	}
-	free(fullname);
+    int tempdir_fd = (int)TEMP_FAILURE_RETRY(open(tempdir, O_RDONLY |
+						  O_NOFOLLOW));
+    if(tempdir_fd == -1){
+      perror_plus("open");
+    } else {
+#ifdef __GLIBC__
+#if __GLIBC_PREREQ(2, 15)
+      int numentries = scandirat(tempdir_fd, ".", &direntries,
+				 notdotentries, alphasort);
+#else  /* not __GLIBC_PREREQ(2, 15) */
+      int numentries = scandir(tempdir, &direntries, notdotentries,
+			       alphasort);
+#endif	/* not __GLIBC_PREREQ(2, 15) */
+#else	/* not __GLIBC__ */
+      int numentries = scandir(tempdir, &direntries, notdotentries,
+			       alphasort);
+#endif	/* not __GLIBC__ */
+      if(numentries >= 0){
+	for(int i = 0; i < numentries; i++){
+	  ret = unlinkat(tempdir_fd, direntries[i]->d_name, 0);
+	  if(ret == -1){
+	    fprintf_plus(stderr, "unlinkat(open(\"%s\", O_RDONLY),"
+			 " \"%s\", 0): %s\n", tempdir,
+			 direntries[i]->d_name, strerror(errno));
+	  }
+	}
+	
+	/* need to clean even if 0 because man page doesn't specify */
+	free(direntries);
+	if(numentries == -1){
+	  perror_plus("scandir");
+	}
+	ret = rmdir(tempdir);
+	if(ret == -1 and errno != ENOENT){
+	  perror_plus("rmdir");
+	}
       }
-    }
-
-    /* need to clean even if 0 because man page doesn't specify */
-    free(direntries);
-    if (numentries == -1){
-      perror_plus("scandir");
-    }
-    ret = rmdir(tempdir);
-    if(ret == -1 and errno != ENOENT){
-      perror_plus("rmdir");
+      TEMP_FAILURE_RETRY(close(tempdir_fd));
     }
   }
   

=== modified file 'plugins.d/mandos-client.xml'
--- plugins.d/mandos-client.xml	2011-12-31 23:05:34 +0000
+++ plugins.d/mandos-client.xml	2014-06-22 02:19:30 +0000
@@ -2,7 +2,7 @@
 <!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
 	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
 <!ENTITY COMMANDNAME "mandos-client">
-<!ENTITY TIMESTAMP "2012-01-01">
+<!ENTITY TIMESTAMP "2014-06-22">
 <!ENTITY % common SYSTEM "../common.ent">
 %common;
 ]>
@@ -34,6 +34,8 @@
       <year>2008</year>
       <year>2009</year>
       <year>2012</year>
+      <year>2013</year>
+      <year>2014</year>
       <holder>Teddy Hogeborn</holder>
       <holder>Björn Påhlsson</holder>
     </copyright>
@@ -64,11 +66,13 @@
 	><replaceable>PORT</replaceable></option></arg>
       </group>
       <sbr/>
-      <group>
+      <group rep='repeat'>
 	<arg choice="plain"><option>--interface
-	<replaceable>NAME</replaceable></option></arg>
-	<arg choice="plain"><option>-i
-	<replaceable>NAME</replaceable></option></arg>
+	<replaceable>NAME</replaceable><arg rep='repeat'
+	>,<replaceable>NAME</replaceable></arg></option></arg>
+	<arg choice="plain"><option>-i <replaceable>NAME</replaceable
+	><arg rep='repeat'>,<replaceable>NAME</replaceable></arg
+	></option></arg>
       </group>
       <sbr/>
       <group>
@@ -137,10 +141,10 @@
       communicates with <citerefentry><refentrytitle
       >mandos</refentrytitle><manvolnum>8</manvolnum></citerefentry>
       to get a password.  In slightly more detail, this client program
-      brings up a network interface, uses the interface’s IPv6
-      link-local address to get network connectivity, uses Zeroconf to
-      find servers on the local network, and communicates with servers
-      using TLS with an OpenPGP key to ensure authenticity and
+      brings up network interfaces, uses the interfaces’ IPv6
+      link-local addresses to get network connectivity, uses Zeroconf
+      to find servers on the local network, and communicates with
+      servers using TLS with an OpenPGP key to ensure authenticity and
       confidentiality.  This client program keeps running, trying all
       servers on the network, until it receives a satisfactory reply
       or a TERM signal.  After all servers have been tried, all
@@ -148,23 +152,21 @@
       will wait indefinitely for new servers to appear.
     </para>
     <para>
-      The network interface is selected like this: If an interface is
-      specified using the <option>--interface</option> option, that
-      interface is used.  Otherwise, <command>&COMMANDNAME;</command>
-      will choose any interface that is up and running and is not a
-      loopback interface, is not a point-to-point interface, is
-      capable of broadcasting and does not have the NOARP flag (see
+      The network interfaces are selected like this: If any interfaces
+      are specified using the <option>--interface</option> option,
+      those interface are used.  Otherwise,
+      <command>&COMMANDNAME;</command> will use all interfaces that
+      are not loopback interfaces, are not point-to-point interfaces,
+      are capable of broadcasting and do not have the NOARP flag (see
       <citerefentry><refentrytitle>netdevice</refentrytitle>
       <manvolnum>7</manvolnum></citerefentry>).  (If the
       <option>--connect</option> option is used, point-to-point
-      interfaces and non-broadcast interfaces are accepted.)  If no
-      acceptable interfaces are found, re-run the check but without
-      the <quote>up and running</quote> requirement, and manually take
-      the selected interface up (and later take it down on program
-      exit).
+      interfaces and non-broadcast interfaces are accepted.)  If any
+      used interfaces are not up and running, they are first taken up
+      (and later taken down again on program exit).
     </para>
     <para>
-      Before a network interface is selected, all <quote>network
+      Before network interfaces are selected, all <quote>network
       hooks</quote> are run; see <xref linkend="network-hooks"/>.
     </para>
     <para>
@@ -218,45 +220,50 @@
 	    assumed to separate the address from the port number.
 	  </para>
 	  <para>
-	    This option is normally only useful for testing and
-	    debugging.
+	    Normally, Zeroconf would be used to locate Mandos servers,
+	    in which case this option would only be used when testing
+	    and debugging.
 	  </para>
 	</listitem>
       </varlistentry>
       
       <varlistentry>
 	<term><option>--interface=<replaceable
-	>NAME</replaceable></option></term>
+	>NAME</replaceable><arg rep='repeat'>,<replaceable
+	>NAME</replaceable></arg></option></term>
 	<term><option>-i
-	<replaceable>NAME</replaceable></option></term>
+	<replaceable>NAME</replaceable><arg rep='repeat'>,<replaceable
+	>NAME</replaceable></arg></option></term>
 	<listitem>
 	  <para>
-	    Network interface that will be brought up and scanned for
-	    Mandos servers to connect to.  The default is the empty
-	    string, which will automatically choose an appropriate
-	    interface.
+	    Comma separated list of network interfaces that will be
+	    brought up and scanned for Mandos servers to connect to.
+	    The default is the empty string, which will automatically
+	    use all appropriate interfaces.
 	  </para>
 	  <para>
-	    If the <option>--connect</option> option is used, this
-	    specifies the interface to use to connect to the address
-	    given.
+	    If the <option>--connect</option> option is used, and
+	    exactly one interface name is specified (except
+	    <quote><literal>none</literal></quote>), this specifies
+	    the interface to use to connect to the address given.
 	  </para>
 	  <para>
 	    Note that since this program will normally run in the
 	    initial RAM disk environment, the interface must be an
 	    interface which exists at that stage.  Thus, the interface
-	    can not be a pseudo-interface such as <quote>br0</quote>
-	    or <quote>tun0</quote>; such interfaces will not exist
-	    until much later in the boot process, and can not be used
-	    by this program, unless created by a <quote>network
-	    hook</quote> — see <xref linkend="network-hooks"/>.
+	    can normally not be a pseudo-interface such as
+	    <quote>br0</quote> or <quote>tun0</quote>; such interfaces
+	    will not exist until much later in the boot process, and
+	    can not be used by this program, unless created by a
+	    <quote>network hook</quote> — see <xref
+	    linkend="network-hooks"/>.
 	  </para>
 	  <para>
 	    <replaceable>NAME</replaceable> can be the string
-	    <quote><literal>none</literal></quote>; this will not use
-	    any specific interface, and will not bring up an interface
-	    on startup.  This is not recommended, and only meant for
-	    advanced users.
+	    <quote><literal>none</literal></quote>; this will make
+	    <command>&COMMANDNAME;</command> only bring up interfaces
+	    specified <emphasis>before</emphasis> this string.  This
+	    is not recommended, and only meant for advanced users.
 	  </para>
 	</listitem>
       </varlistentry>
@@ -314,7 +321,7 @@
 	>SECONDS</replaceable></option></term>
 	<listitem>
 	  <para>
-	    After bringing the network interface up, the program waits
+	    After bringing a network interface up, the program waits
 	    for the interface to arrive in a <quote>running</quote>
 	    state before proceeding.  During this time, the kernel log
 	    level will be lowered to reduce clutter on the system
@@ -507,7 +514,7 @@
 	      It is not necessary to print any non-executable files
 	      already in the network hook directory, these will be
 	      copied implicitly if they otherwise satisfy the name
-	      requirement.
+	      requirements.
 	    </para>
 	  </listitem>
 	</varlistentry>
@@ -547,11 +554,12 @@
 	  <term><envar>DEVICE</envar></term>
 	  <listitem>
 	    <para>
-	      The network interface, as specified to
+	      The network interfaces, as specified to
 	      <command>&COMMANDNAME;</command> by the
-	      <option>--interface</option> option.  If this is not the
-	      interface a hook will bring up, there is no reason for a
-	      hook to continue.
+	      <option>--interface</option> option, combined to one
+	      string and separated by commas.  If this is set, and
+	      does not contain the interface a hook will bring up,
+	      there is no reason for a hook to continue.
 	    </para>
 	  </listitem>
 	</varlistentry>
@@ -660,8 +668,8 @@
     </para>
     <informalexample>
       <para>
-	Normal invocation needs no options, if the network interface
-	is <quote>eth0</quote>:
+	Normal invocation needs no options, if the network interfaces
+	can be automatically determined:
       </para>
       <para>
 	<userinput>&COMMANDNAME;</userinput>
@@ -669,8 +677,8 @@
     </informalexample>
     <informalexample>
       <para>
-	Search for Mandos servers (and connect to them) using another
-	interface:
+	Search for Mandos servers (and connect to them) using one
+	specific interface:
       </para>
       <para>
 	<!-- do not wrap this line -->
@@ -740,8 +748,9 @@
     <para>
       It will also help if the checker program on the server is
       configured to request something from the client which can not be
-      spoofed by someone else on the network, unlike unencrypted
-      <acronym>ICMP</acronym> echo (<quote>ping</quote>) replies.
+      spoofed by someone else on the network, like SSH server key
+      fingerprints, and unlike unencrypted <acronym>ICMP</acronym>
+      echo (<quote>ping</quote>) replies.
     </para>
     <para>
       <emphasis>Note</emphasis>: This makes it completely insecure to
@@ -840,7 +849,7 @@
 	      <para>
 		This client uses IPv6 link-local addresses, which are
 		immediately usable since a link-local addresses is
-		automatically assigned to a network interfaces when it
+		automatically assigned to a network interface when it
 		is brought up.
 	      </para>
 	    </listitem>

=== modified file 'plugins.d/password-prompt.c'
--- plugins.d/password-prompt.c	2011-12-31 23:05:34 +0000
+++ plugins.d/password-prompt.c	2014-03-29 02:38:15 +0000
@@ -2,8 +2,8 @@
 /*
  * Password-prompt - Read a password from the terminal and print it
  * 
- * Copyright © 2008-2012 Teddy Hogeborn
- * Copyright © 2008-2012 Björn Påhlsson
+ * Copyright © 2008-2014 Teddy Hogeborn
+ * Copyright © 2008-2014 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -72,16 +72,6 @@
 /* Needed for conflict resolution */
 const char plymouth_name[] = "plymouthd";
 
-__attribute__((format (gnu_printf, 2, 3), nonnull(1)))
-int fprintf_plus(FILE *stream, const char *format, ...){
-  va_list ap;
-  va_start (ap, format);
-  
-  TEMP_FAILURE_RETRY(fprintf(stream, "Mandos plugin %s: ",
-			     program_invocation_short_name));
-  return TEMP_FAILURE_RETRY(vfprintf(stream, format, ap));
-}
-
 /* Function to use when printing errors */
 __attribute__((format (gnu_printf, 3, 4)))
 void error_plus(int status, int errnum, const char *formatstring,
@@ -92,7 +82,7 @@
   
   va_start(ap, formatstring);
   ret = vasprintf(&text, formatstring, ap);
-  if (ret == -1){
+  if(ret == -1){
     fprintf(stderr, "Mandos plugin %s: ",
 	    program_invocation_short_name);
     vfprintf(stderr, formatstring, ap);
@@ -222,7 +212,7 @@
   struct dirent **direntries = NULL;
   int ret;
   ret = scandir("/proc", &direntries, is_plymouth, alphasort);
-  if (ret == -1){
+  if(ret == -1){
     error_plus(1, errno, "scandir");
   }
   free(direntries);
@@ -313,7 +303,7 @@
     fprintf(stderr, "Starting %s\n", argv[0]);
   }
 
-  if (conflict_detection()){
+  if(conflict_detection()){
     if(debug){
       fprintf(stderr, "Stopping %s because of conflict\n", argv[0]);
     }

=== modified file 'plugins.d/plymouth.c'
--- plugins.d/plymouth.c	2011-12-31 23:05:34 +0000
+++ plugins.d/plymouth.c	2014-03-29 02:38:15 +0000
@@ -2,8 +2,8 @@
 /*
  * Plymouth - Read a password from Plymouth and output it
  * 
- * Copyright © 2010-2012 Teddy Hogeborn
- * Copyright © 2010-2012 Björn Påhlsson
+ * Copyright © 2010-2014 Teddy Hogeborn
+ * Copyright © 2010-2014 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -84,7 +84,7 @@
   
   va_start(ap, formatstring);
   ret = vasprintf(&text, formatstring, ap);
-  if (ret == -1){
+  if(ret == -1){
     fprintf(stderr, "Mandos plugin %s: ",
 	    program_invocation_short_name);
     vfprintf(stderr, formatstring, ap);
@@ -156,7 +156,7 @@
 
 __attribute__((nonnull (2, 3)))
 bool exec_and_wait(pid_t *pid_return, const char *path,
-		   const char **argv, bool interruptable,
+		   const char * const *argv, bool interruptable,
 		   bool daemonize){
   int status;
   int ret;
@@ -179,7 +179,7 @@
     int i = 0;
     for (; argv[i]!=NULL; i++){
       tmp = realloc(new_argv, sizeof(const char *) * ((size_t)i + 1));
-      if (tmp == NULL){
+      if(tmp == NULL){
 	error_plus(0, errno, "realloc");
 	free(new_argv);
 	_exit(EX_OSERR);
@@ -290,12 +290,12 @@
   if(proc_id == 0){
     struct dirent **direntries = NULL;
     ret = scandir("/proc", &direntries, is_plymouth, alphasort);
-    if (ret == -1){
+    if(ret == -1){
       error_plus(0, errno, "scandir");
     }
-    if (ret > 0){
+    if(ret > 0){
       ret = sscanf(direntries[0]->d_name, "%" SCNuMAX, &proc_id);
-      if (ret < 0){
+      if(ret < 0){
 	error_plus(0, errno, "sscanf");
       }
     }
@@ -312,7 +312,7 @@
   return 0;
 }
 
-const char **getargv(pid_t pid){
+const char * const * getargv(pid_t pid){
   int cl_fd;
   char *cmdline_filename;
   ssize_t sret;
@@ -379,7 +379,7 @@
     return NULL;
   }
   argz_extract(cmdline, cmdline_len, argv); /* Create argv */
-  return (const char **)argv;
+  return (const char * const *)argv;
 }
 
 int main(__attribute__((unused))int argc,
@@ -460,7 +460,7 @@
   }
   kill_and_wait(plymouth_command_pid);
   
-  const char **plymouthd_argv;
+  const char * const *plymouthd_argv;
   pid_t pid = get_pid();
   if(pid == 0){
     error_plus(0, 0, "plymouthd pid not found");

=== modified file 'plugins.d/splashy.c'
--- plugins.d/splashy.c	2011-12-31 23:05:34 +0000
+++ plugins.d/splashy.c	2014-03-29 02:38:15 +0000
@@ -2,8 +2,8 @@
 /*
  * Splashy - Read a password from splashy and output it
  * 
- * Copyright © 2008-2012 Teddy Hogeborn
- * Copyright © 2008-2012 Björn Påhlsson
+ * Copyright © 2008-2014 Teddy Hogeborn
+ * Copyright © 2008-2014 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -70,7 +70,7 @@
   
   va_start(ap, formatstring);
   ret = vasprintf(&text, formatstring, ap);
-  if (ret == -1){
+  if(ret == -1){
     fprintf(stderr, "Mandos plugin %s: ",
 	    program_invocation_short_name);
     vfprintf(stderr, formatstring, ap);

=== modified file 'plugins.d/usplash.c'
--- plugins.d/usplash.c	2011-12-31 23:05:34 +0000
+++ plugins.d/usplash.c	2014-03-29 02:38:15 +0000
@@ -2,8 +2,8 @@
 /*
  * Usplash - Read a password from usplash and output it
  * 
- * Copyright © 2008-2012 Teddy Hogeborn
- * Copyright © 2008-2012 Björn Påhlsson
+ * Copyright © 2008-2014 Teddy Hogeborn
+ * Copyright © 2008-2014 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -67,7 +67,7 @@
   
   va_start(ap, formatstring);
   ret = vasprintf(&text, formatstring, ap);
-  if (ret == -1){
+  if(ret == -1){
     fprintf(stderr, "Mandos plugin %s: ",
 	    program_invocation_short_name);
     vfprintf(stderr, formatstring, ap);

